/**
* Orbit controls for CanvsMap.  This file was originally "OrbitControls.js" and was provided
* with Three.js.  However, the class needed to be modified to work with CanvasMap, specifically
* to deal with selections, which are handled within the view.  This meant that the view needs
* to filter the selections within the view and only pass them to the controls if nothing was selected
*
* This class uses a THREE.Quaternion() class to manage the motions with spherical coordinates.
*
* target - the coordinate that the view is orbiting around (the center)
* TheObject - the current camera position
*
* Private properties:
* - scope - a copy of the this pointer
*
* @module CM3ControlsOrbit
* @Copyright HSU, Jim Graham, 2019
*/

// This set of controls performs orbiting, dollying (zooming), and panning.
// Unlike TrackballControls, it maintains the "up" direction this.TheObject.up (+Y by default).
//
//    Orbit - left mouse / touch: one finger move
//    Zoom - middle mouse, or mousewheel / touch: two finger spread or squish
//    Pan - right mouse, or arrow keys / touch: three finter swipe

function CM3ControlsOrbit() 
{
	CM3Controls.call(this);
	
	//***************************************************************************
	// Settings
	//***************************************************************************
	this.Settings.CM3ControlsOrbit=	
	{
	};
	//***************************************************************************
	// Private definitions
	//***************************************************************************
	// The four arrow keys
	this.keys={ LEFT: 37, UP: 38, RIGHT: 39, BOTTOM: 40 };

	// Mouse buttons
	this.mouseButtons={ ORBIT: THREE.MOUSE.LEFT, ZOOM: THREE.MOUSE.MIDDLE, PAN: THREE.MOUSE.RIGHT };

	//***************************************************************************
	// Private properties (most of these should be settings)
	//***************************************************************************
	// Set to false to disable this control
	this.enabled=true;

	// "target" sets the location of focus, where the this.TheObject orbits around
	this.target=new THREE.Vector3();

	// How far you can zoom in and out ( OrthographicCamera only )
	this.minZoom=0;
	this.maxZoom=Infinity;

	// How far you can orbit vertically, upper and lower limits.
	// Range is 0 to Math.PI radians.
	//this.minPolarAngle=0; // radians
//	this.maxPolarAngle=Math.PI/2; // radians

	// How far you can orbit horizontally, upper and lower limits.
	// If set, must be a sub-interval of the interval [ - Math.PI, Math.PI ].
	this.minAzimuthAngle=- Infinity; // radians
	this.maxAzimuthAngle=Infinity; // radians

	// Set to true to enable damping (inertia)
	// If damping is enabled, you must call controls.update() in your animation loop
	this.enableDamping=true;
	this.dampingFactor=0.25;

	// This option actually enables dollying in and out; left as "zoom" for backwards compatibility.
	// Set to false to disable zooming
	this.enableZoom=true;
	this.zoomSpeed=1.0;

	// Set to false to disable rotating
	this.enableRotate=true;
	this.rotateSpeed=1.0;

	// Set to false to disable panning
	this.enablePan=true;
	this.keyPanSpeed=7.0;	// pixels moved per arrow key push

	// Set to true to automatically rotate around the target
	// If auto-rotate is enabled, you must call controls.update() in your animation loop
	this.autoRotate=false;
	this.autoRotateSpeed=2.0; // 30 seconds per round when fps is 60

	// Set to false to disable use of the keys
	this.enableKeys=true;

	//***********************************************************************************
	// private properties
	//***********************************************************************************

	var scope=this; // used by event functions to find the this pointer

	var changeEvent={ type: 'change' };
	var startEvent={ type: 'start' };
	var endEvent={ type: 'end' };

	var STATE={ NONE : - 1, ROTATE : 0, DOLLY : 1, PAN : 2, TOUCH_ROTATE : 3, TOUCH_DOLLY : 4, TOUCH_PAN : 5 };

	var state=STATE.NONE;

	// acronum for epsilon, which just represents a small number
	// the view will only be updated if the this.TheObject's position has changed more than this value
	var EPS=0.000001;

	// current position in spherical coordinates
	//var spherical=new THREE.Spherical();
	var sphericalDelta=new THREE.Spherical();

	// when dollying in and out, scale is set as a fraction (i.e. 1.1, 0.9) and then the radius is mulitplied by the scale in Update() and then scale is reset.
	var scale=1; 
	var panOffset=new THREE.Vector3();
	var zoomChanged=false;

	var rotateStart=new THREE.Vector2();
	var rotateEnd=new THREE.Vector2();
	var rotateDelta=new THREE.Vector2();

	var panStart=new THREE.Vector2();
	var panEnd=new THREE.Vector2();
	var panDelta=new THREE.Vector2();

	var dollyStart=new THREE.Vector2();
	var dollyEnd=new THREE.Vector2();
	var dollyDelta=new THREE.Vector2();
	//***********************************************************************************
	// private methods
	//***********************************************************************************

	this.getPolarAngle=function () {

		return phi;

	};

	this.getAzimuthalAngle=function () {

		return theta;

	};

	this.reset=function () {

		scope.target.copy( scope.target0 );
		scope.TheObject.position.copy( scope.position0 );
		scope.TheObject.zoom=scope.zoom0;

		scope.TheObject.updateProjectionMatrix();
		//scope.dispatchEvent( changeEvent );

		scope.Update();

		state=STATE.NONE;

	};
	this.getState=function() { return(state); } 
	
	//***********************************************************************************
	// private functions
	//***********************************************************************************

	function getAutoRotationAngle() {

		return 2 * Math.PI / 60 / 60 * scope.autoRotateSpeed;

	}

	function getZoomScale() 
	{
		return Math.pow( 0.95, scope.zoomSpeed );
	}

	//*************************************************************************
	// Private functions? 
	//*************************************************************************
	function rotateLeft( angle ) {

		sphericalDelta.theta -= angle;

	}

	function rotateUp( angle ) {

		sphericalDelta.phi -= angle;

	}

	var panLeft=function() {

		var v=new THREE.Vector3();

		return function panLeft( distance, objectMatrix ) {

			v.setFromMatrixColumn( objectMatrix, 0 ); // get X column of objectMatrix
			v.multiplyScalar( - distance );

			panOffset.add( v );

		};

	}();

	var panUp=function() {

		var v=new THREE.Vector3();

		return function panUp( distance, objectMatrix ) {

			v.setFromMatrixColumn( objectMatrix, 1 ); // get Y column of objectMatrix
			v.multiplyScalar( distance );

			panOffset.add( v );

		};

	}();

	// deltaX and deltaY are in pixels; right and down are positive
	var pan=function() 
	{
		var offset=new THREE.Vector3();

		return function( deltaX, deltaY ) {

			var element=scope.domElement === document ? scope.domElement.body : scope.domElement;

			if ( scope.TheObject instanceof THREE.PerspectiveCamera ) 
			{
				// perspective
				var position=scope.TheObject.position;
				offset.copy( position ).sub( scope.target );
				var targetDistance=offset.length();

				// half of the fov is center to top of screen
				targetDistance *= Math.tan( ( scope.TheObject.fov / 2 ) * Math.PI / 180.0 );

				// we actually don't use screenWidth, since perspective camera is fixed to screen height
				panLeft( 2 * deltaX * targetDistance / element.clientHeight, scope.TheObject.matrix );
				panUp( 2 * deltaY * targetDistance / element.clientHeight, scope.TheObject.matrix );

			} else if ( scope.TheObject instanceof THREE.OrthographicCamera ) {

				// orthographic
				panLeft( deltaX * ( scope.TheObject.right - scope.TheObject.left ) / scope.TheObject.zoom / element.clientWidth, scope.TheObject.matrix );
				panUp( deltaY * ( scope.TheObject.top - scope.TheObject.bottom ) / scope.TheObject.zoom / element.clientHeight, scope.TheObject.matrix );

			} else {

				// camera neither orthographic nor perspective
				console.warn( 'WARNING: OrbitControls.js encountered an unknown camera type - pan disabled.' );
				scope.enablePan=false;

			}

		};

	}();

	/**
	* @private
	*/
	function dollyIn( dollyScale ) {

		if ( scope.TheObject instanceof THREE.PerspectiveCamera ) {

			scale /= dollyScale;

		} else if ( scope.TheObject instanceof THREE.OrthographicCamera ) {

			scope.TheObject.zoom=Math.max( scope.minZoom, Math.min( scope.maxZoom, scope.TheObject.zoom * dollyScale ) );
			scope.TheObject.updateProjectionMatrix();
			zoomChanged=true;

		} else {

			console.warn( 'WARNING: OrbitControls.js encountered an unknown camera type - dolly/zoom disabled.' );
			scope.enableZoom=false;

		}

	}

	/**
	* @private
	*/
	function dollyOut( dollyScale ) {

		if ( scope.TheObject instanceof THREE.PerspectiveCamera ) {

			scale *= dollyScale;

		} else if ( scope.TheObject instanceof THREE.OrthographicCamera ) {

			scope.TheObject.zoom=Math.max( scope.minZoom, Math.min( scope.maxZoom, scope.TheObject.zoom / dollyScale ) );
			scope.TheObject.updateProjectionMatrix();
			zoomChanged=true;

		} else {

			console.warn( 'WARNING: OrbitControls.js encountered an unknown camera type - dolly/zoom disabled.' );
			scope.enableZoom=false;

		}

	}

	//*************************************************************************
	// Private event functions
	// These are called by the event handlers
	//*************************************************************************
	/*
	* @private
	*/
	this.handleMouseMoveRotate=function ( event ) 
	{
		rotateEnd.set( event.clientX, event.clientY );
		rotateDelta.subVectors( rotateEnd, rotateStart );

if (rotateDelta.y<0)
{
	var j=12;
}
		// find the amount to scale the angles based on Min/Max distances (e.g. our distance above the earth)
		var ScalingFactor=1;
		
		var MinDistance=this.GetSetting("CM3ControlsOrbit","MinDistance",0);
		var MaxDistance=this.GetSetting("CM3ControlsOrbit","MaxDistance",Infinity);
		
		if ((MaxDistance!=Infinity)&&(MinDistance!=0))
		{
			var Range=MaxDistance-MinDistance;
			
			var element=scope.domElement === document ? scope.domElement.body : scope.domElement;
	
			if ( scope.TheObject instanceof THREE.PerspectiveCamera ) 
			{
				// perspective
				var position=scope.TheObject.position;
				var offset=new THREE.Vector3();
				offset.copy( position ).sub( scope.target );
				var spherical=new THREE.Spherical();
				spherical.setFromVector3( offset );
				
				ScalingFactor=(spherical.radius-MinDistance)/Range;
			}
		}
		// find the fraction of 360 degrees we are moving (-1 to 1?)
		var HorizontalChange=rotateDelta.x / scope.domElement.clientWidth * scope.rotateSpeed*ScalingFactor;
		var VerticalChange=rotateDelta.y / scope.domElement.clientHeight * scope.rotateSpeed*ScalingFactor;
		
		// rotating across whole screen goes 360 degrees around
		rotateLeft( 2 * Math.PI *  HorizontalChange);

		// rotating up and down along whole screen attempts to go 360, but limited to 180
		rotateUp( 2 * Math.PI *  VerticalChange);

		rotateStart.copy( rotateEnd );

		scope.Update();

	}

	function handleMouseMoveDolly( event ) {

		//console.log( 'handleMouseMoveDolly' );

		dollyEnd.set( event.clientX, event.clientY );

		dollyDelta.subVectors( dollyEnd, dollyStart );

		if ( dollyDelta.y > 0 ) {

			dollyIn( getZoomScale() );

		} else if ( dollyDelta.y < 0 ) {

			dollyOut( getZoomScale() );

		}

		dollyStart.copy( dollyEnd );

		scope.Update();

	}

	function handleMouseMovePan( event ) {

		//console.log( 'handleMouseMovePan' );

		panEnd.set( event.clientX, event.clientY );

		panDelta.subVectors( panEnd, panStart );

		pan( panDelta.x, panDelta.y );

		panStart.copy( panEnd );

		scope.Update();

	}

	function handleTouchStartRotate( event ) {

		//console.log( 'handleTouchStartRotate' );

		rotateStart.set( event.touches[ 0 ].pageX, event.touches[ 0 ].pageY );

	}

	function handleTouchStartDolly( event ) {

		//console.log( 'handleTouchStartDolly' );

		var dx=event.touches[ 0 ].pageX - event.touches[ 1 ].pageX;
		var dy=event.touches[ 0 ].pageY - event.touches[ 1 ].pageY;

		var distance=Math.sqrt( dx * dx + dy * dy );

		dollyStart.set( 0, distance );

	}

	function handleTouchStartPan( event ) {

		//console.log( 'handleTouchStartPan' );

		panStart.set( event.touches[ 0 ].pageX, event.touches[ 0 ].pageY );

	}

	function handleTouchMoveRotate( event ) {

		//console.log( 'handleTouchMoveRotate' );

		rotateEnd.set( event.touches[ 0 ].pageX, event.touches[ 0 ].pageY );
		rotateDelta.subVectors( rotateEnd, rotateStart );

		var element=scope.domElement === document ? scope.domElement.body : scope.domElement;

		// rotating across whole screen goes 360 degrees around
		rotateLeft( 2 * Math.PI * rotateDelta.x / element.clientWidth * scope.rotateSpeed );

		// rotating up and down along whole screen attempts to go 360, but limited to 180
		rotateUp( 2 * Math.PI * rotateDelta.y / element.clientHeight * scope.rotateSpeed );

		rotateStart.copy( rotateEnd );

		scope.Update();

	}

	function handleTouchMoveDolly( event ) {

		//console.log( 'handleTouchMoveDolly' );

		var dx=event.touches[ 0 ].pageX - event.touches[ 1 ].pageX;
		var dy=event.touches[ 0 ].pageY - event.touches[ 1 ].pageY;

		var distance=Math.sqrt( dx * dx + dy * dy );

		dollyEnd.set( 0, distance );

		dollyDelta.subVectors( dollyEnd, dollyStart );

		if ( dollyDelta.y > 0 ) {

			dollyOut( getZoomScale() );

		} else if ( dollyDelta.y < 0 ) {

			dollyIn( getZoomScale() );

		}

		dollyStart.copy( dollyEnd );

		scope.Update();

	}

	function handleTouchMovePan( event ) {

		//console.log( 'handleTouchMovePan' );

		panEnd.set( event.touches[ 0 ].pageX, event.touches[ 0 ].pageY );

		panDelta.subVectors( panEnd, panStart );

		pan( panDelta.x, panDelta.y );

		panStart.copy( panEnd );

		scope.Update();

	}

	function handleTouchEnd( event ) {

		//console.log( 'handleTouchEnd' );

	}

	//***********************************************************************************
	// Object methods
	//***********************************************************************************
	this.dispose=function() {

	};

	//***************************************************************************
	// Protected CM3Control Functions
	//***************************************************************************
	this.Initialize=function( TheObject, domElement ) 
	{
		// this is the object that is being moved by the controls (typically a camera)
		this.TheObject=TheObject;
	
		// dom element the scene is displayed in (typically the Canvas)
		this.domElement=( domElement !== undefined ) ? domElement : document;
		// for reset
		this.target0=this.target.clone();
		this.position0=this.TheObject.position.clone();
		this.zoom0=this.TheObject.zoom;

		// force an update at start
	
		this.Update();

	}

	/**
	* Called by event handlers after settings have changed to update the position and orientation
	* of the camera.
	* @private
	*/
	this.Update=function() 
	{
		var Result=false;
		
		var offset=new THREE.Vector3();

		// so camera.up is the orbit axis
		var quat=new THREE.Quaternion().setFromUnitVectors( this.TheObject.up, new THREE.Vector3( 0, 1, 0 ) );
		var quatInverse=quat.clone().inverse();

		var lastPosition=new THREE.Vector3();
		var lastQuaternion=new THREE.Quaternion();

		// get the current position of the camera
		var position=scope.TheObject.position;

		offset.copy( position ).sub( scope.target ); // find the offset vectors from the target to the camera position

		// rotate offset to "y-axis-is-up" space
		offset.applyQuaternion( quat );

		//*************************************************
		// get the new camera position in spherical coordinates (radius, phi, theta)
		
		// angle from z-axis around y-axis
		var spherical=new THREE.Spherical();
		spherical.setFromVector3( offset );

		if ( scope.autoRotate && state === STATE.NONE ) 
		{
			rotateLeft( getAutoRotationAngle() );
		}

		// Rotate the cameras position based on current deltas
		spherical.theta += sphericalDelta.theta;
		spherical.phi += sphericalDelta.phi;

		//*************************************************
		// restrict theta to be between desired limits
		spherical.theta=Math.max( scope.minAzimuthAngle, Math.min( scope.maxAzimuthAngle, spherical.theta ) );

		// restrict phi to be between desired limits
		var MinLatitude=this.GetSetting("CM3ControlsOrbit","MinLatitude",-90);
		var MaxLatitude=this.GetSetting("CM3ControlsOrbit","MaxLatitude",90);
		
		var MinPolarAngle=(MinLatitude+90)/45;
		var MaxPolarAngle=(MaxLatitude+90)/45;
		
		//spherical.phi=Math.max( MinPolarAngle, Math.min( MaxPolarAngle, spherical.phi ) );

		spherical.makeSafe();

		var MinDistance=this.GetSetting("CM3ControlsOrbit","MinDistance",0);
		var MaxDistance=this.GetSetting("CM3ControlsOrbit","MaxDistance",Infinity);
		
		if ((MaxDistance!=Infinity)&&(MinDistance!=0))
		{
			var Temp=spherical.radius-MinDistance;
			Temp*= scale;
			spherical.radius=MinDistance+Temp;
		}
		else
		{
			spherical.radius *= scale;
		}
		
		scale=1; // reset for the next update

		// restrict radius to be between desired limits
		
		spherical.radius=Math.max( MinDistance, Math.min( MaxDistance, spherical.radius ) );

		//*************************************************
		// move target to panned location
		scope.target.add( panOffset );
		
		panOffset.set( 0, 0, 0 ); // reset for the next update

		//*************************************************
		// find the new position of the camera
		offset.setFromSpherical( spherical );

		// rotate offset back to "camera-up-vector-is-up" space
		offset.applyQuaternion( quatInverse );

		// move the camera to the new position by taking the target's position and adding an offset to the camera
		position.copy( scope.target ).add( offset );

		var MinElevation=this.GetSetting("CM3Controls","MinElevation",0);
		var MaxElevation=this.GetSetting("CM3Controls","MaxElevation",Infinity);
		
		var Up=scope.TheObject.up;
		
		if (Up.y==1)
		{
			if (position.y<MinElevation) position.y=MinElevation;
			if (position.y>MaxElevation) position.y=MaxElevation;
		}
		else
		{
			if (position.z<MinElevation) position.z=MinElevation;
			if (position.z>MaxElevation) position.z=MaxElevation;
		}
		// turn the camera to look at the target
		scope.TheObject.lookAt( scope.target );

		//*************************************************
		// reduce the rate of turning for the next time if damping enabled
		// otherwise, stop the turning
		
		if ( scope.enableDamping === true ) 
		{
			sphericalDelta.theta *= ( 1 - scope.dampingFactor );
			sphericalDelta.phi *= ( 1 - scope.dampingFactor );
		} 
		else 
		{
			sphericalDelta.set( 0, 0, 0 );
		}

		//*************************************************
		// update condition is:
		// min(camera displacement, camera rotation in radians)^2 > EPS
		// using small-angle approximation cos(x/2)=1 - x^2 / 8

//		if (scope.TheObject.position.y<0) scope.TheObject.position.y=0;
		
		if ( zoomChanged ||
			lastPosition.distanceToSquared( scope.TheObject.position ) > EPS ||
			8 * ( 1 - lastQuaternion.dot( scope.TheObject.quaternion ) ) > EPS ) {

			//scope.dispatchEvent( changeEvent );

			lastPosition.copy( scope.TheObject.position );
			lastQuaternion.copy( scope.TheObject.quaternion );
			zoomChanged=false;

			Result=true;
		}

		return(Result);
	}
	
	this.GetPosition=function() 
	{ 
		var Result=new THREE.Vector3(this.TheObject.position.x,this.TheObject.position.y,this.TheObject.position.z);
		return(Result); 
	}
	this.SetPosition=function(x,y,z)
	{
		this.TheObject.position.set(x,y,z); // 
		this.TheObject.updateProjectionMatrix();
		this.Update();
	}
	this.SetRotation=function(x,y,z)
	{
		this.TheObject.rotation.set(x,y,z);
	}
	this.LookAt=function(x,y,z)
	{
		this.target=new THREE.Vector3(x,y,z);
		this.TheObject.lookAt(x,y,z);
		this.Update();
	}

	//*************************************************************************
	// Public CM3Controls Functions
	//*************************************************************************
	this.ZoomIn=function()
	{
		dollyIn(1);
	}
	this.ZoomOut=function()
	{
		dollyOut(1);
	}
	//*************************************************************************
	// Public CM3Controls event handlers
	//*************************************************************************
	//
	// event handlers - FSM: listen for events and reset state
	//

	this.MouseDown=function ( event )  // jjg I had to add this to make mousedown public so I could call it from the view
	{
		var Used=false;
		
		var MouseActive=this.GetSetting("CM3Controls","MouseActive",true);

		if (MouseActive)
		{
			if ( event.button === scope.mouseButtons.ORBIT ) 
			{
				if ( scope.enableRotate)
				{
					rotateStart.set( event.clientX, event.clientY );
		
					state=STATE.ROTATE;
				}
			} 
			else if ( event.button === scope.mouseButtons.ZOOM ) 
			{
				if ( scope.enableZoom)
				{
					dollyStart.set( event.clientX, event.clientY ); //handleMouseDownDolly( event );
	
					state=STATE.DOLLY;
				}
			} 
			else if ( event.button === scope.mouseButtons.PAN ) 
			{
				if ( scope.enablePan)
				{
					panStart.set( event.clientX, event.clientY );
	
					state=STATE.PAN;
				}
			}
			Used=true;
		}

		return(Used);
	}

	this.MouseMove=function ( event ) // jjg I had to add this to make mousedown public so I could call it from the view
	{
		var Used=false;
		
		var MouseActive=this.GetSetting("CM3Controls","MouseActive",true);

		if (MouseActive)
		{
			if ( state === STATE.ROTATE ) 
			{
				this.handleMouseMoveRotate( event );
			} 
			else if ( state === STATE.DOLLY ) 
			{
				handleMouseMoveDolly( event );
			} 
			else if ( state === STATE.PAN ) 
			{
				handleMouseMovePan( event );
			}
			Used=true;
		}

		return(Used);
	}

	this.MouseUp=function ( event ) // jjg I had to add this to make mousedown public so I could call it from the view
	{
		var Used=false;
		
		var MouseActive=this.GetSetting("CM3Controls","MouseActive",true);

		if (MouseActive)
		{
			state=STATE.NONE;
			
			Used=true;
		}

		return(Used);
	}

	this.MouseWheel=function ( event ) // jjg I had to add this to make mousedown public so I could call it from the view
	{
		var Used=false;
		
		var MouseActive=this.GetSetting("CM3Controls","MouseActive",true);

		if (MouseActive)
		{
			if ( scope.enabled === false || scope.enableZoom === false || ( state !== STATE.NONE && state !== STATE.ROTATE ) ) return;
	
			var delta=0;
	
			if ( event.wheelDelta !== undefined )  // WebKit / Opera / Explorer 9
			{
				delta=event.wheelDelta;
			} 
			else if ( event.detail !== undefined )  // Firefox
			{
				delta=- event.detail;
			}
	
			// if there was a change, zoom in or out
			if ( delta > 0 ) 
			{
				dollyOut( getZoomScale() );
			} 
			else if ( delta < 0 ) 
			{
				dollyIn( getZoomScale() );
			}
	
			scope.Update();
			
			if (event.preventDefault)  event.preventDefault();
			Used=true;
		}

		return(Used);
	}

	this.KeyDown=function ( event ) // jjg I had to add this to make mousedown public so I could call it from the view
	{
		var Used=false;
		
		if ( scope.enabled === false || scope.enableKeys === false || scope.enablePan === false ) return;

		switch ( event.keyCode ) {

			case scope.keys.UP:
				pan( 0, scope.keyPanSpeed );
				scope.Update();
				break;

			case scope.keys.BOTTOM:
				pan( 0, - scope.keyPanSpeed );
				scope.Update();
				break;

			case scope.keys.LEFT:
				pan( scope.keyPanSpeed, 0 );
				scope.Update();
				break;

			case scope.keys.RIGHT:
				pan( - scope.keyPanSpeed, 0 );
				scope.Update();
				break;

		}
		//console.log("Keydown");

		return(Used);
	}
	this.KeyUp=function ( event ) 
	{
		return(false);
	}

	this.TouchStart=function ( event )  // jjg I had to add this to make mousedown public so I could call it from the view
	{
		var Used=false;
		

		switch ( event.touches.length ) 
		{

			case 1:	// one-fingered touch: rotate

				if ( scope.enableRotate)
				{
					handleTouchStartRotate( event );
	
					state=STATE.TOUCH_ROTATE;
				}
				break;

			case 2:	// two-fingered touch: dolly

				if ( scope.enableZoom)
				{
					handleTouchStartDolly( event );

					state=STATE.TOUCH_DOLLY;
				}
				break;
			case 3: // three-fingered touch: pan

				if ( scope.enablePan)
				{
					handleTouchStartPan( event );

					state=STATE.TOUCH_PAN;
				}
				break;
			default:
				state=STATE.NONE;
		}

		return(Used);
	}

	this.TouchMove=function ( event ) // jjg I had to add this to make mousedown public so I could call it from the view
	{
		var Used=false;

		switch ( event.touches.length ) 
		{
			case 1: // one-fingered touch: rotate

				if ( scope.enableRotate && state === STATE.TOUCH_ROTATE )
				{
					handleTouchMoveRotate( event );
				}
				break;
			case 2: // two-fingered touch: dolly

				if ( scope.enableZoom && state === STATE.TOUCH_DOLLY )
				{
					handleTouchMoveDolly( event );
				}
				break;

			case 3: // three-fingered touch: pan

				if ( scope.enablePan && state === STATE.TOUCH_PAN )
				{
					handleTouchMovePan( event );
				}
				break;
			default:
				state=STATE.NONE;
		}

		return(Used);
	}

	this.TouchEnd=function ( event )  // jjg I had to add this to make mousedown public so I could call it from the view
	{
		var Used=false;

		handleTouchEnd( event );

		state=STATE.NONE;

		return(Used);
	}


};

CM3ControlsOrbit.prototype=Object.create( CM3Controls.prototype );
CM3ControlsOrbit.prototype.constructor=CM3ControlsOrbit;

//*************************************************************************
// CMBase Functions
//*************************************************************************
CM3ControlsOrbit.prototype.CM3Controls_GetSettingsDefinitions=CM3Controls.prototype.GetSettingsDefinitions; // save the superclass function

CM3ControlsOrbit.prototype.GetSettingsDefinitions=function() 
{
	var Result=this.CM3Controls_GetSettingsDefinitions(); // get the settings definitions from the superclass
	
	for (Key in CM3ControlsOrbit.SettingDefintions)
	{
		Result[Key]=CM3ControlsOrbit.SettingDefintions[Key];
	}

	return(Result); 
}

//******************************************************************
// Definitions
//******************************************************************
/**
* Below are the settings definitions.
* @public, @settings
*/
CM3ControlsOrbit.SettingDefintions=
{
	CM3ControlsOrbit:
	{
		MinDistance: { Name:"Min Distance",Type:CMBase.DATA_TYPE_FLOAT, Default:0 },
		MaxDistance: { Name:"Max Distance",Type:CMBase.DATA_TYPE_FLOAT, Default:Infinity },
		MinLatitude: { Name:"Min Latitude",Type:CMBase.DATA_TYPE_FLOAT, Default:-90 },
		MaxLatitude: { Name:"Max Latitude",Type:CMBase.DATA_TYPE_FLOAT, Default:90 },
	},
};