/**************************************************************************************************
* CM3Item Base Class
*
* Abstract class for 3D items
* @module CM3Item
* @Copyright HSU, Jim Graham, 2019
**************************************************************************************************/
//**************************************************************************************************
// Constructor
//**************************************************************************************************
/*
* Constructor
* @public, @constructs
*/
function CM3Item()
{
	CMItem.call(this);
	
	this.ThePosition=new THREE.Vector3(0,0,0);
	
	this.TheMaterial=null; // can get set by the user
	
	// other properties
	
	this.TheOGLObject=null; // the object the user can interact with
	
	// properties for moving items around
	this.Anchor=new THREE.Vector3();
	this.Dragging=false;
	this.DraggingPlane=new THREE.Plane(); // Plane that objects are dragged within (normal plane to the camera at the item intersected with mouse?
	this.IntersectionPoint=new THREE.Vector3();

	this.Labels=null;
}
CM3Item.prototype=Object.create(CMItem.prototype);

CM3Item.prototype.constructor=CM3Item;

//**************************************************************************************************
// Private functions
//**************************************************************************************************
/**
* Private function to initlaize the item's OpenGL object
* @protected
*/
CM3Item.prototype.Initialize=function()
{
	if (this.TheOGLObject==null)
	{
		this.SetOGLObjectVisible(false); // make the object invisible while we are changing things
		
		var TheGeometry=this.GetGeometry();
		
		// setup the material
		
		this.TheMaterial=this.GetMaterial();
		
		if ((TheGeometry!=null)&&(this.TheMaterial!=null))
		{
			var TheMesh=this.TheOGLObject;
			
			// create the object that will go into the scene
			
			if (TheMesh!=null) // modify the existing object
			{
				TheMesh.material=this.TheMaterial;
				TheMesh.geometry=TheGeometry;
				
				TheMesh.position.copy( this.ThePosition );
			}
			else // create a new object
			{
				TheMesh = new THREE.Mesh(TheGeometry, this.TheMaterial);
				TheMesh.TheCM3Item=this;
				
				var Rotation=this.GetSetting("Item","Rotation");
				
				if (Rotation!=undefined)
				{
					TheMesh.rotateX(Rotation[0]);
					TheMesh.rotateY(Rotation[1]);
					TheMesh.rotateZ(Rotation[2]);
				}
				var TheGeo=this.GetParent(CM3Geo);
				var TheProjector=TheGeo.GetProjector();
				
				if ((typeof(CM3ProjectorSpherical)!="undefined")&&(TheProjector instanceof CM3ProjectorSpherical))
				{
					TheMesh=TheProjector.GetPositionedMesh(TheMesh,this.ThePosition.x,this.ThePosition.y,this.ThePosition.z);
				}
				else
				{
					TheMesh.position.copy( this.ThePosition );
				}
			}
			//
			var Scale=this.GetSetting("Item","Scale");
			
			if (Scale!=undefined)
			{
				TheMesh.scale.set(Scale[0],Scale[1],Scale[2]);
			}
			this.TheOGLObject=TheMesh;
		}
		this.SetOGLObjectVisible(true); // make the object visible
	}
};
/**
* This mouse down is different from 2D as we know this item had the mouse clicked down in it.
* @private
*/
CM3Item.prototype.MouseDownOld=function(TheView,TheEvent,TheRayTracer)
{
	var Used=false;
	
//if (this.GetSetting("Item","Editable"))
	{
		if (this.TheOGLObject!=null)
		{
			var SelectedObjects=TheRayTracer.intersectObject(this.TheOGLObject,false); // find the objects the ray intersects with
			
			if (SelectedObjects.length>0)
			{
				if (TheRayTracer.ray.intersectPlane(this.DraggingPlane, this.IntersectionPoint)) // undocumented!
				{
					TheView.SetCursor('move'); // change the cursor to move the object
					
					// Find the difference between the intersection with the object and the object's origin and save as the anchor point
					this.Anchor.copy( this.IntersectionPoint ).sub( this.TheOGLObject.position ); 
					
					this.Dragging=true;
					
					Used=true;
					TheEvent.stopPropagation();
				}
			}
		}
	}
	
	return(Used); // used=true
}
/*
*/
CM3Item.prototype.MouseMoveOld=function(TheView,TheEvent,TheRayTracer)
{
	var Used=false;
	
	if (this.TheOGLObject!=null)
	{
		if (this.Dragging)
		{
			if (TheRayTracer.ray.intersectPlane(this.DraggingPlane, this.IntersectionPoint)) // undocumented!
			{
				var NewPosition=new THREE.Vector3();
				NewPosition.copy(this.IntersectionPoint);
				NewPosition.sub(this.Anchor);
				
				// have to use set position so subclasses (e.g. CM3Light) can move their other objects
				
				this.SetSetting("Item","PositionOffset",[NewPosition.x,NewPosition.y,NewPosition.z]); 
				//this.SetPosition(NewPosition.x,NewPosition.y,NewPosition.z); // set the selected objects position
				
				TheEvent.stopPropagation();
				Used=true;
			}
		}
		else
		{
			var SelectedObjects=TheRayTracer.intersectObject(this.TheOGLObject,false); // find the objects the ray intersects with
			
			if (SelectedObjects.length>0)
			{
				if (TheRayTracer.ray.intersectPlane(this.DraggingPlane, this.IntersectionPoint))  // jjg - this appears to be undocumented!
				{
					var TheCamera=TheView.GetCamera();
					
					this.DraggingPlane.setFromNormalAndCoplanarPoint(
						TheCamera.getWorldDirection(this.DraggingPlane.normal), // normal to the this.DraggingPlane
						this.TheOGLObject.position); // point on the this.DraggingPlane (coplanar point)
					
//					TheEvent.stopPropagation();
//					Used=true;
				}
			}
		}
	}
	return(Used);
}
/*
*/
CM3Item.prototype.MouseUpOld=function(TheView,TheEvent,TheRayTracer)
{
	var Used=false;
	
	if (this.Dragging)
	{
		this.Dragging=false;
		TheEvent.stopPropagation();
		Used=true;
	}
	
	return(Used); 
}
//******************************************************************
// CMBase Settings Functions
//******************************************************************
	
CMItem.prototype.CMItem_SetSettings=CMItem.prototype.SetSettings;

CM3Item.prototype.SetSettings=function(NewSettings,TimeSlice) 
{
	this.CMItem_SetSettings(NewSettings,TimeSlice);
	
	this.Repaint();
}
	
CMItem.prototype.CMItem_SetSetting=CMItem.prototype.SetSetting;

CM3Item.prototype.SetSetting=function(Group,Key,Value) 
{
	this.CMItem_SetSetting(Group,Key,Value);
			
	if ((Group=="Item")&&(Key=="Status"))
	{
		if (this.GetSetting("Item","Status")!=Value)
		{
			var Flag=false;
			if (Value!=CM3Item.STATUS_HIDDEN) Flag=true;
			
			// also have to make the OGLObject(s) visible or invisible
			this.SetOGLObjectVisible(Flag);
		}
	}
	else if ((Group=="Item")&&(Key=="Rotation"))
	{
		if (this.TheOGLObject!=null)
		{
			this.TheOGLObject.rotateX(Value[0]);
			this.TheOGLObject.rotateY(Value[1]);
			this.TheOGLObject.rotateZ(Value[2]);
		}
	}
	else if ((Group=="Item")&&(Key=="PositionOffset"))
	{
		this.ThePosition=new THREE.Vector3(Value[0],Value[1],Value[2]);
		
		if (this.TheOGLObject!=null)
		{
			this.TheOGLObject.position.copy( this.ThePosition );
		}
	}
	else if ((Group=="Item")&&(Key=="Scale"))
	{
		if (this.TheOGLObject!=null)
		{
			var Scale=new THREE.Vector3(Value[0],Value[1],Value[2]);
			
			this.TheOGLObject.scale.set(this.ScaleX,this.ScaleY,this.ScaleZ);
		}
	}
	else 
	{
		this.Repaint();
	}
}
/**
* Utility function to call old event handlers
* @public
*/
CM3Item.prototype.SendMessageToDescendants=function(Message,AdditionalInfo)
{
	Used=false;
	
	switch (Message)
	{
	case CMView.MESSAGE_MOUSE_DOWN:
		Used=this.MouseDownOld(AdditionalInfo.TheView,AdditionalInfo.TheEvent,AdditionalInfo.TheRayTracer);
		break;
	case CMView.MESSAGE_MOUSE_MOVE:
		Used=this.MouseMoveOld(AdditionalInfo.TheView,AdditionalInfo.TheEvent,AdditionalInfo.TheRayTracer);
		break;
	case CMView.MESSAGE_MOUSE_UP:
		Used=this.MouseUpOld(AdditionalInfo.TheView,AdditionalInfo.TheEvent,AdditionalInfo.TheRayTracer);
		break;
	}
	return(Used);
}

//**************************************************************************************************
// Public STItem Functions
//**************************************************************************************************
//CM3Item.prototype.CMItem_SetVisible=CMItem.prototype.SetVisible;

CM3Item.prototype.SetOGLObjectVisible=function(Flag)
{
	//if (this.GetVisible()!=Flag)
	{
		//this.CMItem_SetVisible(Flag);
		
		if (this.TheOGLObject!=null)
		{
			var TheScene=this.GetParent(CM3Scene);
			var TheGeo=this.GetParent(CM3Geo);
			
			if (Flag) // becoming visible
			{
				if (this.TheOGLObject!=null)
				{
					TheGeo.AddOGLObjectToGeo(this.TheOGLObject); // add the control object 
				}
			}
			else  
			{
				if (this.TheOGLObject!=null)
				{
					TheGeo.RemoveOGLObjectFromGeo(this.TheOGLObject); // remove the control object  
				}
			}
		}
	}
	
};

CM3Item.prototype.Paint=function(TheView) 
{
	//if (this.TheTextureDataset!=null)
	{
		var TheStyle=this.GetStyle(TheView);
		
		if (TheStyle!==null) { TheView.SetStyle(TheStyle); }
	
		this.Initialize();
		
		if (TheStyle!==null) { TheView.RestoreStyle(); }
	}
}


//**************************************************************************************************
// CM3Item functions fur subclasses to override
//**************************************************************************************************
CM3Item.prototype.GetGeometry=function()
{
	return(null);
}
CM3Item.prototype.GetMaterial=function()
{
	if (this.TheMaterial==null)
	{
		var TheColor=this.GetSetting("Style","fillStyle","rgb(255,255,255)");
		
		var opacity=1;
		var transparent=false;
		
		if (typeof(TheColor) === 'string')
		{
			var TheValues=CMUtilities.GetRGBAValuesFromRGBA(TheColor);
	
			opacity=TheValues.Transparency;
			
			if (opacity!=1) transparent=true;
			
			TheColor="rgb("+TheValues.Red+","+TheValues.Green+","+TheValues.Blue+")";
		}
		
		this.TheMaterial= new THREE.MeshPhongMaterial( { 
			color: TheColor ,
			opacity:opacity,
			transparent :transparent
		} ) ;
	}
	return(this.TheMaterial);
}

//**************************************************************************************************
// Public CM3Item functions
//**************************************************************************************************
/**
* Provide a custom material for the item
*/
CM3Item.prototype.SetMaterial=function(TheMaterial)
{
	this.TheMaterial=TheMaterial;
}

CM3Item.prototype.SetPosition=function(X,Y,Z)
{
	this.SetSetting("Item","PositionOffset",[X,Y,Z]);
}
CM3Item.prototype.SetRotation=function(X,Y,Z)
{
	this.SetSetting("Item","Rotation",[X,Y,Z]);
}
CM3Item.prototype.SetScale=function(X,Y,Z)
{
	this.SetSetting("Item","Scale",[X,Y,Z]);
}
CM3Item.prototype.SetOGLObject=function(NewOGLObject)
{
	this.TheOGLObject=NewOGLObject;
}
CM3Item.prototype.GetOGLObject=function()
{
	return(this.TheOGLObject);
}

CM3Item.prototype.AddLabel=function(TheLabel)
{
	TheLabel.SetParent(this);
	
	if (this.TheLabels==null) this.TheLabels=[];

	var Index=this.TheLabels.length;
	
	this.TheLabels.push(TheLabel);
	TheLabel.Initialize();
	
	return(Index);
}

