/**************************************************************************************************
* CM3ItemBox Class
*
* @module CM3ItemBox
* @Copyright HSU, Jim Graham, 2019
**************************************************************************************************/
//**************************************************************************************************
// Constructor
//**************************************************************************************************

/**
 * Container class for a 3D item that fits in a box (e.g. boxes, cones, cylinders, pyramids)
 */
function CM3ItemBox(Type) 
{
	CM3Item.call(this);
	
	if (Type==undefined) Type=CM3ItemBox.BOX;
	
	//
	
	this.Type=Type;
	
	switch (this.Type)
	{
	case CM3ItemBox.PYRAMID:
		this.SetName("Pyramid");
		this.TimeSlices[0].Settings.Dimensions=	
		{
			Width:10,
			Height:20,
		}
		break;
	case CM3ItemBox.BOX:
		this.SetName("Box");
		this.TimeSlices[0].Settings.Dimensions=	
		{
			Width:10,
			Height:10,
			Depth:10
		}
		break;
	case CM3ItemBox.CONE:
		this.SetName("Cone");
		this.TimeSlices[0].Settings.Dimensions=	
		{
			TopRadius:2,
			BottomRadius:15,
			Height:20,
		}
		break;
	case CM3ItemBox.CYLINDER:
		this.SetName("Cylinder");
		this.TimeSlices[0].Settings.Dimensions=	
		{
			Radius:10,
			Height:20,
		}
		break;
	case CM3ItemBox.SPHERE:
		this.SetName("Sphere");
		this.TimeSlices[0].Settings.Dimensions=	
		{
			Radius:20,
		}
		break;
	}
}
CM3ItemBox.prototype=Object.create(CM3Item.prototype);

CM3ItemBox.prototype.constructor=CM3ItemBox;
//**************************************************************************************************
// Public Defniitions
//**************************************************************************************************

CM3ItemBox.PYRAMID=1;
CM3ItemBox.BOX=2;
CM3ItemBox.CONE=3;
CM3ItemBox.CYLINDER=4;
CM3ItemBox.SPHERE=5;

//******************************************************************
// CMBase Settings Functions
//******************************************************************
CM3ItemBox.prototype.CM3Item_GetSettingsDefinitions=CM3Item.prototype.GetSettingsDefinitions;

CM3ItemBox.prototype.GetSettingsDefinitions=function() 
{
	var Result=this.CM3Item_GetSettingsDefinitions();
	
	// remove the text and shadow settings
	delete(Result.Text);
	delete(Result.Shadow);
	
	switch (this.Type)
	{
	case CM3ItemBox.PYRAMID:
		Result["Dimensions"]=
		{
			Width: { Name:"Width",Type:CMBase.DATA_TYPE_FLOAT, Default:10 },
			Height: { Name:"Height",Type:CMBase.DATA_TYPE_FLOAT, Default:20 },
			NumSides: { Name:"Num Sides",Type:CMBase.DATA_TYPE_FLOAT, Default:4 },
		}
		break;
	case CM3ItemBox.BOX:
		Result["Dimensions"]=
		{
			Width: { Name:"Width",Type:CMBase.DATA_TYPE_FLOAT, Default:20 },
			Height: { Name:"Height",Type:CMBase.DATA_TYPE_FLOAT, Default:20 },
			Depth: { Name:"Depth",Type:CMBase.DATA_TYPE_FLOAT, Default:20 },
		}
		break;
	case CM3ItemBox.CONE:
		Result["Dimensions"]=
		{
			TopRadius: { Name:"Top Radius",Type:CMBase.DATA_TYPE_FLOAT, Default:2 },
			BottomRadius: { Name:"Bottom Radius",Type:CMBase.DATA_TYPE_FLOAT, Default:20 },
			Height: { Name:"Height",Type:CMBase.DATA_TYPE_FLOAT, Default:20 },
			NumSides: { Name:"Num Sides",Type:CMBase.DATA_TYPE_FLOAT, Default:4 },
		}
		break;
	case CM3ItemBox.CYLINDER:
		Result["Dimensions"]=
		{
			Radius: { Name:"Radius",Type:CMBase.DATA_TYPE_FLOAT, Default:10 },
			Height: { Name:"Height",Type:CMBase.DATA_TYPE_FLOAT, Default:20 },
			NumSides: { Name:"Num Sides",Type:CMBase.DATA_TYPE_FLOAT, Default:4 },
		}
		break;
	case CM3ItemBox.SPHERE:
		Result["Dimensions"]=
		{
			Radius: { Name:"Radius",Type:CMBase.DATA_TYPE_FLOAT, Default:20 },
			NumSides: { Name:"Num Sides",Type:CMBase.DATA_TYPE_FLOAT, Default:4 },
		}
		break;
	}

	return(Result); 
}
//**************************************************************************************************
// CM3Item functions
//**************************************************************************************************
CM3ItemBox.prototype.GetGeometry=function()
{
	var TheGeometry=null;
	
	var Dimensions=this.TimeSlices[0].Settings.Dimensions;
		
	switch (this.Type)
	{
	case CM3ItemBox.PYRAMID:
		var NumSides=this.GetSetting("Dimensions","NumSides",4);
		TheGeometry = new THREE.CylinderBufferGeometry( 0, Dimensions.Width/2, Dimensions.Height, NumSides ); // cylinder with closed top and 4 sides = pyramid
		break;
	case CM3ItemBox.BOX:
		TheGeometry = new THREE.BoxBufferGeometry( Dimensions.Width, Dimensions.Height, Dimensions.Depth );
		break;
	case CM3ItemBox.CONE:
		var NumSides=this.GetSetting("Dimensions","NumSides",30);
		TheGeometry = new THREE.CylinderBufferGeometry( Dimensions.TopRadius, Dimensions.BottomRadius,Dimensions.Height, NumSides ); // cylinder with a closed top
		break;
	case CM3ItemBox.CYLINDER:
		var NumSides=this.GetSetting("Dimensions","NumSides",30);
		TheGeometry = new THREE.CylinderBufferGeometry( Dimensions.Radius, Dimensions.Radius, Dimensions.Height, NumSides );
		break;
	case CM3ItemBox.SPHERE:
		var NumSides=this.GetSetting("Dimensions","NumSides",32);
		TheGeometry = new THREE.SphereBufferGeometry( Dimensions.Radius, NumSides, NumSides );
		break;
	}
	return(TheGeometry);
}

//******************************************************************
// CM3ItemBox Functions
//******************************************************************
CM3ItemBox.prototype.GetName=function()  
{ 
	var Result="Untitled";
	
	switch (this.Type)
	{
	case CM3ItemBox.PYRAMID:
		Result="Pyramid";
		break;
	case CM3ItemBox.BOX:
		Result="Box";
		break;
	case CM3ItemBox.CONE:
		Result="Cone";
		break;
	case CM3ItemBox.CYLINDER:
		Result="Cylinder";
		break;
	case CM3ItemBox.SPHERE:
		Result="Sphere";
		break;
	}
	return(Result); 
}



//**************************************************************************************************
// Public functions
//**************************************************************************************************
