/******************************************************************************************************************
* CM3LayerGraticule
*
* by: Jim Graham
* 
* This class required a level of complexity beyond what I initially expected.
* The problem is that we have to:
*	- Draw graticule lines based on the zoom level, projection, and area viewed
*  - Paint coordinates along the border based on where the graticules intersect with the border
*  - Paint coordinates along the outside of the graticule grid if it is visible (i.e. does not 
*    intersect with the border.
*
* graticles are made visible by setting the overall style to transparent.  The graticules must be drawn to get the intersections for the exterior and border
* @module CM3LayerGraticule 
* @Copyright HSU, Jim Graham, 2019
******************************************************************************************************************/

/**
* Below are the additional settings definitions for STLayers.  Setings for the basic drawing style and text
* style are inherited from STItem
* @public, @settings
*/
CM3LayerGraticule.SettingDefintions=
{
	Graticule3D: // 3d settings 
	{
		LineWidth: { Name:"Line Width",Type:CMBase.DATA_TYPE_FLOAT, Default:20 },
		HeightAboveSeaLevel: { Name:"Height Above Sea Level",Type:CMBase.DATA_TYPE_FLOAT, Default:20 },
		LabelHeightAboveSeaLevel: { Name:"Height Above Sea Level",Type:CMBase.DATA_TYPE_FLOAT, Default:40 },
		
		LongitudePosition: { Name:"Longitude Position",Type:CMBase.DATA_TYPE_ENUMERATED, Options:CMLayer.LABEL_DIRECTIONS,Default:null }, // Position of the longitude text, null for centered
		LongitudeXOffset: { Name:"Longitude X Offset",Type:CMBase.DATA_TYPE_FLOAT, Default:50 },
		LongitudeYOffset: { Name:"Longitude Y Offset",Type:CMBase.DATA_TYPE_FLOAT, Default:0 },
		
		LatitudePosition: { Name:"Latitude Position",Type:CMBase.DATA_TYPE_ENUMERATED, Options:CMLayer.LABEL_DIRECTIONS,Default:"T" },
		LatitudeXOffset: { Name:"Latitude X Offset",Type:CMBase.DATA_TYPE_FLOAT, Default:0 },
		LatitudeYOffset: { Name:"Latitude Y Offset",Type:CMBase.DATA_TYPE_FLOAT, Default:50 },
	},
	Text3D:
	{
		color: { Name:"Color",Type:CMBase.DATA_TYPE_COLOR, Default:0xffffff }, // size of the text
		opacity: { Name:"Opacity",Type:CMBase.DATA_TYPE_FLOAT, Default:1 }, // 1 for opaque, 0 for tranparent
		size: { Name:"Size",Type:CMBase.DATA_TYPE_FLOAT, Default:1 }, // size of the text
		HaloSize: { Name:"Halo Size",Type:CMBase.DATA_TYPE_FLOAT, Default:10 }, // size of the text
		GlobalFontIndex: { Name:"Global Font Index",Type:CMBase.DATA_TYPE_INTEGER, Default:2 }, // global font to use
		
		height: { Name:"height",Type:CMBase.DATA_TYPE_FLOAT, Default:2 }, // amount to extrude the object
		curveSegments: { Name:"curveSegments",Type:CMBase.DATA_TYPE_FLOAT, Default:4 }, // number of points on the curve
		bevelEnabled: { Name:"bevelEnabled",Type:CMBase.DATA_TYPE_BOOLEAN, Default:true }, // beveled instead of 90 degree change, keep true for good looking text
		bevelSegments: { Name:"bevelSegments",Type:CMBase.DATA_TYPE_FLOAT, Default:3 }, // number of bevel layers (more makes a smoother bevel but draws slower
		bevelSize: { Name:"bevelSize",Type:CMBase.DATA_TYPE_FLOAT, Default:0.1 },// width of bevel (distance from shape outline), keep > 0
		bevelThickness: { Name:"bevelThickness",Type:CMBase.DATA_TYPE_FLOAT, Default:0.1 }, // how deep the bevel goes into the extrusion ,keep > 0
	}
};


//**************************************************************************************************
// Constructor
//**************************************************************************************************
/*
* Constructor
* @public, @constructs
*/
function CM3LayerGraticule() 
{
	CMLayerGraticule.call(this);
	
	// settings
	this.TimeSlices[0].Settings.Graticule3D=	
	{
	};
	this.TimeSlices[0].Settings.Text3D=	
	{
	};
	
	// default spacing is 15 degrees
	this.LatitudesWithLatitudeLabels=[-75-60,-45,-30,-15,0,15,30,45,60,75]; // which longitudes have latitude markins
	this.LatitudeSpacings=[45,45,45,45,45,45,45,45,45,45,45]; // spacing for each entry in LatitudesWithLabels
	this.LatitudeLongitudinalOffset=7.5;
	
	this.LatitudesWithLongitudeLabels=[-22.5,22.5]; // which longitudes have latitude markings
	this.LongitudeSpacings=[15,15]; // spacing for each entry in LongitudesWithLabels
	
	//***********************************************
	// temp properties for the currnet clipping box setup in SetupBoundsPoly()
	
	//********************************************************
	this.LineMesh=null;
	this.LineMeshInGeo=false;
	
	this.TextMeshes=null;
	this.TextMeshesInGeo=false;
	this.TextBufferGeometrySettings=null;
	
	this.TextHaloSettings=null;
	
}
CM3LayerGraticule.prototype=Object.create(CMLayerGraticule.prototype); // inherit prototype functions from PanelBase()

CM3LayerGraticule.prototype.contructor=CM3LayerGraticule; // override the constructor to go to ours

//**************************************************************************
// Private Functions for managing 3d data
//**************************************************************************
/**
* Get the cross-section for the line
* @private
*/
CM3LayerGraticule.prototype.GetLineMaterial=function()
{
	// if the material has not yet been created or set, create one from the settings
	if (this.TheMaterial==undefined)
	{
		var Color=this.GetSetting("Style","lineStyle",0xcccccc);
//		var LineWidth=this.GetSetting("Graticule3D","LineWidth",12);
//		var LineWidth=1; // 
		
		var Result=CM3Utilities.Get3DColorFromCanvasColor(Color);
//		Color=CMUtilities.GetHexColorFromColor(Color);
		
		this.TheMaterial= new THREE.MeshPhongMaterial( { 
			color: Result.RGB ,
			 side:THREE.FrontSide, 
			opacity:Result.Opacity,
			transparent :Result.Transparent
		} ) ;
/*		this.TheMaterial= new THREE.LineBasicMaterial( // makes a "flat" surface
		{ 
			color: Result.RGB ,
			opacity:Result.Opacity,
			transparent:Result.Transparent,
			linewidth:LineWidth
		} );
*/	}
	return(this.TheMaterial);
}
/**
* Get the cross-section for the line
* @private
*/
CM3LayerGraticule.prototype.GetLineCrossSection=function(Factor)
{
//	if (this.TheCrossSection==undefined)
	{
		var TheShape = new THREE.Shape();
		
		var Size=this.GetSetting("Graticule3D","LineWidth",10);
		
		Size=Size*Factor;
		
		TheShape.moveTo( -Size/2, Size/2 );  // top left
		TheShape.lineTo( Size/2, Size/2 ); // top right
		TheShape.lineTo( Size/2, -Size/2 ); // bottom right
		TheShape.lineTo( -Size/2, -Size/2 ); // bottom left
		TheShape.lineTo( -Size/2, Size/2 ); // top left
/*		for (var i=Math.PI/8;i<Math.PI*2;i+=Math.PI/8)
		{
			var x=Math.sin(i)*Size/2;
			var y=Math.cos(i)*Size/2;
			
			TheShape.lineTo( x,y);
		}
		TheShape.lineTo( 0, Size/2 );
*/		
		this.TheCrossSection=TheShape;
	}
	return(this.TheCrossSection);
}
/**
* Setup the mesh to render the 3D graticules
* @private
*/
CM3LayerGraticule.prototype.GetPointsForParallels=function(Latitude,West,East,DegreeSpacing,TheGeo,HeightAboveSeaLevel)
{
	var ThePoints=[];
	for (var Column=West;Column<=East;Column+=DegreeSpacing) // draw each column from west to east
	{
		var Longitude=Column;
		
		// draw each of the segments
		var Spacing=DegreeSpacing/10;
		
		for (var i=0;i<10;i++)
		{
			var Result=TheGeo.ProjectFromGeographic(Longitude, Latitude, HeightAboveSeaLevel);
			
			var Vector=new THREE.Vector3(Result[0],Result[1],Result[2]);
			
			ThePoints.push(Vector);
			
			Longitude+=Spacing;
		}
	}
	// add one more point to bring us back to the begining
	var Result=TheGeo.ProjectFromGeographic(Longitude, Latitude, HeightAboveSeaLevel);
			
	var Vector=new THREE.Vector3(Result[0],Result[1],Result[2]);
	
	ThePoints.push(Vector);

	return(ThePoints);
}
/**
* Setup the mesh to render the 3D graticules
* @private
*/
CM3LayerGraticule.prototype.GetMesh=function()
{
	if (this.LineMesh==null)
	{
		var TheGeo=this.GetParent(CMGeo);
		
		// make sure we have a projector
		var TheProjector=TheGeo.GetProjector();
		if (TheProjector==null) TheProjector=new CMProjector(); // geographic
		
		// get the desired max bounds
		// this needs to be changed to use just the clipping boxes
		var ClippingBoxes=TheProjector.GetClippingPolys();
		var ClippingBox=ClippingBoxes[0]; // Xmin,Ymin,XMax,YMax
		
		// create the group for all the lines
		this.LineMesh = new THREE.Object3D();
		
		var HeightAboveSeaLevel=this.GetSetting("Graticule3D","HeightAboveSeaLevel",20);

		//  get the cross sectional shape
		//var shape = CM3Utilities.GetRegularPolygonShape(4,Size,0,0,0);

		var shape=this.GetLineCrossSection(10);
		
		// for halo
		var shape2=this.GetLineCrossSection(2);
		var HaloMaterial= new THREE.MeshPhongMaterial( { 
			color: 0x000000 ,
			opacity:1,
			transparent :false
		} ) ;
		
		// get the material
		var TheMaterial=this.GetLineMaterial();
		
		var East=this.GetSetting("Graticule","East",180);
		var West=this.GetSetting("Graticule","West",-180);
		var North=this.GetSetting("Graticule","North",90);
		var South=this.GetSetting("Graticule","South",-90);
		var DegreeSpacing=this.GetSetting("Graticule","DegreeSpacing",CMLayerGraticule.MAX_SPACING_DEGREES);
		
		//**************************************************************
		// draw lines of latitude (parallels)
		for (var Row=South;Row<=North;Row+=DegreeSpacing) // draw each row from south to north
		{
			var Latitude=Row;
			
			// draw the parallel
			var ThePoints=this.GetPointsForParallels(Latitude,West,East,DegreeSpacing,TheGeo,HeightAboveSeaLevel);
			
			// create the path with the extrusion
			var ExtrusionPath =  new THREE.CatmullRomCurve3( ThePoints );
			ExtrusionPath.curveType = 'catmullrom';
			ExtrusionPath.closed = false;
				
			var extrudeSettings = {
				steps: 100,
				bevelEnabled: false,
				extrudePath: ExtrusionPath
			};
			
			var geometry = new THREE.ExtrudeBufferGeometry( shape, extrudeSettings );
			
			var TheLine = new THREE.Mesh( geometry, TheMaterial );
			
			this.LineMesh.add(TheLine);
			
			// add the halo
			var ThePoints=this.GetPointsForParallels(Latitude,West,East,DegreeSpacing,TheGeo,HeightAboveSeaLevel-40);
			
			// create the path with the extrusion
			var ExtrusionPath =  new THREE.CatmullRomCurve3( ThePoints );
			ExtrusionPath.curveType = 'catmullrom';
			ExtrusionPath.closed = false;
				
			var extrudeSettings = {
				steps: 100,
				bevelEnabled: false,
				extrudePath: ExtrusionPath
			};
			
			var geometry = new THREE.ExtrudeBufferGeometry( shape2, extrudeSettings );
			
			var TheLine = new THREE.Mesh( geometry, HaloMaterial );
			
			this.LineMesh.add(TheLine);
		}
		//*********************************************************************
		// draw lines of longitude (merdians)
		for (var Column=West;Column<East;Column+=DegreeSpacing) // draw each column from west to east
		{
			var ThePoints=[];
			
			for (var Row=South;Row<North;Row+=DegreeSpacing) // draw each row from south to north
			{
				var Latitude=Row;
				var Longitude=Column;
				
				var Spacing=DegreeSpacing/10;
				
				for (var i=0;i<10;i++)
				{
					var Result=TheGeo.ProjectFromGeographic(Column, Latitude, HeightAboveSeaLevel);
					
					var Vector=new THREE.Vector3(Result[0],Result[1],Result[2]);
					
					ThePoints.push(Vector);
				
					Latitude+=Spacing;
				}
			}
			// create the path for the lines
			var ExtrusionPath =  new THREE.CatmullRomCurve3( ThePoints );
			ExtrusionPath.curveType = 'catmullrom';
			ExtrusionPath.closed = false;
				
			var extrudeSettings = {
				steps: 36, // number of line segments
				bevelEnabled: false,
				extrudePath: ExtrusionPath
			};
			// create the geometry
			var geometry = new THREE.ExtrudeBufferGeometry( shape, extrudeSettings );
			
			// create the mesh with the geometry and TheMaterial
			var TheLine = new THREE.Mesh( geometry, TheMaterial );
			
//			edges = new THREE.VertexNormalsHelper( TheLine, 200, 0x00ff00, 1 );
//			edges = new THREE.FaceNormalsHelper( mesh, 2, 0x00ff00, 1 );
//			this.LineMesh.add(edges);
			
			this.LineMesh.add(TheLine);
		}
	
	}
	
	
	return(this.LineMesh);
}
//**************************************************************************************
// Private functions for managing the text
//**************************************************************************************

/**
* Get the material for the lat/lon text
* @private
*/
CM3LayerGraticule.prototype.GetTextMaterial=function(color)
{
	// if the material has not yet been created or set, create one from the settings
	if (this.TextMaterial==undefined)
	{
		if (color==undefined) color=this.GetSetting("Text3D","color",0xff0000);
		var opacity=this.GetSetting("Text3D","opacity",1);
		
		var Transparent=false;
		if (opacity!=1) Transparent=true;
		
		this.TextMaterial= new THREE.MeshPhongMaterial( { 
			color: color ,
			opacity:opacity,
			transparent :Transparent
		} ) ;
	}
	return(this.TextMaterial);
}
/**
* Get the material for the halo behind the lat/lon text
* @private
*/
CM3LayerGraticule.prototype.GetHaloMaterial=function(color)
{
	// if the material has not yet been created or set, create one from the settings
	if (this.HaloMaterial==undefined)
	{
		if (color==undefined) color=this.GetSetting("Text3D","color",0xff0000);
		var opacity=this.GetSetting("Text3D","opacity",1);
		
		var Transparent=false;
		if (opacity!=1) Transparent=true;
	
		this.HaloMaterial = new THREE.MeshPhongMaterial( { 
			color: color ,
			opacity:opacity,
			transparent :true
		} ) ;
	}
	return(this.HaloMaterial);
}

/**
* Get the settings for the text buffer geometry
* @private
*/
CM3LayerGraticule.prototype.GetTextBufferGeometrySettings=function()
{
	if (this.TextBufferGeometrySettings==null)
	{
		if ((CM3Text.GlobalFonts!=undefined)&&(CM3Text.GlobalFonts.length>0))
		{
			var GlobalFontIndex=this.GetSetting("Text3D","GlobalFontIndex",0);
			var size=this.GetSetting("Text3D","size",12);
			
			var height=this.GetSetting("Text3D","height",1);
			var curveSegments=this.GetSetting("Text3D","curveSegments",10);
			var bevelSegments=this.GetSetting("Text3D","bevelSegments",5);
			var bevelSize=this.GetSetting("Text3D","bevelSize",0.1);
			var bevelThickness=this.GetSetting("Text3D","bevelThickness",0.1);
		
			this.TextBufferGeometrySettings= { 
				font: CM3Text.GlobalFonts[GlobalFontIndex],
				size: size,
				height: height,
				curveSegments: curveSegments,
				bevelEnabled: true,  // beveled instead of 90 degree change
				bevelSegments: bevelSegments,  // number of bevel layers (more makes a smoother bevel but draws slower
				bevelSize: bevelSize,  // width of bevel (distance from shape outline
				bevelThickness: bevelThickness // how deep the bevel goes into the extrusion 
			}
			
		}
	}
	return(this.TextBufferGeometrySettings);
}
/**
* Get the settings for the halo around the text
* @private
*/
CM3LayerGraticule.prototype.GetTextHaloSettings=function()
{
	if (this.TextHaloSettings==null)
	{
		if ((CM3Text.GlobalFonts!=undefined)&&(CM3Text.GlobalFonts.length>0))
		{
			var GlobalFontIndex=this.GetSetting("Text3D","GlobalFontIndex",0);
			var size=this.GetSetting("Text3D","size",12);
		
			var HaloSize=this.GetSetting("Text3D","HaloSize",10);
			
			var height=this.GetSetting("Text3D","height",1);
			var curveSegments=this.GetSetting("Text3D","curveSegments",10);
			var bevelSegments=this.GetSetting("Text3D","bevelSegments",5);
			var bevelThickness=this.GetSetting("Text3D","bevelThickness",0.1);
		
			this.TextHaloSettings= { 
				font: CM3Text.GlobalFonts[GlobalFontIndex],
				size: size,
				height: height,
				curveSegments: curveSegments,
				bevelEnabled: true,  // beveled instead of 90 degree change
				bevelSegments: bevelSegments,  // number of bevel layers (more makes a smoother bevel but draws slower
				bevelSize: HaloSize,  // width of bevel (distance from shape outline
				bevelThickness: bevelThickness // how deep the bevel goes into the extrusion 
			}
			
		}
	}
	return(this.TextHaloSettings);
}
/**
* Sets up the text geometry
* @private
*/
CM3LayerGraticule.prototype.GetTextGeometry=function(TheText,Settings)
{
	var TheGeometry=new THREE.TextBufferGeometry( TheText, Settings);

	TheGeometry.computeBoundingBox();
	TheGeometry.computeVertexNormals();

	return(TheGeometry);
}
/**
* Gets the mesh for one text entry.
* @private
*/
CM3LayerGraticule.prototype.GetTextMesh=function(TheGeo,TheText,Settings,textMaterial,Longitude,Latitude,Height,Position,XOffset,YOffset)
{
	var LongitudeInRadians=Longitude/180*Math.PI;
	var LatitudeInRadians=Latitude/180*Math.PI;
	
	// setup the geometry
	var TheGeometry=this.GetTextGeometry(TheText,Settings);
	
	// get the mesh
	var TheLabelMesh = new THREE.Mesh( TheGeometry, textMaterial );
	
	TheLabelMesh.rotateX(Math.PI/2); // rotate the text upright
	TheLabelMesh.rotateY(LongitudeInRadians+Math.PI/2); // rotate the text to lay flat on the earth along the lines of latitude
	TheLabelMesh.rotateX(-LatitudeInRadians); // rotate the text to ly along the lines of longitude
	
	//***********************************
	// setup the position
	
	// get the width and height of the label box
	var TheBoundingBox=TheGeometry.boundingBox;
	var Min=TheBoundingBox.min;
	var Max=TheBoundingBox.max;
	
	var LabelWidth=Max.x-Min.x;
	var LabelHeight=Max.y-Min.y;
	
	// setup the default x and y locations of the label (centered)
	var X=LabelWidth/2;
	var Y=LabelHeight/2;
	
	// find the x and y locations based on the desired position of the label
	switch (Position)
	{
	case "TL":
	case "L":
	case "BL":
		X=-LabelWidth;
		break;
	case "BR":
	case "TR":
	case "R":
		X=0;
		break;
	}
	switch (Position)
	{
	case "TL":
	case "T":
	case "TR":
		Y=0;
		break;
	case "BR":
	case "B":
	case "BL":
		Y=-LabelHeight;
		break;
	}
	// add the desired offsets
	X-=XOffset;
	Y-=YOffset;
	
	// convert the locations of the labels to degrees
	var LatOffsetInRadians=Y/CM3ProjectorSpherical.Radius; // get the fraction of the circumference
	var LatOffsetInDegrees=LatOffsetInRadians*180/Math.PI;
	
	var LonOffsetInRadians=X/CM3ProjectorSpherical.Radius; // get the fraction of the circumference
	var LonOffsetInDegrees=LonOffsetInRadians*180/Math.PI;
	
	// find the new position
	var Result=TheGeo.ProjectFromGeographic(Longitude-LonOffsetInDegrees, Latitude-LatOffsetInDegrees, Height);
	
	var ThePosition=new THREE.Vector3(Result[0],Result[1],Result[2]);
																
	TheLabelMesh.position.copy( ThePosition );

	return(TheLabelMesh);
}
/**
* Sets up the text meshes
* @private
*/
CM3LayerGraticule.prototype.GetText=function()
{
	if (this.TextMesh==null)
	{
		if ((CM3Text.GlobalFonts!=undefined)&&(CM3Text.GlobalFonts.length>0))
		{
			this.TextMesh = new THREE.Object3D();
			
			var TheGeo=this.GetParent(CMGeo);
			
			var Settings=this.GetTextBufferGeometrySettings();
		
			var HaloSettings=this.GetTextHaloSettings();
			
			// get the material
			var textMaterial = this.GetTextMaterial();
		
			var HaloMaterial = this.GetHaloMaterial(0x000000);
		
			// get settings
			var FontSize=this.GetSetting("Text3D","size",12);
			
			var HeightAboveSeaLevel=this.GetSetting("Graticule3D","LabelHeightAboveSeaLevel");
			var DegreeSpacing=this.GetSetting("Graticule","DegreeSpacing",CMLayerGraticule.MAX_SPACING_DEGREES);
			
			var East=this.GetSetting("Graticule","East",180);
			var West=this.GetSetting("Graticule","West",-180);
			var North=this.GetSetting("Graticule","North",90);
			var South=this.GetSetting("Graticule","South",-90);
			
			var TextSpacing=this.GetSetting("Graticule","TextSpacing",180);
			
			var LongitudePosition=this.GetSetting("Graticule3D","LongitudePosition");
			var LongitudeXOffset=this.GetSetting("Graticule3D","LongitudeXOffset");
			var LongitudeYOffset=this.GetSetting("Graticule3D","LongitudeYOffset");
			
			var LatitudePosition=this.GetSetting("Graticule3D","LatitudePosition");
			var LatitudeXOffset=this.GetSetting("Graticule3D","LatitudeXOffset");
			var LatitudeYOffset=this.GetSetting("Graticule3D","LatitudeYOffset");
			
			//********************************************************************************
			// render the longitude labels long the equator
			
			for (var i=0;i<this.LatitudesWithLongitudeLabels.length;i++) // for each latitude that has longitude labels
			{
				var Latitude=this.LatitudesWithLongitudeLabels[i];
				
				var LongitudeSpacing=this.LongitudeSpacings[i];
				
				if (LongitudeSpacing==360) Longitude=0; // just one label
				else Longitude=-180;
				
				while (Longitude<180)
				{
					var LongitudeOffset=0;//.3; // needs to be based on font size, jjg
					
					// setup the text
					var TheText=Math.abs(Longitude)+String.fromCharCode(176);
					
					if (Longitude<0) TheText+=" W";
					else TheText+=" E";
					
					var TheLabelMesh=this.GetTextMesh(TheGeo,TheText,Settings,textMaterial,Longitude, Latitude, HeightAboveSeaLevel*1.1,
						LongitudePosition,LongitudeXOffset,LongitudeYOffset);
					
					// add the text to the mesh with all the labels
					this.TextMesh.add(TheLabelMesh);
	
					if (true)
					{
						var TheLabelMesh=this.GetTextMesh(TheGeo,TheText,HaloSettings,HaloMaterial,Longitude, Latitude, HeightAboveSeaLevel*0.9,
							LongitudePosition,LongitudeXOffset,LongitudeYOffset);
						
						this.TextMesh.add(TheLabelMesh);
					}
					Longitude+=LongitudeSpacing;
				}
			}
		
			//********************************************************************************
			// render the latitude labels long the 0 and -180 meridians
			
			var LatitudeLongitudinalOffset=this.LatitudeLongitudinalOffset; // jjg - needed? setting?
			var LatitudeLatitudianalOffset=.8;
			
			for (var i=0;i<this.LatitudesWithLatitudeLabels.length;i++)
			{
				var Latitude=this.LatitudesWithLatitudeLabels[i];
				
				var LongitudeSpacing=this.LatitudeSpacings[i];
				
				if (LongitudeSpacing==360) Longitude=0; // just one
				else Longitude=-180;
				
				while (Longitude<180)
				{
					// setup the text
					var TheText=Math.abs(Latitude)+String.fromCharCode(176);
					
					if (Latitude<0) TheText+=" S";
					else TheText+=" N";
					
					//********************************
					// setup the geometry
					var TheLabelMesh=this.GetTextMesh(TheGeo,TheText,Settings,textMaterial,
						Longitude+LatitudeLongitudinalOffset, Latitude, HeightAboveSeaLevel*1.1,
						LatitudePosition,LatitudeXOffset,LatitudeYOffset);
	
					this.TextMesh.add(TheLabelMesh); // add the text to the mesh with all the labels
					
					//********************************
					if (true)
					{
						var TheLabelMesh=this.GetTextMesh(TheGeo,TheText,HaloSettings,HaloMaterial,
							Longitude+LatitudeLongitudinalOffset, Latitude, HeightAboveSeaLevel*0.9,
							LatitudePosition,LatitudeXOffset,LatitudeYOffset);
						
						this.TextMesh.add(TheLabelMesh);// add the text to the mesh with all the labels
					}
					Longitude+=LongitudeSpacing;
				}
			}
		}
	}
}
//**************************************************************************
// Private high level functions
//**************************************************************************
/**
* Private function to handle making the meshes hidden or visible
* @private
*/
CM3LayerGraticule.prototype.SetOGLObjectVisible=function(Flag)
{
	if (typeof(CM3Scene)!== 'undefined')
	{
		var TheScene=this.GetParent(CM3Scene);
		
		if ((TheScene!=null))
		{
			if (Flag)  // add the mesh to the TheOGLScene to make it visible
			{
				this.GetMesh();
				
				// add the line mesh
				if ((this.LineMesh!=null)&&(this.LineMeshInGeo==false))
				{
					TheScene.GetGeo().AddOGLObjectToGeo(this.LineMesh);
					
					this.LineMeshInGeo=true;
				}
				// add the text meshes
				this.GetText();
				
				if ((this.TextMesh!=null)&&(this.TextMeshesInGeo==false))
				{
					TheScene.GetGeo().AddOGLObjectToGeo(this.TextMesh);
					
					this.TextMeshesInGeo=true;
				}
			}
			else // remove the meshfrom the TheOGLScene to make it invisible
			{
				// remove the line mesh
				if (this.LineMeshInGeo)
				{
					TheScene.GetGeo().RemoveOGLObjectFromGeo(this.LineMesh);
					
					this.LineMeshInGeo=false;
				}
				// remove the text meshes
				if (this.TextMeshesInGeo)
				{
					TheScene.GetGeo().RemoveOGLObjectFromGeo(this.TextMesh);
					
					this.TextMeshesInGeo=false;
				}
			}
		}
	}
}
//**************************************************************************
// CMBase Functions
//**************************************************************************

CM3LayerGraticule.prototype.CMLayerGraticule_SettingsChanged=CM3LayerGraticule.prototype.SettingsChanged;

CM3LayerGraticule.prototype.SettingsChanged=function()
{
	this.CMLayerGraticule_SettingsChanged();
	
	this.ClippingBoxProjectedEastings=null;
}
// we want the CMItem definitions but not the CMLayer definitions
CM3LayerGraticule.prototype.CMLayerGraticule3D_GetSettingsDefinitions=CM3LayerGraticule.prototype.GetSettingsDefinitions;

CM3LayerGraticule.prototype.GetSettingsDefinitions=function() 
{
	var Result=this.CMLayerGraticule3D_GetSettingsDefinitions();
	
	for (Key in CM3LayerGraticule.SettingDefintions)
	{
		Result[Key]=CM3LayerGraticule.SettingDefintions[Key];
	}

	return(Result); 
}
//***********************************************************************
// CMItem functions
//***********************************************************************
/*CM3LayerGraticule.prototype.CMLayerGraticule_SetVisible=CMLayerGraticule.prototype.SetVisible;

CM3LayerGraticule.prototype.SetVisible=function(Flag)
{
	if (this.GetVisible()!=Flag)
	{
		this.CMLayerGraticule_SetVisible(Flag);
		
		this.MakeVisible(Flag);
	}
}*/


//******************************************************************
// CMLayer Functions
//******************************************************************
CM3LayerGraticule.prototype.IsVisible=function(TheView) 
{
	return(this.GetVisible()); // jjg for now
}

/**
* Paint one set of graticules based on a clipping bounds in geographic
*
* Approach:
* - GetLatLonRangeFromClippingBounds
* - GetProjectedCoordinateGrid
*
* @protected
*/
CM3LayerGraticule.prototype.Paint=function(TheView) 
{
	// 3D painting
	
	if ((this.IsVisible(TheView))&&(typeof(CM3View)!=='undefined'))
	{
		if (TheView instanceof CM3View)
		{
			this.SetOGLObjectVisible(true);
		}
	}
}
//******************************************************************
// CM3LayerGraticule Functions
//******************************************************************
/**
* Set a custom material for the line
* @public
*/
CM3LayerGraticule.prototype.SetLineMaterial=function(TheMaterial)
{
	this.TheMaterial=TheMaterial;
}

CM3LayerGraticule.prototype.SetTextMaterial=function(TheMaterial)
{
	this.TextMaterial=TheMaterial;
}

