/**************************************************************************************************
* CM3LayerTerrain Class
*
*  Manages a layer that contains an image of the terrain
*  This can include:
*  - A DEM for elevation values
*  - An image to overlay on the DEM
*  - A bump map to add surface texture
*  - A reflectance map to change how light interacts with the surface
*
* @module CM3LayerTerrain
* @Copyright HSU, Jim Graham, 2019
**************************************************************************************************/
//**************************************************************************************************
// CM3LayerTerrain Class
//
// Manages a layer that contains an image of the terrain
// This can include:
// - A DEM for elevation values
// - An image to overlay on the DEM
// - A bump map to add surface texture
// - A reflectance map to change how light interacts with the surface
//**************************************************************************************************

//**************************************************************************************************
// Constructor
//**************************************************************************************************
/*
* Constructor
* @public, @constructs
*/
function CM3LayerTerrain()
{
	CMLayer.call(this);

	this.TheSurface=null;
	this.TheGeometry=null;
	
	this.TimeSlices[0].Settings.Elevation=	
	{
		Exaggeration:1,
		Offset:0
	};
	this.TimeSlices[0].Settings.TerrainMaterial=	
	{
		Transparent:false,
		Color:"rgb(255,255,255)"
	};
};

CM3LayerTerrain.prototype=Object.create(CMLayer.prototype); // inherit prototype functions from PanelBase()

CM3LayerTerrain.prototype.contructor=CM3LayerTerrain; // override the constructor to go to ours
//**************************************************************************************************
// Definitions
//**************************************************************************************************

CM3LayerTerrain.SettingDefintions=
{
	Elevation:
	{
		Exaggeration: { Name:"Exaggeration",Type:CMBase.DATA_TYPE_FLOAT, Default:1 },
		Offset: { Name:"Offset",Type:CMBase.DATA_TYPE_FLOAT, Default:0 },
	},
	Terrain:
	{
		RoundEdges: { Name:"Round Edges",Type:CMBase.DATA_TYPE_BOOLEAN, Default:1 },
	},
	TerrainMaterial:
	{
		Opacity: { Name:"Opacity",Type:CMBase.DATA_TYPE_FLOAT, Default:1 },
		Transparent: { Name:"Transparent",Type:CMBase.DATA_TYPE_BOOLEAN, Default:1 },
		Color: { Name:"Color",Type:CMBase.DATA_TYPE_COLOR, Default:"rgb(255,255,255)" },
		lineCap: { Name:"Side",Type:CMBase.DATA_TYPE_ENUMERATED, Options:[THREE.DoubleSide,THREE.FrontSide,THREE.BackSide],Default:THREE.DoubleSide },
		
		map: { Name:"Overlaid Image (map)",Type:CMBase.DATA_TYPE_STRING,Default:null },
		bumpMap: { Name:"Bump Map",Type:CMBase.DATA_TYPE_STRING,Default:null },
		bumpScale: { Name:"Bump Scale",Type:CMBase.DATA_TYPE_INTEGER,Default:10 },
		specularMap: { Name:"Reflectance Map",Type:CMBase.DATA_TYPE_STRING,Default:null },
		specular: { Name:"Specular Color",Type:CMBase.DATA_TYPE_COLOR,Default:'grey' }
	}
}
//**************************************************************************************************
// Private CM3LayerTerrain Functions
//**************************************************************************************************
/**
* Sets up the material for the terrain based on current settings.
* @private
*/
CM3LayerTerrain.prototype.SetupMaterial=function()
{
	var texloader = new THREE.TextureLoader();
	
	// create the geometry and set the array into it to create the positions (verticies) and normals
	
	var Opacity=this.GetSetting("TerrainMaterial","Opacity",1);
	var Color=this.GetSetting("TerrainMaterial","Color",0xffffff);
	var Transparent=this.GetSetting("TerrainMaterial","Transparent",false);
	var Side=this.GetSetting("TerrainMaterial","Side",THREE.DoubleSide);
	
	var SolidColorMaterial=new THREE.MeshLambertMaterial({
		  color: Color,
		  side:Side ,
		  opacity:Opacity,
		  transparent :false
	});	
	//
	var TheBumpMap=this.GetSetting("TerrainMaterial","bumpMap",null);
	
	if (TheBumpMap!=null) 
	{
		SolidColorMaterial.bumpMap=texloader.load(TheBumpMap); // "texture" of the image
		
		var TheBumpscale=this.GetSetting("TerrainMaterial","bumpScale",10);
		
		SolidColorMaterial.bumpScale = TheBumpscale;
	}
	//
	var TheReflectanceMap=this.GetSetting("TerrainMaterial","specularMap",null);
	
	if (TheReflectanceMap!=null) 
	{
		SolidColorMaterial.specularMap=texloader.load(TheReflectanceMap) // reflectrance
		
		var Color=this.GetSetting("TerrainMaterial","specular",'grey');
		
		SolidColorMaterial.specular=new THREE.Color(Color);
	}
	// create a material with the photographic image of the earth as the "map" (texture)
	var TheImage=this.GetSetting("TerrainMaterial","map",null);
	
	if (TheImage!=null) 
	{
		SolidColorMaterial.map=texloader.load(TheImage); // image
	}

	return(SolidColorMaterial);
}
/**
* Setup a rectangular geometry based on the GeoJSON object provided
* @private
*/
CM3LayerTerrain.prototype.SetupGeometry2=function(TheGeoJSONObject) 
{
	var TheGeo=this.GetParent(CM3Geo);
	var TheElevationOffset=this.GetSetting("Elevation","Offset");
	var TheElevationExaggeration=this.GetSetting("Elevation","Exaggeration");
	
	// get the array of coordinates and elevation values and the mask array form the GeoJSON
	var TheResult=CM3SurfaceGeometry.GetArrayFromGeoJSON(TheGeoJSONObject,TheElevationOffset,TheElevationExaggeration);
	var TheArray=TheResult[0];
	var TheMask=TheResult[1];
	
	// project and flip the y-placement of the coordinates in the array
	for (var Row=0;Row<TheArray.length;Row++)
	{
		var TheRow=TheArray[Row];
		
		for (var Column=0;Column<TheRow.length;Column++)
		{
			var Index=TheRow.length-Column-1;
			var Index=Column;
			
			var Lon=TheRow[Index][0];
			var Lat=TheRow[Index][1];
			var Elevation=TheRow[Index][2]+TheElevationOffset;
			
			// project the coordinate
			
			var Result=TheGeo.ProjectFromGeographic(Lon, Lat, Elevation);
			
			if (Result!=null)
			{
				TheRow[Column]=Result;
			}
		}
	}
	
	var NumColumns=TheArray[0].length;
	var NumRows=TheArray.length;
	
	var RoundEdges=this.GetSetting("Terrain","RoundEdges");
	
	var TheGeometry=new CM3SurfaceGeometry(NumColumns-1,NumRows-1,TheMask,0,1,0,1,RoundEdges);
	
	TheGeometry.SetPositions(TheArray);
				
	return(TheGeometry);
}
//**************************************************************************************************
// Public CMBase Functions
//**************************************************************************************************
CM3LayerTerrain.prototype.CMLayer_GetSettingsDefinitions=CMLayer.prototype.GetSettingsDefinitions; // save the superclass function

CM3LayerTerrain.prototype.GetSettingsDefinitions=function() 
{
	var Result=this.CMLayer_GetSettingsDefinitions(); // get the settings definitions from the superclass
	
	for (Key in CM3LayerTerrain.SettingDefintions)
	{
		Result[Key]=CM3LayerTerrain.SettingDefintions[Key];
	}
	return(Result); 
}

//CM3LayerTerrain.prototype.CMLayer_SetVisible=CMLayer.prototype.SetOGLObjectVisible;

CM3LayerTerrain.prototype.SetOGLObjectVisible=function(Flag)
{
	//if (this.GetVisible()!=Flag)
	{
		//this.CMLayer_SetVisible(Flag);
		
		var TheScene=this.GetParent(CM3Scene);
		
		if (Flag)  // add the mesh to the TheOGLScene to make it visible
		{
			if (this.TheSurface!=null)
			{
				TheScene.GetGeo().AddOGLObjectToGeo(this.TheSurface);
			}
		}
		else // remove the meshfrom the TheOGLScene to make it invisible
		{
			if (this.TheSurface!=null)
			{
				TheScene.GetGeo().RemoveOGLObjectFromGeo(this.TheSurface);
			}
		}
	}
}
//**************************************************************************************************
// Public CMItem Functions
//**************************************************************************************************

CM3LayerTerrain.prototype.Paint=function(TheView) 
{
	if (this.TheTextureDataset!=null)
	{
		var TheStyle=this.GetStyle(TheView);
		
		if (TheStyle!==null) { TheView.SetStyle(TheStyle); }
		
		this.TheTextureDataset.Paint(this,TheView,false);
		
		if (TheStyle!==null) { TheView.RestoreStyle(); }
	}
}

//**************************************************************************************************
// Public CMLayer Functions
//**************************************************************************************************

/**
* Standard function to load data from a URL.
* @public
*/
CM3LayerTerrain.prototype.SetURL=function(DEM_URL) 
{
	//**********************************************************************
	// create the new data set or get an existing one
	if (this.TheDEMDataset==null)
	{
		this.TheDEMDataset=CMDataset.GetDataObject(DEM_URL,CMDataset.RASTER);
	}
	
	// Call the layer's OnLoad() function when the dataset loads the data
	this.TheDEMDataset.AddListener(CMDataset.MESSAGE_DATASET_LOADED,this,function(TheDataset,ThisLayer,AdditionalInfo)
	{
		var TheGeoJSONObject=TheDataset.GetData();
		
		// setup the geometry
		ThisLayer.TheGeometry=ThisLayer.SetupGeometry2(TheGeoJSONObject);
		
		// create the maeterial
		var SolidColorMaterial=ThisLayer.SetupMaterial();

		// setup the mesh
		var TheSurface=new THREE.Mesh(ThisLayer.TheGeometry,SolidColorMaterial);
		
		var PositionOffset=ThisLayer.GetSetting("Item","PositionOffset");
		if (PositionOffset!=null) 
		{
			TheSurface.translateX(PositionOffset[0]);
			TheSurface.translateY(PositionOffset[1]);
			TheSurface.translateZ(PositionOffset[2]);
		}

		// add the mesh to the Geo
		var TheGeo=ThisLayer.GetParent(CM3Geo);
		
		if (ThisLayer.TheSurface!=null)
		{
			TheGeo.RemoveOGLObjectFromGeo(ThisLayer.TheSurface);
			ThisLayer.TheSurface=null;
		}
		ThisLayer.TheSurface=TheSurface;
		
		// setup the layer's bounds
		var Bounds=TheGeoJSONObject.Bounds; // XMin, XMax, YMin, YMax
		
		ThisLayer.SetBounds(Bounds); // set the layers bounds
		
		// if the object is visible, add it to the geo
		if (ThisLayer.GetVisible())
		{
			TheGeo.AddOGLObjectToGeo(TheSurface);
		}
		// call the onload function to zoom to bounds, repaint, etc.
		ThisLayer.OnLoad();
	});
									 
	// add a projector if one is provided
	//if (TheProjector!=undefined) this.TheDEMDataset.SetProjector(TheProjector);
	
	this.TheDEMDataset.SetURL(DEM_URL);

}
//**************************************************************************************************
// Public CM3LayerTerrain Functions
//**************************************************************************************************
/**
* Allows tiles to find the DEM dataset that is associated with this terrain.
* @public
*/
CM3LayerTerrain.prototype.GetDEMDataset=function() 
{
	return(this.TheDEMDataset);
}

/**
* Function to load data from CanvasMap pryamid files.
* @public
*/
CM3LayerTerrain.prototype.SetPyramidURL=function(DEM_URL,TextureURL,DataSetType,TheProjector) 
{
	var TheScene=this.GetParent(CMScene);
	
	//**********************************************************************
	// create the new data set or get an existing one
	if (this.TheDEMDataset==null)
	{
		this.TheDEMDataset=CMDataset.GetDataObject(DEM_URL,DataSetType);
		//this.TheDEMDataset.SetParent(TheScene);  
	}
									 
	// Call the layer's OnLoad() function when the dataset loads the data
	this.TheDEMDataset.AddListener(CMDataset.MESSAGE_DATASET_LOADED,this,function(TheDataset,ThisLayer,AdditionalInfo)
	{
		//****************
		var Bounds=TheDataset.GetBounds(); // XMin, XMax, YMin, YMax
		
		ThisLayer.SetBounds(Bounds); // set the layers bounds
	});
	//
	this.TheDEMDataset.AddListener(CMDataset.MESSAGE_DATASET_TILE_LOADED,this,function(TheTile,ThisLayer,AdditionalInfo)
	{
		ThisLayer.GetScene().Repaint();
	});
									 
	// add a projector if one is provided
	if (TheProjector!=undefined) this.TheDEMDataset.SetProjector(TheProjector);
	
	this.TheDEMDataset.SetURL(DEM_URL);
	//**********************************************************************
	// create the new data set or get an existing one
	if (this.TheTextureDataset==null)
	{
		this.TheTextureDataset=CMDataset.GetDataObject(TextureURL,DataSetType);
		//this.TheTextureDataset.SetParent(TheScene); 
	}
									 
	// Call the layer's OnLoad() function when the dataset loads the data
	this.TheTextureDataset.AddListener(CMDataset.MESSAGE_DATASET_LOADED,this,function(TheDataset,ThisLayer,AdditionalInfo)
	{
		var Bounds=TheDataset.GetBounds(); // XMin, XMax, YMin, YMax
		
		ThisLayer.SetBounds(Bounds); // set the layers bounds
		//****************
		ThisLayer.OnLoad();
		ThisLayer.GetScene().LayerListChanged(); // jjg - a bit of a hack to get the layer list to repaint once we have the data (i.e. can determine if the GeoJSON data is string).
	});
	//
	this.TheTextureDataset.AddListener(CMDataset.MESSAGE_DATASET_TILE_LOADED,this,function(TheDataset,ThisLayer,TheTextureTile)
	{
		ThisLayer.GetScene().Repaint();
		// Pass in a function that is called as soon as the DEM data is available
		// Create the mesh for TheTile and add to the Geo 
	});
									 
	// add a projector if one is provided
	if (TheProjector!=undefined) this.TheTextureDataset.SetProjector(TheProjector);
	
	this.TheTextureDataset.SetURL(TextureURL);
}

