/**************************************************************************************************
* CM3Light Class
* Container class for a 3D light object.  Allows the light to be moved by the user if desired.
*
* @module CM3Light
* @Copyright HSU, Jim Graham, 2019
**************************************************************************************************/
//**************************************************************************************************
// Constructor
//**************************************************************************************************

/*
* Constructor
* @public, @constructs
*/
function CM3Light(Type) 
{
	CM3Item.call(this);

	this.Type=Type;
	
	switch (Type)
	{
	case CM3Light.TYPE_POINT:
		this.SetName("Point");
		break;
	case CM3Light.TYPE_SPOT:
		this.SetName("Spot");
		break;
	case CM3Light.TYPE_DIRECTION:
		this.SetName("Direction");
		break;
	case CM3Light.TYPE_AMBIENT:
		this.SetName("Ambient");
		break;
	case CM3Light.TYPE_HEMISPHERE:
		this.SetName("Hemisphere");
		break;
	}
	this.ControlObjectIsInOGL=false;

	this.TheOGLLight=null;
	this.TheOGLHelper=null; 
	this.LightIsInOGL=false;
	
	this.TimeSlices[0].Settings.ControlObject=	
	{
	};
	this.Initialized=false;
};
CM3Light.prototype = new CM3Item();

CM3Light.prototype.constructor = CM3Light;
//**************************************************************************************************
// Public Definitions
//**************************************************************************************************

CM3Light.TYPE_POINT=1;
CM3Light.TYPE_SPOT=2;
CM3Light.TYPE_DIRECTION=3;
CM3Light.TYPE_AMBIENT=4;
CM3Light.TYPE_HEMISPHERE=5;

CM3Light.SettingDefintions=
{
	Light:
	{
//		Color: { Name:"Fill Color",Type:CMBase.DATA_TYPE_CSS_STYLE, Default:"rgb(255,255,255)" },
		Intensity: { Name:"Intensity",Type:CMBase.DATA_TYPE_FLOAT, Default:1 }, // 0 to 1
		On: { Name:"On",Type:CMBase.DATA_TYPE_BOOLEAN, Default:true }, // 
		AddHelper: { Name:"Add Helper",Type:CMBase.DATA_TYPE_BOOLEAN, Default:false }, // 
	},			
	ControlObject:
	{
		Radius: { Name:"Radius",Type:CMBase.DATA_TYPE_FLOAT, Default:10 }
	}
};
//**************************************************************************************************
// Private functions
//**************************************************************************************************
/**
* @private
*/
CM3Light.prototype.UpdateOGL=function()
{
	var TheScene=this.GetParent(CM3Scene);
	
	if (TheScene!=null)
	{
		// manage making the control object visible or not
		var Status=this.GetSetting("Item","Status",CMItem.STATUS_HIDDEN);
		
		if (this.TheOGLObject!=null)
		{
			if (Status!=CMItem.STATUS_HIDDEN) // visible
			{
				if (this.ControlObjectIsInOGL==false) // not in OGL, add it
				{
					TheScene.AddOGLObject(this.TheOGLObject); // add the control object 
					
					this.ControlObjectIsInOGL=true;
				}
			}
			else  // hidden
			{
				if (this.ControlObjectIsInOGL) // in OGL, remote it
				{
					if (this.TheOGLObject!=null)
					{
						TheScene.RemoveOGLObject(this.TheOGLObject); // remove the control object  
						
						this.ControlObjectIsInOGL=false;
					}
				}
			}
		}
		// manage turning the light on and off by adding it and removing it from the OGL
		
		var On=this.GetSetting("Light","On",true);
		
		if (On)
		{
			if (this.LightIsInOGL==false) // in OGL, remote it
			{
				TheScene.AddOGLObject(this.TheOGLLight);
				
				if (this.TheOGLHelper!=null) TheScene.AddOGLObject(this.TheOGLHelper);
				
				this.LightIsInOGL=true;
			}
		}
		else
		{
			if (this.LightIsInOGL)
			{
				TheScene.RemoveOGLObject(this.TheOGLLight);
				
				if (this.TheOGLHelper!=null) TheScene.RemoveOGLObject(this.TheOGLHelper);
				
				this.LightIsInOGL=false;
			}
		}
	}
}

/**
* TheSize - radius
* @private
*/
CM3Light.prototype.CreateTheOGLControlObject=function(TheSize)
{
	var ArrowPointRadius=TheSize;
	var ArrowPointLength=TheSize*3;
	
	var ThePosition=this.TheOGLLight.position;
	
	var FillColor=this.GetSetting("Style","fillStyle","rgb(255,255,255)");
	var TheColor=TheColor=CMUtilities.GetRGBAFromAnyColor(FillColor);
	TheColor=new THREE.Color(TheColor);
	
	var RedMaterial=new THREE.MeshPhongMaterial( {
		color: TheColor,
		emissive: 0x072534,
		side: THREE.FrontSide,
		flatShading: THREE.FlatShading
	} );
	
	var TheGeometry=null;
	
	switch (this.Type)
	{
	case CM3Light.TYPE_POINT:
		TheGeometry = new THREE.SphereBufferGeometry(TheSize,8,8); // top radius,bottom 		
		break;
	case CM3Light.TYPE_SPOT:
		TheGeometry = new THREE.CylinderGeometry( 0,ArrowPointRadius,ArrowPointLength,4,1 ); // top radius,bottom 		
		break;
	case CM3Light.TYPE_DIRECTION:
		TheGeometry = new THREE.CylinderGeometry( 0,ArrowPointRadius,ArrowPointLength,4,1 ); // top radius,bottom 		
		break;
	}
	
	this.TheOGLObject=new THREE.Mesh( TheGeometry, RedMaterial );
	
	var TheGeo=this.GetParent(CM3Geo);
	if (TheGeo!=null)
	{
		var TheProjector=TheGeo.GetProjector();
		
		if ((typeof(CM3ProjectorSpherical)!="undefined")&&(TheProjector instanceof CM3ProjectorSpherical))
		{
			this.TheOGLObject=TheProjector.GetPositionedMesh(this.TheOGLObject,this.ThePosition.x,this.ThePosition.y,this.ThePosition.z);
		}
		else
		{
			this.TheOGLObject.position.set(this.ThePosition.x,this.ThePosition.y,this.ThePosition.z);
		}
	}
	else this.TheOGLObject.position.set(this.ThePosition.x,this.ThePosition.y,this.ThePosition.z);
	
	return(this.TheOGLObject);
};
/**
* @private
*/
CM3Light.prototype.Initialize=function()
{
	if (this.TheOGLLight==null) // has not been initialized
	{
		this.Initialized=true;
		
		var LightObjectRadius=this.GetSetting("ControlObject","Radius",1);
		var FillColor=this.GetSetting("Style","fillStyle","rgb(255,255,255)");
		var Intensity=this.GetSetting("Light","Intensity",1);
		var AddHelper=this.GetSetting("Light","AddHelper");
		
		var TheColor=TheColor=CMUtilities.GetRGBAFromAnyColor(FillColor);
		TheColor=new THREE.Color(TheColor);
		
		switch (this.Type)
		{
		case CM3Light.TYPE_POINT:
			this.TheOGLLight=new THREE.PointLight(TheColor,Intensity); // color, intensity, distance, decay
			break;
		case CM3Light.TYPE_SPOT:
			this.TheOGLLight = new THREE.SpotLight( TheColor,Intensity ); //color, intensity, distanc, angle, penumbra, decay 
			
			this.TheOGLLight.castShadow = true;
			
			this.TheOGLLight.shadow.mapSize.width = 1024;
			this.TheOGLLight.shadow.mapSize.height = 1024;
			
			this.TheOGLLight.shadow.camera.near = 500;
			this.TheOGLLight.shadow.camera.far = 4000;
			this.TheOGLLight.shadow.camera.fov = 30;
			break;
		case CM3Light.TYPE_DIRECTION:
			this.TheOGLLight = new THREE.DirectionalLight( TheColor,Intensity ); // color, intensity
			if (AddHelper)
			{
				this.TheOGLHelper=new THREE.DirectionalLightHelper( this.TheOGLLight, 5 );
			}
			break;
		case CM3Light.TYPE_AMBIENT:
			this.TheOGLLight = new THREE.AmbientLight( TheColor,Intensity );
			break;
		case CM3Light.TYPE_HEMISPHERE:
			this.TheOGLLight = new THREE.HemisphereLight( TheColor,0x080820,Intensity ); // color, ground color, intensity
			break;
		}

		// create the object that the user can move around
		
		if (this.Type!=CM3Light.TYPE_AMBIENT)
		{
			var TheGeo=this.GetParent(CM3Geo);
			if (TheGeo!=null)
			{
				var TheProjector=TheGeo.GetProjector();
				
				if ((typeof(CM3ProjectorSpherical)!="undefined")&&(TheProjector instanceof CM3ProjectorSpherical))
				{
					this.TheOGLLight=TheProjector.GetPositionedMesh(this.TheOGLLight,this.ThePosition.x,this.ThePosition.y,this.ThePosition.z);
				}
				else
				{
					this.TheOGLLight.position.set(this.ThePosition.x,this.ThePosition.y,this.ThePosition.z);
				}
			}
			else this.TheOGLLight.position.set(this.ThePosition.x,this.ThePosition.y,this.ThePosition.z);
		}
		if ((this.Type==CM3Light.TYPE_POINT)||(this.Type==CM3Light.TYPE_SPOT))
		{
			this.TheOGLObject=this.CreateTheOGLControlObject(LightObjectRadius);
			
			this.TheOGLObject.TheCM3Item=this;
		}
		// turn the light on
		this.SetOGLObjectVisible(true);
		
//		if (this.GetVisible()) this.AddOrRemoveObject(true);

		this.UpdateOGL();
	}
}

//******************************************************************
// CMBase Settings Functions
//******************************************************************
CM3Light.prototype.CM3Item_GetSettingsDefinitions=CM3Item.prototype.GetSettingsDefinitions; // save the superclass function

CM3Light.prototype.GetSettingsDefinitions=function() 
{
	var Result=this.CM3Item_GetSettingsDefinitions(); // get the settings definitions from the superclass
	
	for (Key in CM3Light.SettingDefintions)
	{
		Result[Key]=CM3Light.SettingDefintions[Key];
	}

	return(Result); 
}
	
CMItem.prototype.CMItem_SetSettings=CMItem.prototype.SetSettings;

CM3Light.prototype.SetSettings=function(NewSettings,TimeSlice) 
{
	this.CMItem_SetSettings(NewSettings,TimeSlice);
	
	if (this.TheOGLLight!=null) // if the user changed the light, set it right away
	{
		var TheColor=this.GetSetting("Style","fillStyle");
		TheColor=CMUtilities.GetRGBAFromAnyColor(TheColor);
		TheColor=new THREE.Color(TheColor);
		
		this.TheOGLLight.color=TheColor;
		
		this.Repaint();
		
		this.UpdateOGL();
	}
}
	
CMItem.prototype.CMItem_SetSetting=CMItem.prototype.SetSetting;

CM3Light.prototype.SetSetting=function(Group,Key,Value) 
{
	this.CMItem_SetSetting(Group,Key,Value);
	
	if ((Group=="Item")&&(Key=="Status"))
	{
		this.UpdateOGL();
	}
	else if ((Group=="Style")&&(Key=="fillStyle"))
	{
		if (this.TheOGLLight!=null) // if the user changed the light, set it right away
		{
			var TheColor=CMUtilities.GetRGBAFromAnyColor(Value);
			TheColor=new THREE.Color(TheColor);
			
			this.TheOGLLight.color=TheColor;
			
			this.Repaint();
			
			this.UpdateOGL();
		}
	}
	else if ((Group=="Item")&&(Key=="PositionOffset"))
	{
		var NewPosition=new THREE.Vector3(Value[0],Value[1],Value[2]);
	
		this.ThePosition=NewPosition;
		
		if (this.TheOGLLight!=null)
		{
			this.TheOGLLight.position.copy( NewPosition );
					
			if (this.TheOGLObject!=null)
			{
				this.TheOGLObject.position.copy(NewPosition);
			}
		}
	}
	else if ((Group=="Item")&&(Key=="Rotation"))
	{
		if (this.TheOGLLight!=null)
		{
			this.TheOGLLight.rotateX(Value[0]);
			this.TheOGLLight.rotateY(Value[1]);
			this.TheOGLLight.rotateZ(Value[2]);
		}
	}
}

//**************************************************************************************************
// Public STItem Functions
//**************************************************************************************************

/**
* for 3D items, making them visible means removing the control item
*/
//CM3Light.prototype.CMItem_SetVisible=CMItem.prototype.SetVisible;

CM3Light.prototype.SetOGLObjectVisible=function(Flag)
{
	//if ((this.GetVisible()!=Flag)||(this.Initialized==false))
	{
		//this.CMItem_SetVisible(Flag);
		
		this.Initialize();
		
		this.UpdateOGL();
	}
//	this.CMItem_SetVisible(Flag);
	
};


CM3Light.prototype.Paint=function(TheView) 
{
	this.Initialize();
}

//**************************************************************************************************
// Public functions
//**************************************************************************************************
/**
*
*/
CM3Light.prototype.SetPosition=function(X,Y,Z)
{
	if (X instanceof THREE.Vector3) 
	{
		throw("CM3Light.prototype.SetPosition depricated");
	}
	this.SetSetting("Item","PositionOffset",[X,Y,Z]);
}
