/******************************************************************************
* PathExtrusion
*
* Creates an extruded 3d object along a path with a specified cross section.
* This class is used to extrude a variety of objects including polylines.
*
* Note: This class makes use of the tan2(y,x) function.  This function is documented
* as the angle in the x,y plane being from the x-axis.  This is revsered because
* geographers measure x,y from the y axis (north) (a.k.a. the longitude).  Similarly,
* we measure the angle from the x,y plane (the equator).
*
* Lontidue=tan2(x,y)
* Latitude=tan2(z,xy) - where xy is the distance in the xy plane
* Latitude=sin(z/r)
*
* Resources:
* http://edwilliams.org/avform.htm#Intermediate - Ed Williams page on aviation formulas
*  this contains the key equation for computing segments along a great circle.
* https://en.wikipedia.org/wiki/Spherical_coordinate_system - Wikipedia's spherical
*  coorcinate page. Remember that this is for typical geometry based calculations
*  and we are using geographic lats and lons so the positions of the angles are different
* https://en.wikipedia.org/wiki/Atan2 - Wikipedia page on atan2().  This documentation
*  does not excatly match the JavaScript function as the JS function is valid for all
*  posible values of the angle.
* https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Math/atan2
*  Mizilla documatation for JS atan2 function.  Again, our angles are reversed.
*
* @module CM3PathExtrusion
* @Copyright HSU, Jim Graham, 2019
******************************************************************************/

//******************************************************************************
// Default Constructor
//******************************************************************************

CM3PathExtrusion=function()
{
	CMBase.call(this);
		
	// settings
	this.Settings=	
	{
		PathExtrusion:
		{
			Closed:false,
			EndCap:"Flat",
			CrossSectionAngle:90,
			RadiusOfRotation:40,
			DegreesPerSegment:45,
		},
	};
	
	this.PathArray=null;
	this.CrossSectionArray=null;
	
	//***********************************************
	// properties
	this.TheGeometry=null;
	
	// properties for managing the vertexes and their associated normals
	// both arrays contain sequences of 3 entries representing the 3 elements of 
	// a vertex (coordinate) or a normal (direction vector)
	this.NumVertices=0;
	this.VertexArray=null; // vertex array
	this.NormalsArray=null; // there is always one normal per vertex
	
	// properties for managing the indexes that represent the faces
	// the array contains sequences of 3 entries representing indexes
	// intoe the VertexArray and NormalsArray for the arrays above.
	this.IndexArray=null; // array of indexes into the vertex (and normal) arrays
	this.NumIndexes=0;
	
	// for dealing with intersections
	this.NewPathArray=null;
};
CM3PathExtrusion.prototype = Object.create( CMBase.prototype );

CM3PathExtrusion.prototype.constructor = CM3PathExtrusion;

//**************************************************************************************************
// Setting Definitions
//**************************************************************************************************

CM3PathExtrusion.SettingDefintions=
{
	PathExtrusion:
	{
		Type: { Name:"Type",Type:CMBase.DATA_TYPE_ENUMERATED, Options:["Flat","Pointy","ThreeJS"],Default:"Flat" },
		Closed: { Name:"Closed",Type:CMBase.DATA_TYPE_BOOLEAN, Default:false }, 
		EndCap: { Name:"End Cap",Type:CMBase.DATA_TYPE_ENUMERATED, Options:["Flat","Open","Pointy"],Default:"Flat" },
		CrossSectionAngle: { Name:"Cross Section Rotation",Type:CMBase.DATA_TYPE_FLOAT,Default:0 }, // Angle to rotate the cross section can be rotated about its axis.  This can be used to create a 2 sided cross section and then rotate it so it creates a flat surface similar to a road.
		RadiusOfRotation: { Name:"Radius Of Rotation",Type:CMBase.DATA_TYPE_FLOAT,Default:40 }, // Minimum radius for the path to curve.  Segments that turn faster will be shortened or removed.
		DegreesPerSegment: { Name:"DegreesPerSegment",Type:CMBase.DATA_TYPE_FLOAT,Default:45 }, // Minimum angle in a curve before another segment is added.  The lower the value, the smoother the curve but also the more complex.
		SegementsPerLine: { Name:"SegementsPerLine",Type:CMBase.DATA_TYPE_INTEGER,Default:5 }, // ThreeJS type places segements evenly along the extrusion so this factor if mulitplied by the number of lines in the path to smooth it out
	},
}
;
//******************************************************************************
// Definitions
//******************************************************************************

CM3PathExtrusion.END_NONE=0;
CM3PathExtrusion.END_FLAT=1;

// not implemented yet
CM3PathExtrusion.JOIN_NONE=0;
CM3PathExtrusion.JOIN_FLAT=1;
CM3PathExtrusion.JOIN_ROUND=2;
CM3PathExtrusion.JOIN_MITER=3;

var XAxis=new THREE.Vector3( 1,0, 0 );
var YAxis=new THREE.Vector3( 0,1, 0 );
var ZAxis=new THREE.Vector3( 0,0, 1 );

//**************************************************************************************
// Basic Geometry functions
//**************************************************************************************

CM3PathExtrusion.GetSegmentXYLength=function(P1,P2)
{
	var DX=P1.x-P2.x;
	var DY=P1.y-P2.y;
	
	return(Math.sqrt(DX*DX+DY*DY));
}
CM3PathExtrusion.GetSegmentLength=function(P1,P2)
{
	var DX=P1.x-P2.x;
	var DY=P1.y-P2.y;
	var DZ=P1.z-P2.z;
	
	return(Math.sqrt(DX*DX+DY*DY+DZ*DZ));
}
/**
* Get a new point that is on the line segement P0 to P1.  
* The new point will be the distance D from P0.
* If the distance D is greater than the length of the line segment,
* 	the new point would be beyond the segment so null is returned
*/
CM3PathExtrusion.GetNewPoint=function(D,P0,P1)
{
	var Result=null;
	
	var Distance=CM3PathExtrusion.GetSegmentLength(P0,P1);
	
	var Factor=1; 
	
	if (Distance!=0) 
	{
		// find the proportion of the length of the segment that we need to 
		// move from P1 toward P0
		
		Factor=D/Distance;
	
		if (Factor<0.5) // we are still within the length of the segment
		{
			var DX=P1.x-P0.x;
			var DY=P1.y-P0.y;
			var DZ=P1.z-P0.z;
			
			var x=P0.x+(DX*Factor);
			var y=P0.y+(DY*Factor);
			var z=P0.z+(DZ*Factor);
			
			Result=new THREE.Vector3(x,y,z);
		}
	}
	return(Result);
}
/**
* Makes sure we have a positive angle
*/
CM3PathExtrusion.FixAngle=function(Angle)
{
	while (Angle<0) Angle+=Math.PI*2;
	while (Angle>Math.PI*2) Angle-=Math.PI*2;
	return(Angle);
}
//******************************************************************************
// Private functions to find lat/lon of direction vectors
//******************************************************************************

/*
* Docs shows Math.tan2(y,x) but this is with the angle being from
* the x-axis, since we measure from the y axis, the parameters
* are swapped
*/
CM3PathExtrusion.GetSegmentLongitude=function(P1,P2)
{
	// find the deltas from the current point to the next point
	var dx2=P2.x-P1.x;
	var dy2=P2.y-P1.y;
	
	// x and y are reversed from the docs because we are getting the angle from the y-axis instead of the x-axis
	var Longitude=Math.atan2(dx2,dy2); 
	
	return(Longitude);
}
/**
* Documentation shows Math.tan2(y,x).  Since we measure form the y-axis
* this becomes Math.tan2(x,y).  Because we are looking for the angle in the
* z-plane, this becomes Math.tan2(dxy,dz).
*/
CM3PathExtrusion.GetSegmentLatitude=function(P1,P2)
{
	// find the deltas from the current point to the next point
	var dx2=P2.x-P1.x;
	var dy2=P2.y-P1.y;
	var dz2=P2.z-P1.z;
	
	// get the z angle plane
	dxy=Math.sqrt(dx2*dx2+dy2*dy2); // find the distance on the xy plane
	
	var Latitude=Math.atan2(dxy,dz2); // find the angle (effectively the latitude up from the equator)

	return(Latitude);
}
/**
* Finds the angles in the XY plane (Z=0 plane, Longitude) and the "tilt" from the vertcal (Latitude)
* in the direction from one point (P1) to another point (P2)
* @private
* returns - object Angles, with:
* - Longitude - angle of the face of the plane that is perpendicular to the direction
* of the specified line segment.  The angle is rotated-90 degrees from the direction of the segment
*/
CM3PathExtrusion.GetStartAngles=function(P1,P2)
{
	var Angles={};
	
	// get the angle of the bisecting plane projected to the xy plane
	var Longitude=CM3PathExtrusion.GetSegmentLongitude(P1,P2); // negative to go in direction of the start (jjg - but why?)
	
	Latitude=CM3PathExtrusion.GetSegmentLatitude(P1,P2);
	
	Angles.Longitude=Longitude;
	Angles.Latitude=Latitude;

	return(Angles);
}
//**************************************************************************************
// Functions to compute bisecting angles
//**************************************************************************************
/**
* Finds the 3D angle between the two vectors
* Used algorithm from: http://www.nabla.hr/PC-LinePlaneIn3DSp2.htm#Angle%20between%20lines
*/
CM3PathExtrusion.GetAngleBetweenVectors=function(P0,P1,P2)
{
	var dx1=P1.x-P0.x; // a1
	var dy1=P1.y-P0.y; // b1
	var dz1=P1.z-P0.z; // c1
	
	var dx2=P1.x-P2.x; // a2
	var dy2=P1.y-P2.y; // b2
	var dz2=P1.z-P2.z; // c2
	
	var Length1=Math.sqrt(dx1*dx1+dy1*dy1+dz1*dz1);
	var Length2=Math.sqrt(dx2*dx2+dy2*dy2+dz2*dz2);
	
	var CosAngle=(dx1*dx2+dy1*dy2+dz1*dz2)/(Length1*Length2);
	
	var Angle=Math.acos(CosAngle);
	
	return(Angle);
}

/**
* Finds the angle between the two vectors
*/
CM3PathExtrusion.GetXYAnglesBetweenVectors=function(P0,P1,P2)
{
	var Longitude1=CM3PathExtrusion.GetSegmentLongitude(P0,P1);
	
	var Longitude2=CM3PathExtrusion.GetSegmentLongitude(P1,P2);
	
	var BisectingAngleXY=Longitude2-Longitude1;
	while (BisectingAngleXY<0) BisectingAngleXY+=Math.PI*2;
	while (BisectingAngleXY>Math.PI*2) BisectingAngleXY-=Math.PI*2;
	
	return({
		Longitude1:Longitude1,
		Longitude2:Longitude2,
		BisectingAngleXY:BisectingAngleXY,
	});
}

/**
* Find the angles between the vectors based on a projection into a plane that is
* perpendicular to the z plane
*/
CM3PathExtrusion.GetZAnglesBetweenVectors=function(P0,P1,P2)
{
	// get the angle for each vector
	var Latitude1=CM3PathExtrusion.GetSegmentLatitude(P0,P1);
	
	var Latitude2=CM3PathExtrusion.GetSegmentLatitude(P1,P2);
	
	// find the angle between the vectors
	var BisectingAngleZ=Latitude2-Latitude1;
	while (BisectingAngleZ<0) BisectingAngleZ+=Math.PI*2;
	while (BisectingAngleZ>Math.PI*2) BisectingAngleZ-=Math.PI*2;
	
	return({
		Latitude1:Latitude1,
		Latitude2:Latitude2,
		BisectingAngleZ:BisectingAngleZ,
	});
}
//******************************************************************************
// Private functions to find the point of rotation
//******************************************************************************

/**
* find the RotationPoint based on the RotationPoint being on the same
* line as P1 and P11.  The distance between P1 and P11 is used relative
* to the distance to the RotationPoint to find the new point
*/
CM3PathExtrusion.GetRotationPoint=function(P1,P11,DistanceToRotationPoint)
{
	// find how far we move in each direction from P1 to P11
	var dx=P11.x-P1.x;
	var dy=P11.y-P1.y;
	var dz=P11.z-P1.z;
	
	// find the overall distance from P1 to P11
	var P1ToP11Distance=Math.sqrt(dx*dx+dy*dy+dz*dz);
	
	// find the factor that relates the distance from P1 to P11 to the distance from P11 to the Rotation Point
	var Factor=DistanceToRotationPoint/P1ToP11Distance;
	
	dx*=Factor;
	dy*=Factor;
	dz*=Factor;
	
	var RotationPoint=new THREE.Vector3(P1.x+dx,P1.y+dy,P1.z+dz);
	
	return(RotationPoint);
}

//******************************************************************************
// Private functions to manage the PathArray
//******************************************************************************
/**
* Get the previous point along the path.  If PathIndex is at the start of the path
* the last point in the path will be returned.
*
* @private
*/
CM3PathExtrusion.prototype.GetPreviousPathPoint=function(PathIndex)
{
	var PathArray=this.PathArray;
	
	// find the previous coordinate
	if (PathIndex-1>=0) P0=PathArray[PathIndex-1];
	else P0=PathArray[PathArray.length-1];
	
	return(P0);
}

/**
* Get the next point along the path.  If PathIndex is at the end of the path
* the first point in the path will be returned.
*
* @private
*/
CM3PathExtrusion.prototype.GetNextPathPoint=function(PathIndex)
{
	var PathArray=this.PathArray;
	if (PathIndex+1<PathArray.length) P2=PathArray[PathIndex+1];
	else P2=PathArray[0];
	
	return(P2);
}
//******************************************************************************
// Private functions to manage the VertexArray
//******************************************************************************
/**
* Allocate the vertex and index arrays.
*/
CM3PathExtrusion.prototype.AllocateArrays=function(NumVertices,NumTriangles)
{
	this.VertexArray= new Float32Array( NumVertices * 3 );
	this.NormalsArray= new Float32Array( NumVertices * 3 );
	
	this.IndexArray = new Uint32Array( NumTriangles * 3 ); // there are 3 coords per triangle, 2 triangles per side, this.CrossSectionArray.length-1 sides per tube(row)
	
	this.NumVertices=0;
	this.NumIndexes=0;
	
}
/**
* Size the Vertex, and Normal arrays to make sure we can add
* another entry in each.  The arrays are sized by a factor of 2 to keep 
* performance high
*/
CM3PathExtrusion.prototype.SizeVertexArrays=function()
{
	if (this.VertexArray.length<=(this.NumVertices+1)*3)
	{
		var NewSize=this.VertexArray.length*2;
		
		while (NewSize<(this.NumVertices+1)*3) NewSize*=2;
		
		var TempVertexArray=new Float32Array( NewSize);
		var TempNormalsArray=new Float32Array( NewSize);
		
		for (var i=0;i<this.NumVertices*3;i++)
		{
			TempVertexArray[i]=this.VertexArray[i];
			TempNormalsArray[i]=this.NormalsArray[i];
		}
		this.VertexArray=TempVertexArray;
		this.NormalsArray=TempNormalsArray;
	}
}
CM3PathExtrusion.prototype.SizeIndexArray=function()
{
	if (this.IndexArray.length<=(this.NumIndexes+6)) // have to increase the size of the index array
	{
		var TempIndexArray=new Uint32Array( this.IndexArray.length*2);
		
		for (var i=0;i<this.NumIndexes;i++)
		{
			TempIndexArray[i]=this.IndexArray[i];
		}
		this.IndexArray=TempIndexArray;
	}
}

/**
* Adds a vertex using a point and a vector to the new point.
*
* Note the normals are not currect using this approach
*
* @private
*/
CM3PathExtrusion.prototype.AddVertexFromVector=function(P1,TheVector)
{
	var Result=this.NumVertices; // save the current NumVertices to return as the index to the vertex that is going to be added
	
	this.SizeVertexArrays(this.NumVertices);
	
	var ThreeIndex=this.NumVertices*3;
	
	if ((P1.x+TheVector.x==Infinity)||(P1.y+TheVector.y==Infinity)||(P1.z+TheVector.z==Infinity))
	{
		console.log("CM3PathExtrusion.AddVertexFromVector: coordinate value should not be Infinity");
	}
	// move the vector to the correct position and set it as the position
	this.VertexArray[ThreeIndex]=P1.x+TheVector.x;
	this.VertexArray[ThreeIndex+1]=P1.y+TheVector.y;
	this.VertexArray[ThreeIndex+2]=P1.z+TheVector.z;
	
	// set the normal to go in the direction of the cross section point (only works correctly for round cross-sections)
	var NormalVector=new THREE.Vector3(TheVector.x,TheVector.y,TheVector.z );
//	NormalVector.set( TheVector.x,TheVector.y,TheVector.z );
	NormalVector.normalize();
	
	this.NormalsArray[ThreeIndex]=NormalVector.x*32767;
	this.NormalsArray[ThreeIndex+1]=NormalVector.y*32767;
	this.NormalsArray[ThreeIndex+2]=NormalVector.z*32767;
	
	// error checking
	for (var i=0;i<3;i++)
	{
		if (isNaN(this.VertexArray[ThreeIndex+i])) 
		{
			console.log("CM3PathExtrusion.AddVertexFromVector this.VertexArray["+(ThreeIndex+i)+"] is NaN");
		}
	}
	for (var i=0;i<3;i++)
	{
		if (isNaN(this.NormalsArray[ThreeIndex+i])) 
		{
			console.log("CM3PathExtrusion.AddVertexFromVector this.NormalsArray["+(ThreeIndex+i)+"] is NaN");
		}
	}
	
	this.NumVertices++;
	
	return(Result);
}
/**
* Adds an entry to the positions array.
*
* Note the normals are not currect using this approach
*
* @private
*/
CM3PathExtrusion.prototype.AddVertex=function(P1,NormalVector)
{
	var Result=this.NumVertices; // return the index to the new vertex
	
	var ThreeIndex=this.NumVertices*3;
	
	this.SizeVertexArrays();
	
	this.VertexArray[ThreeIndex]=P1.x;
	this.VertexArray[ThreeIndex+1]=P1.y;
	this.VertexArray[ThreeIndex+2]=P1.z;
	
	NormalVector.normalize();
	
	this.NormalsArray[ThreeIndex]=NormalVector.x*32767;
	this.NormalsArray[ThreeIndex+1]=NormalVector.y*32767;
	this.NormalsArray[ThreeIndex+2]=NormalVector.z*32767;
	
	// error checking
	for (var i=0;i<3;i++)
	{
		if (isNaN(this.VertexArray[ThreeIndex+i])) 
		{
			console.log("CM3PathExtrusion.AddVertex this.VertexArray["+(ThreeIndex+i)+"] is NaN");
		}
	}
	for (var i=0;i<3;i++)
	{
		if (isNaN(this.NormalsArray[ThreeIndex+i])) 
		{
			console.log("CM3PathExtrusion.AddVertex this.NormalsArray["+(ThreeIndex+i)+"] is NaN");
		}
	}
	this.NumVertices++;
	
	return(Result);
}
//******************************************************************************
// Private functions to manage the IndexesArray
//******************************************************************************

/**
* Adds a triangular fan to the indexes using a ring of vertices and then a center vertex.
*
* @private
*/
CM3PathExtrusion.prototype.AddTriangleFan=function(StartIndex1,NumTriangles,CoordinateIndex,ReverseDirection)
{
	var Index1Next;
	
	var Index1=StartIndex1;
	
	for (var SectionIndex=0;SectionIndex<NumTriangles;SectionIndex++)
	{
		if (SectionIndex+1<NumTriangles)  // get the next index to a coordinate
		{
			Index1Next=Index1+1;
		}
		else // go back to the start
		{
			Index1Next=StartIndex1;
		}
	
		this.SizeIndexArray();
		// add the next triangle
		if (ReverseDirection)
		{
			this.IndexArray[this.NumIndexes++]=CoordinateIndex;
			this.IndexArray[this.NumIndexes++]=Index1Next;
			this.IndexArray[this.NumIndexes++]=Index1;
		}
		else
		{
			this.IndexArray[this.NumIndexes++]=Index1;
			this.IndexArray[this.NumIndexes++]=Index1Next;
			this.IndexArray[this.NumIndexes++]=CoordinateIndex;
		}
		
		Index1++;
	}
}

/**
* Adds triangles to connect two rings of vertices to create a triangular strip.
*
* @param StartIndex1 - Index to first point in the first loop
* @param StartIndex2 - Index to first point in the first loop
* @param NumTriangles - Number of triangles to add
* @private
*/
CM3PathExtrusion.prototype.AddTrianguleStrip=function(StartIndex1,StartIndex2,NumTriangles)
{
	var Index1Next;
	var Index2Next;
	
	var Index1=StartIndex1;
	var Index2=StartIndex2;
	
	for (var SectionIndex=0;SectionIndex<NumTriangles;SectionIndex++)
	{
		if (SectionIndex+1<NumTriangles)  // get the next index to a coordinate
		{
			Index1Next=Index1+1;
			Index2Next=Index2+1;
		}
		else // go back to the start
		{
			Index1Next=StartIndex1;
			Index2Next=StartIndex2;
		}
	
		this.SizeIndexArray();
		
		this.IndexArray[this.NumIndexes++]=Index2Next;
		this.IndexArray[this.NumIndexes++]=Index1Next;
		this.IndexArray[this.NumIndexes++]=Index1;
		
		this.IndexArray[this.NumIndexes++]=Index1;
		this.IndexArray[this.NumIndexes++]=Index2;
		this.IndexArray[this.NumIndexes++]=Index2Next;
		
		Index1++;
		Index2++;
	}
}
//*********************************************************************************
// Extrudes with "flat" connections
//*********************************************************************************
/**
*	Converts the cross section to if the specifed parameters
* @param Angles - contains the longitude and latitude of the line segment.
*/
CM3PathExtrusion.prototype.GetCrossSectionVectors=function(CrossSectionAngle,Angles)
{
	var TheVectors=[];
	
	var NumCrossSectionCoors=this.CrossSectionArray.length;
	
	// jjg- I'm not sure why this is required
	var Latitude=-Math.PI/2+Angles.Latitude;
	
	// rotate the longidue to be along the face of the cross section
	var Longitude=Math.PI/2-Angles.Longitude;
	
	//*********************************************************************************
	// go around the cross section for each segment of the current tube (start of the tube)
	for (var SectionIndex=0;SectionIndex<NumCrossSectionCoors;SectionIndex++)
	{
		var TheVector=new THREE.Vector3();
		
		// get the vector for the cross section and make a copy of it
		TheVector.copy(this.CrossSectionArray[SectionIndex]);
		
		// rotate the cross section if desired by the user
		if (CrossSectionAngle!=0)  TheVector.applyAxisAngle( ZAxis, CrossSectionAngle );
	
		// apply the angles
		TheVector.applyAxisAngle( ZAxis,-Math.PI/2 ); // rotate the cross section so that x becomes y 
		
		TheVector.applyAxisAngle( YAxis,Math.PI/2 ); // "stand up" the cross section" (cross section is now on the yz plane ready to be rotated)
		
		// apply the angles to tilt and rotate to match the vectors
		TheVector.applyAxisAngle( YAxis,Latitude ); // // tilt the axis about the y-axis
		
		TheVector.applyAxisAngle( ZAxis,Longitude ); // rotate to the correct angle
		
		// add the vector to the positions and normals
	//	this.AddVertexFromVector(P1,TheVector);
		
		TheVectors.push(TheVector);
	}
	return(TheVectors);
}
CM3PathExtrusion.prototype.AddCrossSection=function(P1,TheVectors)
{
	var StartIndex1=this.NumVertices; // save the start of the first loop for the first section of faces
	for (var i=0;i<TheVectors.length;i++)
	{
		this.AddVertexFromVector(P1,TheVectors[i]);
	}
	return(StartIndex1);
}

/**
* Adds a flat end to the extrusion at P2 based on TheVectors which represent vectors from
* P2 to the edges of the flat end.  A center vertex will be added a P2 and normals will point
* in the direction from P1 to P2
*/
CM3PathExtrusion.prototype.AddFlatEnd=function(P2,P1,TheVectors,ReverseDirection)
{
	// set the normal to go in the direction of the cross section point (only works correctly for round cross-sections)
	var NormalVector=new THREE.Vector3(P2.x-P1.x,P2.y-P1.y,P2.z-P1.z); 
	NormalVector.normalize();

	var StartIndex=this.NumVertices; // save the start of the first loop for the first section of faces
	
	for (var i=0;i<TheVectors.length;i++)
	{
		var TheVertex=TheVectors[i];
		
		var NewVector=new THREE.Vector3(P2.x+TheVertex.x,P2.y+TheVertex.y,P2.z+TheVertex.z);
		
		this.AddVertex(NewVector,NormalVector);
	}
	var CenterVertexIndex=this.AddVertex(P2,NormalVector); //add the center point
		
	// add a triangle fan to the end of the tube
	this.AddTriangleFan(StartIndex,TheVectors.length,CenterVertexIndex,ReverseDirection);
}
//*******************************************************************************************
// Private functions to find the angles at the connecting coordinates
//*******************************************************************************************
/**
* Finds the angles in the XY plane (Z=0 plane) and the "tilt" from the vertcal (Latitude)
* based on the plane that bisects the intersection of two tubes.
* @private
*/
CM3PathExtrusion.GetBisectingAngles=function(P0,P1,P2)
{
	// find the angles between the vectors
	var Result=CM3PathExtrusion.GetXYAnglesBetweenVectors(P0,P1,P2);
	BisectingAngle=Result.BisectingAngleXY;
	S1Angle=Result.Longitude1;
	S2Angle=Result.Longitude2;
	
	// find the angle that bisects the two vectors at their intersection
	var HalfBisectingAngle=BisectingAngle/2;
	
	//var RotationAngleAboutZ=S1Angle-Math.PI/2+HalfBisectingAngle;
		
	//*********************************************************************************
	// find the factors
	
	// find the amount to scale the cross section so it matches the intersection angle
	XFactor=1/Math.cos(HalfBisectingAngle); // factor to scale cross section in x direction
	if (XFactor>5) XFactor=5; // limit the amount of pointiness
	
	// compuate the y factor based on a projection of the vectors in a plane that is perpendicular to the xy plane and
	// contains the vectors.
	
	var Result=CM3PathExtrusion.GetZAnglesBetweenVectors(P0,P1,P2);
	var BisectingAngleZ=Result.BisectingAngleZ;
	
	// find the angle that bisects the two vectors at their intersection
	YFactor=1;
	var HalfBisectingAngleZ=BisectingAngleZ/2;
	YFactor=1/Math.cos(HalfBisectingAngleZ); // factor to scale cross section in x direction
	if (YFactor<0) YFactor=-YFactor;
	
	//*********************************************************************************
	// find the angles to rotate the bisecting planes
	var Vector1=new THREE.Vector3(P1.x-P0.x,P1.y-P0.y,P1.z-P0.z); // vector from P1 to P0
	var Vector2=new THREE.Vector3(P1.x-P2.x,P1.y-P2.y,P1.z-P2.z); // vector from P1 to P2
	
	// normalize the vectors (make the distance along the vectors the same)
	Vector1.normalize();
	Vector2.normalize();
	
	// these vectors can now be treated as two points
	// find the vector that goes from one point to the other
	
	var NormalToPlane=new THREE.Vector3(Vector1.x-Vector2.x,Vector1.y-Vector2.y,Vector1.z-Vector2.z);
	NormalToPlane.normalize();

	// these deltas now represent the offset from the normal for the z-plane to the normal for the bisecting plane
	var DX=NormalToPlane.x;
	var DY=NormalToPlane.y;
	var DZ=NormalToPlane.z;
	
	// rotate the cross section to match the bisecting plane around the z axis (the cross section will still be vertical)
	Longitude=Math.atan2(DY,DX); // effectively, latitude of the vector that is perpendicular to the bisecting plane
	
	// rotate the cross section to match the bisecting plane around the x axis (this tilts the plan into position)
	var XYDistance=Math.sqrt(DX*DX+DY*DY); // distance projected to the XY plane (i.e. Z=0)
	
	Latitude=-Math.atan2(DZ,XYDistance); // opposite over adjacent
	
	Angles={};
	Angles.Longitude=Longitude;
	Angles.Latitude=Latitude;

	return(Angles);
}
/**
* Create an extrusion along a series of line segements (path).  This approach
* creates tubes for each line segment where the ends of the tubes are perpendicular
* to the line segments direction.  The tubes will overlap at the corners and there will
* be gaps on the outside of the turns.  The gaps are filled with another series of surfaces.
*
* Note: this version connects the ends by finding the bisecting plane between the vectors
* on either side of the current point.  This gives the long pointy extensions to the extrusion
*
* @public
*/
CM3PathExtrusion.prototype.ComputeGeometryBisecting=function()
{
	var Closed=this.GetSetting("PathExtrusion","Closed",false);
	var EndCap=this.GetSetting("PathExtrusion","EndCap","Flat");
	var CrossSectionAngle=this.GetSetting("PathExtrusion","CrossSectionAngle",0)*Math.PI/180;
	
	// Compute the number of vertexes and triangles
	
	var NumCrossSectionCoors=this.CrossSectionArray.length; // number of pairs of faces around the extrusion (i.e. number of triangles per fan/2)
	
	var NumTubes=this.PathArray.length-1; // number of sections along the extrusion (i.e. number of triangle fans)
	if (Closed) NumTubes+=1; // add a section to close the extrusion
	
	var NumTriangles=NumCrossSectionCoors*NumTubes*2; // number of trianglular faces in each section
	
	var NumVertices=NumCrossSectionCoors*this.PathArray.length; // total number of verticies
	
	// create the position, normal, and index arrays
	
	this.VertexArray= new Float32Array( NumVertices * 3 );
	this.NormalsArray= new Int16Array( NumVertices * 3 );
	
	this.IndexArray = new Uint32Array( NumTriangles * 3 );
	
	// create the temporary variables
	
	var TheVector=new THREE.Vector3();
	
	var NewVector=new THREE.Vector3();
	
	var StartIndex1=-1; // CoordinateIndex for the first coordinate on the end cross section
	var StartIndex2=0; // CoordinateIndex for the first coordinate on the start cross section
	
	var P0,P1,P2; // variables for each point's vector along the path
	
	// add each segment to the geometry
	
	var NumPathSegments=this.PathArray.length;
	var Longitude;
	var Latitude;
	var XFactor=1;
	var YFactor=1;
	
	// render the 3d extrusion along the path
	for (var PathIndex=0;PathIndex<NumPathSegments;PathIndex++) // for each point in the extrusion path, 
	{
		// get the points before and after the current one
		P0=this.GetPreviousPathPoint(PathIndex);
		P1=this.PathArray[PathIndex];
		P2=this.GetNextPathPoint(PathIndex);
		
		var Angles={};
		
		// set factors to the default in case the ends are open
		XFactor=1;
		YFactor=1;
	
		// find the rotation vectors 
		if ((Closed==false)&&(PathIndex==0)) // start end is open
		{
			Angles=CM3PathExtrusion.GetStartAngles(P1,P2);
		}
		else if ((Closed==false)&&(PathIndex==NumPathSegments-1)) // start end is open
		{
			Angles=CM3PathExtrusion.GetStartAngles(P1,P0);
			
			// swap direction of angles for the end
			Angles.Longitude=-Angles.Longitude;
			Angles.Latitude=-Angles.Latitude;
		}
		else
		{
			Angles=CM3PathExtrusion.GetBisectingAngles(P0,P1,P2);
		}
		//*********************************************************************************
		// setup the locals
		StartIndex2=this.NumVertices; // save the start of the first loop for the first section of faces
		
		//*********************************************************************************
		// go around the cross section for each segment of the path (start of the tube)
		for (var SectionIndex=0;SectionIndex<NumCrossSectionCoors;SectionIndex++)
		{
			// get the vector for the cross section and make a copy of it
			TheVector.copy(this.CrossSectionArray[SectionIndex]);
			
			// rotate the cross section if desired by the user
			if (CrossSectionAngle!=0) 
			{
				TheVector.applyAxisAngle( ZAxis, CrossSectionAngle );
			}
		
			TheVector.x=TheVector.x*XFactor; 	// scale the cross section based on the amount of angle between the two vectors (works)
			TheVector.y=TheVector.y*YFactor; 	// scale the cross section based on the amount of angle between the two vectors (works)
			
			TheVector.applyAxisAngle( ZAxis,-Math.PI/2 ); // rotate the cross section so that x becomes y 
			
			TheVector.applyAxisAngle( YAxis,Math.PI/2 ); // "stand up" the cross section" (cross section is now on the yz plane ready to be rotated)
			
			// apply the angles to tilt and rotate to match the vectors
			TheVector.applyAxisAngle( YAxis,Angles.Latitude ); // // tilt the axis about the y-axis
			
			TheVector.applyAxisAngle( ZAxis,Angles.Longitude ); // rotate to the correct angle
				
			// add the vector to the positions and normals
			this.AddVertexFromVector(P1,TheVector);
		}
		//*********************************************************************************
		// Create the array of indexes into the positions for each triangle that is a face
		
		if (StartIndex1!=-1)
		{
			this.AddTrianguleStrip(StartIndex1,StartIndex2,NumCrossSectionCoors);
		}
		// the current loop becomes the second loop for the next time around
		StartIndex1=StartIndex2;
	}
	
	if (Closed)
	{
		this.AddTrianguleStrip(StartIndex1,0,NumCrossSectionCoors);
	}
}
//*********************************************************************************
// Extrudes with "flat" connections
//*********************************************************************************
/**
* Completes the connections between segments with sections based on a rotation 
* radius.  This function can also be used to complete the ends with just a flat
* end
*/
CM3PathExtrusion.prototype.ComputeGeometryFlat=function()
{
	//*********************************************************************************
	var Closed=this.GetSetting("PathExtrusion","Closed",false);
	var EndCap=this.GetSetting("PathExtrusion","EndCap","Flat");
	var CrossSectionAngle=this.GetSetting("PathExtrusion","CrossSectionAngle",0);
	CrossSectionAngle=CrossSectionAngle*Math.PI/180;
	var RadiusOfRotation=this.GetSetting("PathExtrusion","RadiusOfRotation",40);
	var DegreesPerSegment=this.GetSetting("PathExtrusion","DegreesPerSegment",12);
	
	//*********************************************************************************
	// Compute the number of vertexes and triangles
	var NumCrossSectionCoors=this.CrossSectionArray.length; // number of pairs of faces around the extrusion (i.e. number of triangles per fan/2)
	var NumSidesPerTube=this.CrossSectionArray.length-1; // number of pairs of faces around the extrusion (i.e. number of triangles per fan/2)
	
	// find the number of tubes 
	var NumTubes=this.PathArray.length-1; // number of sections along the extrusion (i.e. number of triangle fans)
	if (Closed) NumTubes+=1; // add a section to close the extrusion
	
	//  find the number of connectors
	var NumConnectors=this.PathArray.length-2; // number of connectors between each tube
	if (Closed) NumConnectors+=2; //  add a connector before and after the tube that is closing the polyline
	
	// there are two triangles per side of the tbues
	var NumTrianglesPerTube=NumSidesPerTube*2;
	
	var NumTriangles=NumTrianglesPerTube*(NumTubes+NumConnectors)*2*2; // number of trianglular faces in each section (two triangles per tube side, times number of tubes
	
	var NumVertices=(NumCrossSectionCoors)*2*this.PathArray.length*2*2; // there are always two loops or points around each tube
	
	if ((Closed==false)&&(EndCap=="Flat")) // ends closed with a flat cap
	{
		NumVertices+=2;
	}
				
	//*********************************************************************************
	// create the position, normal, and index arrays
	
	this.AllocateArrays(NumVertices,NumTriangles);
	
	//*********************************************************************************
	// create the temporary variables
	
	var StartIndex1=-1; // CoordinateIndex for the first coordinate on the end cross section
	var StartIndex2=-1; // CoordinateIndex for the first coordinate on the start cross section
	
	var P0,P1,P2; // variables for each point's vector along the path
	
	//*********************************************************************************
	// Find the path that keeps the sections from intersecting within RadiusOfRotation
	// This creates NewPathArray
	
	var NewPathArray=[];
	var NewAngles=[];
	var CrossSectionVertexStartIndexes=[];
	
	// these two variables are used to move the P1 to a point mid-way between P0 and P2
	var NumConsecutiveSkips=0;
	var P1Index=0;
	
	P0=this.PathArray[this.PathArray.length-1];
	P1=this.PathArray[P1Index];
	
	for (var PathIndex=0;PathIndex<this.PathArray.length;PathIndex++) // for each tube in the extrusion path
	{
		//*********************************************************************************
		// get the points before and after the current one
		
		if (PathIndex+1<this.PathArray.length) P2=this.PathArray[PathIndex+1];
		else P2=this.PathArray[0];
		
		if (PathIndex==0) // just add the first point
		{
			NewPathArray.push(P1);
			
			var Angles1=CM3PathExtrusion.GetStartAngles(P1,P2);
			
			var TheVectors=this.GetCrossSectionVectors(CrossSectionAngle,Angles1);
			
			// add the cross section for the tube
			StartIndex1=this.AddCrossSection(P1,TheVectors);
			CrossSectionVertexStartIndexes.push(StartIndex1);
			//
			P0=P1;
			P1=P2;
		}
		else if (PathIndex==this.PathArray.length-1) // just add the last point
		{
			NewPathArray.push(P1);
			
			var Angles1=CM3PathExtrusion.GetStartAngles(P0,P1);
			
			var TheVectors=this.GetCrossSectionVectors(CrossSectionAngle,Angles1);
			
			StartIndex2=this.AddCrossSection(P1,TheVectors);
			CrossSectionVertexStartIndexes.push(StartIndex2);
			//
			P0=P1;
			P1=P2;
		}
		else // find the intersections of the two points and add points that do not intersect (i.e. are within RadiusOfRotation distance)
		{
			//*********************************************************************************
			// find the angle between the two line segments in 3D space (i.e. the minimum angle)
			var AngleIn3D=CM3PathExtrusion.GetAngleBetweenVectors(P0,P1,P2);
			var HalfAngleIn3D=AngleIn3D/2;
			
			// find the distance 
			
			var DistanceToNewPoints=RadiusOfRotation/Math.tan(HalfAngleIn3D); 
			var DistanceToRotationPoint=RadiusOfRotation/Math.sin(HalfAngleIn3D); 
			
			//*********************************************************************************
			// Find the new points on the line segments that will be used to start and end the curve
			var P10=CM3PathExtrusion.GetNewPoint(DistanceToNewPoints,P1,P0);
			var P12=CM3PathExtrusion.GetNewPoint(DistanceToNewPoints,P1,P2);
			
			// if the points are within the existing length of the line segment, create the curve
			// (otherwise, P2 will be skipped)
			if ((P10!=null)&&(P12!=null))
			{
				// save the points for debugging
				this.P10=P10;
				this.P12=P12;
				
				//*********************************************************************************
				// find the point that is mid-way between the new points on the line segments
				// this point is co-linear with the RotationPoint and P1
				
				// find the point that is mid-way between the two new points on the line segments
				// this point will be on the line that also contains the intersection point (P1)
				// and the rotation point.
				var P11=new THREE.Vector3((P10.x+P12.x)/2,(P10.y+P12.y)/2,(P10.z+P12.z)/2);
				this.P11=P11;
				
				// find the rotation point based on it's being colinear with P1 and P11
				// and being DistanceToRotationPoint
				this.RotationPoint=CM3PathExtrusion.GetRotationPoint(P1,P11,DistanceToRotationPoint);
				
				//*********************************************************************************
				// find the angles between the two line segements in the XY and Z planes and which direction we are turning
				
				// get the BisectingAngle in the XY plane
				var XYAngles=CM3PathExtrusion.GetXYAnglesBetweenVectors(P0,P1,P2);
				var BisectingAngleXY=Math.PI-XYAngles.BisectingAngleXY; // angle between the two line segmetns (jjg needs work for crossing 360 to 0)
		
				//*********************************************************************************
				// find the rotation vectors for the existing segments (we'll start at Angles1 and add tubes to gradually get to Angles2)
				var AnglesP0P1=CM3PathExtrusion.GetStartAngles(P0,P1); // Longitude - direction vector - 90 degrees
				var AnglesP1P2=CM3PathExtrusion.GetStartAngles(P1,P2); // 
	
				//*********************************************************************************
				// add the first point and it's cross section
				NewPathArray.push(P10);
				
				var TheVectors=this.GetCrossSectionVectors(CrossSectionAngle,AnglesP0P1);
				
				StartIndex1=this.AddCrossSection(P10,TheVectors);
				CrossSectionVertexStartIndexes.push(StartIndex1);
				
				//*********************************************************************************
				// find the number of segments to create based on the overall angle between the segments in 3d space
				
				var CrackAngle=Math.PI-AngleIn3D; // find the angle of the crack opposite where the tubes intersect
				
				var RadiansPerSegment=DegreesPerSegment/180*Math.PI;
				
				var NumSegments=Math.floor(CrackAngle/RadiansPerSegment);
				
				// only have to add other curve segments if we need more than one
				if (NumSegments>1)
				{
					//*********************************************************************************
					// find the value d to compute the line segments along a great arc circle
					
					// get the lat/lon for the first point
					var DX0=P10.x-this.RotationPoint.x;
					var DY0=P10.y-this.RotationPoint.y;
					var DZ0=P10.z-this.RotationPoint.z;
					
					var r0=Math.sqrt(DX0*DX0+DY0*DY0+DZ0*DZ0);
					var lon0=Math.atan2(DY0,DX0); // 
					var lat0=Math.asin(DZ0/r0); // 
					
					// get the lat/lon for the second point
					var DX2=P12.x-this.RotationPoint.x;
					var DY2=P12.y-this.RotationPoint.y;
					var DZ2=P12.z-this.RotationPoint.z;
					
					var r2=Math.sqrt(DX2*DX2+DY2*DY2+DZ2*DZ2);
					var lon2=Math.atan2(DY2,DX2); // 
					var lat2=Math.asin(DZ2/r2); // 
					
					// make sure the lat is within -90 to 90
					if (lat2>Math.PI/2) lat2-=Math.PI/2; //
					
					// find the "distance" (in radians) along the great circle from point 0 to point 1
					var d = Math.acos(Math.sin(lat0) * Math.sin(lat2) + Math.cos(lat0) * Math.cos(lat2) * Math.cos(lon0 - lon2))
					
					//*********************************************************************************
					// add the inbetween cross sections
					
					this.RotationPoints=[]; // setup the array for the rotation points (debugging)
					
					var LastPoint=P10;
					
					for (var j=0;j<NumSegments-1;j++)
					{
						//****************************************************
						// find the next point along the great arc circle
						var f=(j+1)/NumSegments;
						
						var A=Math.sin((1-f)*d)/Math.sin(d);
						var B=Math.sin(f*d)/Math.sin(d);
						
						var DX = A*Math.cos(lat0)*Math.cos(lon0) +  B*Math.cos(lat2)*Math.cos(lon2);
						var DY = A*Math.cos(lat0)*Math.sin(lon0) +  B*Math.cos(lat2)*Math.sin(lon2);
						var DZ = A*Math.sin(lat0)+B*Math.sin(lat2);
						
						DX*=r0;
						DY*=r0;
						DZ*=r0;
						
						// create the new point and save it
						var NewPoint=new THREE.Vector3(this.RotationPoint.x+DX,this.RotationPoint.y+DY,this.RotationPoint.z+DZ);
						this.RotationPoints.push(NewPoint);
						
						NewPathArray.push(NewPoint);
						
						//****************************************************
						NewAngles=CM3PathExtrusion.GetStartAngles(LastPoint,NewPoint);
						
						// find the vectors and then use them to add the cross section
						// cross section is rotated Latitude about y axis
						// rotated Longitude about the z axis
						var TheVectors=this.GetCrossSectionVectors(CrossSectionAngle,NewAngles);
						
						StartIndex1=this.AddCrossSection(NewPoint,TheVectors);
						CrossSectionVertexStartIndexes.push(StartIndex1);
						
						LastPoint=NewPoint;
					}
					
				}
				// add the start of the next tube (this ends the curve)
				
				var TheVectors=this.GetCrossSectionVectors(CrossSectionAngle,AnglesP1P2);
				
				StartIndex1=this.AddCrossSection(P12,TheVectors);
				CrossSectionVertexStartIndexes.push(StartIndex1);
				
				//
				P0=P1;
				P1=P2;
				
				NumConsecutiveSkips=0; // did not skip
			}
			else // there was a problem with having enough room beween P0,P1,P2 to make the turn so skip a point
			{
				if (NumConsecutiveSkips==0) // first time we have skipped
				{
					P1Index=PathIndex;
				}
				else // P1 moves forward at 1/2 the pace of P2 so the points get spread out
				{
//					console.log(NumConsecutiveSkips%2);
					if ((NumConsecutiveSkips%2)==1) // odd skip, move PI forward
					{
						P1Index++;
						P1=this.PathArray[P1Index];
					}
				}
				NumConsecutiveSkips++;
			}
		}
		
	}
	
	//*********************************************************************************
	// Deal with the start of the tube (this creates a triangular fan that is separate from the 
	// other triangles
	
	// if this is the first tube and the extrusion is not closed, and they want a flat cap on it, 
	// add a flat end.
	if ((Closed==false)&&(EndCap=="Flat")) // close the end of the first tube
	{	
		P1=this.PathArray[0];
		P2=this.PathArray[1];
		
		var AnglesP1P2=CM3PathExtrusion.GetStartAngles(P1,P2);
		
		var TheVectors=this.GetCrossSectionVectors(CrossSectionAngle,AnglesP1P2);
	
		this.AddFlatEnd(P1,P2,TheVectors,false);
	}
		
	//*********************************************************************************
	// add the triangle strips that make up the tube
	var NumCrossSections=CrossSectionVertexStartIndexes.length;
	
	for (var i=0;i<NumCrossSections-1;i++)
	{
		this.AddTrianguleStrip(CrossSectionVertexStartIndexes[i],CrossSectionVertexStartIndexes[i+1],NumSidesPerTube);
	}
	// Add the end of the last tube
	
	if (Closed) // Add the connector between this tube and the previous tube
	{
		this.AddTrianguleStrip(StartIndex2,0,NumSidesPerTube);
	}
	else
	{
		if (EndCap=="Flat") // close the end of the first tube
		{	
			P1=this.PathArray[this.PathArray.length-1];
			P2=this.PathArray[this.PathArray.length-2];
			
			var AnglesP2P1=CM3PathExtrusion.GetStartAngles(P2,P1);
			
			var TheVectors=this.GetCrossSectionVectors(CrossSectionAngle,AnglesP2P1);
		
			this.AddFlatEnd(P1,P2,TheVectors,true);
		}
	}
}

//*********************************************************************************
//
//*********************************************************************************
/**
* This approach uses the THREE.js ExtrudeBufferGeometry function
* @private
*/
CM3PathExtrusion.prototype.ExtrudeBufferGeometry=function()
{
	var Closed=this.GetSetting("PathExtrusion","Closed",false);
	var EndCap=this.GetSetting("PathExtrusion","EndCap","Flat");
	var CrossSectionAngle=this.GetSetting("PathExtrusion","CrossSectionAngle",0)*Math.PI/180;
	var SegementsPerLine=this.GetSetting("PathExtrusion","SegementsPerLine");
	
	// get the shape
	var TheShape = new THREE.Shape();
		
	for (var i=0;i<this.CrossSectionArray.length;i++)
	{
		TheVector=this.CrossSectionArray[i];
		
		if (CrossSectionAngle!=0) 
		{
			TheVector.applyAxisAngle( ZAxis, CrossSectionAngle );
		}
		if (i==0)
		{
			TheShape.moveTo( TheVector.x,TheVector.y );  
		}
		else
		{
			TheShape.lineTo( TheVector.x,TheVector.y );  
		}
	}
	
	//
	
	var ExtrusionPath =  new THREE.CatmullRomCurve3( this.PathArray );
	ExtrusionPath.curveType = 'catmullrom';
	ExtrusionPath.closed = Closed;
		
	var extrudeSettings = {
		steps: this.PathArray.length*SegementsPerLine,
		bevelEnabled: false,
		extrudePath: ExtrusionPath
	};
	
	this.TheGeometry = new THREE.ExtrudeBufferGeometry( TheShape, extrudeSettings );
}
//*********************************************************************************
// Main function to compute the extrusion
//*********************************************************************************
/**
* Main function to create the geometry
**/
CM3PathExtrusion.prototype.ComputeGeometry=function()
{
	if (this.TheGeometry==null)
	{
		var Type=this.GetSetting("PathExtrusion","Type","Flat");
		
		//*****************************************
		// Get the positions, normals, and this.IndexArray
		
		switch (Type)
		{
		case "Flat":
			this.ComputeGeometryFlat();
			break;
		case "ThreeJS":
			this.ExtrudeBufferGeometry();
			break;
		case "Pointy":
			this.ComputeGeometryBisecting();
			break;
		default:
			throw("Sorry, the extrusion type "+Type+" is not supported");
			break;
		}
	
		if (Type!="ThreeJS")
		{
			//*****************************************
			// setup the geometry
			
			var TempVertexArray=new Float32Array( this.NumVertices * 3 );
			var TempNormalsArray=new Float32Array( this.NumVertices * 3 );
			 
			for (var i=0;i<this.NumVertices*3;i++)
			{
				TempVertexArray[i]=this.VertexArray[i];
				TempNormalsArray[i]=this.NormalsArray[i];
			}
			this.VertexArray=TempVertexArray;
			this.NormalsArray=TempNormalsArray;
		
			//
			var TempIndexArray=new Uint32Array( this.NumIndexes );
			 
			for (var i=0;i<this.NumIndexes;i++)
			{
				TempIndexArray[i]=this.IndexArray[i];
			}
			this.IndexArray=TempIndexArray;
			
			var ActualVertices=this.VertexArray.length/3;
			
			if (ActualVertices!=this.NumVertices)
			{
				console.log("CM3PathExtrusion.ComputeGeometry: The number of vetices ("+this.NumVertices+") does not match the allocated array size ("+ActualVertices+") ");
			}
			var ActualIndexes=this.IndexArray.length;
			
			if (ActualIndexes!=this.NumIndexes)
			{
				console.log("CM3PathExtrusion.ComputeGeometry: The number of indexes ("+this.NumIndexes+") does not match the allocated array size ("+ActualIndexes+") ");
			}
			//*****************************************
			// setup the geometry
			var TheGeometry=new THREE.BufferGeometry();
			
			var positions = new THREE.BufferAttribute(this.VertexArray, 3 ); // 3 values per position (x,y,z)
			TheGeometry.addAttribute( 'position', positions );
		
			// add the normals to the geometry
			var normals = new THREE.BufferAttribute(this.NormalsArray, 3 ); // 3 values per normal vector
			TheGeometry.addAttribute( 'normal', normals );
			
			// jjg - these do not appear to work
		//TheGeometry.computeFaceNormals(); 
		//	TheGeometry.computeVertexNormals();
			// add the face indexes to the coordinates
			
			TheGeometry.setIndex( new THREE.BufferAttribute( this.IndexArray, 1 ) );
			
			//
			TheGeometry.computeBoundingSphere();
		
			this.TheGeometry=TheGeometry;
		}
	}
}
//************************************************************************************************
// CMBase Functions 
//************************************************************************************************

CM3PathExtrusion.prototype.CMBase_SetSettings=CMBase.prototype.SetSettings;

CM3PathExtrusion.prototype.SetSettings=function(NewSettings,TimeSlice) 
{
	this.CMBase_SetSettings(NewSettings,TimeSlice);
	
	this.TheGeometry=null;
}

CM3PathExtrusion.prototype.CMBase_GetSettingsDefinitions=CMBase.prototype.GetSettingsDefinitions; // save the superclass function

CM3PathExtrusion.prototype.GetSettingsDefinitions=function() 
{
	var Result=this.CMBase_GetSettingsDefinitions(); // get the settings definitions from the superclass
	
	for (GroupKey in CM3PathExtrusion.SettingDefintions)
	{
		Result[GroupKey]=CM3PathExtrusion.SettingDefintions[GroupKey];
	}

	return(Result); 
}

//************************************************************************************************
// Public 
//************************************************************************************************
CM3PathExtrusion.prototype.GetGeometry=function()
{
	this.ComputeGeometry();
	
	return(this.TheGeometry);
}
CM3PathExtrusion.prototype.SetPathArray=function(PathArray)
{
	this.PathArray=PathArray;
	
	this.TheGeometry=null;
}
CM3PathExtrusion.prototype.SetCrossSectionArray=function(CrossSectionArray)
{
	this.CrossSectionArray=CrossSectionArray;
	
	this.TheGeometry=null;
}
CM3PathExtrusion.prototype.GetCrossSectionArray=function()
{
	return(this.CrossSectionArray);
}
