/**************************************************************************************************
* CM3SurfaceGeometry Class
* Creates a rectangular buffered geometry based on a BufferGeometry.  The geometry uses an array of 
* coordinates and a list of face indexes.  Surfaces then use the indexes to save having mulitple
* instances of the coordinates. The geometry will have
* positions,normals, uvs (for textures).
*
* This class is used to create surfaces for DEMs and also extruded geometries
*
* @module CM3SurfaceGeometry
* @Copyright HSU, Jim Graham, 2019
**************************************************************************************************/
//**************************************************************************************************
// CM3SurfaceGeometry class
//
// Creates a rectangular buffered geometry based on a BufferGeometry.  The geometry uses an array of 
// coordinates and a list of face indexes.  Surfaces then use the indexes to save having mulitple
// instances of the coordinates. The geometry will have
// positions,normals, uvs (for textures).
//
// This class is used to create surfaces for DEMs and also extruded geometries
//**************************************************************************************************
//**************************************************************************************************
// Constructor
//**************************************************************************************************
/**
* Creates a CM3SurfaceGeometry object with the specified number of columns and rows.
* The CM3SurfaceGeometry is a subclass of THREE.BufferGeometry.
*
* Adds the following to the object.  The postion and normal values are all zero.
*
*	this.setIndex( FaceIndexArray ); - indexes for faces
*	this.addAttribute( 'position', positions ); - position array (three dimensional coordinates for the surface)
*	this.addAttribute( 'normal', normals ); - normals ot each position
*	this.addAttribute( 'uv', uvs ); - uv locations in a texture 
*
* @public
*/

// NumRows and NumColumns is the number of faces, there are NumRows+1 and NumColumns+1 entries in the vertex and mask arrays

CM3SurfaceGeometry = function ( NumColumns, NumRows,TheMask,StartU,EndU,StartV,EndV,RoundEdges) 
{
	if (StartU==undefined) 
	{
		StartU=0;
		EndU=1;
		StartV=0;
		EndV=1;
	}
	if (RoundEdges==undefined) RoundEdges=false;
	
	this.ThePositions=null;
	
	// construct the object from it's supper class
	THREE.BufferGeometry.call( this );

	// setup the number of rows and columns

	NumColumns = Math.max( 3, Math.floor( NumColumns )  ); // makes columns at least 3
	NumRows = Math.max( 2, Math.floor( NumRows )  ); // rows are at least 2

	this.NumColumns=NumColumns;
	this.NumRows=NumRows;

	// setup the longitude and latitude to be the entire earth
	
	this.type = 'BezierSurfaceGeometry';

	// compute the total number of vertices
	
	var vertexCount = ( ( NumColumns + 1 ) * ( NumRows + 1 ) );

	// create the arrays for the position, normal, and uvs vectors
	
	var positions = new THREE.BufferAttribute( new Float32Array( vertexCount * 3 ), 3 );
	var normals = new THREE.BufferAttribute( new Float32Array( vertexCount * 3 ), 3 );
	
	//this.Positions=positions;
	//this.Normals=normals;
	
	// setup the uv array for the texture to map it to this geometry
	var UVArray=new Float32Array( vertexCount * 2 );
	
	var UVAttribute=new THREE.BufferAttribute( UVArray, 2 );
	
	// create the array to hold the rows of vertex indexes
	
	var VertexIndexRows = []; // temp array rows of indexes to the VertexIndexRows
	var uvArrayIndex=0; // index into the UV array
	var index = 0; // count of the number of total entries (uvArrayIndex*2)

	var UFactor=(EndU-StartU)/NumColumns;
	var VFactor=(EndV-StartV)/NumRows;
	
	for ( var y = 0; y <= NumRows; y ++ ) 
	{
		var VertexIndexRow = []; // temp array to contain one row of vertex indexes

		var v = StartV+(y*VFactor);

		for ( var x = 0; x <= NumColumns; x ++ ) 
		{
			VertexIndexRow.push( index );

			var u=StartU+(x*UFactor);
			
			UVArray[uvArrayIndex++]=u;
			UVArray[uvArrayIndex++]=v;
			
			index ++;
		}
		VertexIndexRows.push( VertexIndexRow );
	}
	// create the array of which faces will be added if masking
	
	var FaceMaskArray=null;
	if (TheMask!=null)
	{
		FaceMaskArray=[];
		
		for ( var y = 0; y < NumRows; y ++ ) 
		{
			var FaceMaskRow=[];
			
			for ( var x = 0; x < NumColumns; x ++ ) 
			{
				if ((TheMask[y][x]!=0)&&(TheMask[y][x+1]!=0)&&(TheMask[y+1][x]!=0)&&(TheMask[y+1][x+1]!=0))
				{
					FaceMaskRow[x]=true;
				}
				else
				{
					FaceMaskRow[x]=false;
				}
			}
			FaceMaskArray.push(FaceMaskRow);
		}
	}
	// create the array of indexes for each face
	
	var FaceIndices = [];

	if (TheMask==null) // no mask, add both triangles for all faces
	{
		for ( var y = 0; y < NumRows; y ++ ) 
		{
			for ( var x = 0; x < NumColumns; x ++ ) 
			{
				var v1 = VertexIndexRows[ y ][ x + 1 ];
				var v2 = VertexIndexRows[ y ][ x ];
				var v3 = VertexIndexRows[ y + 1 ][ x ];
				var v4 = VertexIndexRows[ y + 1 ][ x + 1 ];
	
				FaceIndices.push( v4, v2, v1 );
				FaceIndices.push( v4, v3, v2 );
			}
		}
	}
	else // have a mask
	{
		for ( var y = 0; y < NumRows; y ++ ) 
		{
			for ( var x = 0; x < NumColumns; x ++ ) 
			{
				if (FaceMaskArray[y][x]) // have a mask but all four corners are unmasked
				{
					var TR = VertexIndexRows[ y ][ x + 1 ];
					var TL = VertexIndexRows[ y ][ x ];
					var BL = VertexIndexRows[ y + 1 ][ x ];
					var BR = VertexIndexRows[ y + 1 ][ x + 1 ];
		
					if (RoundEdges==false) // just add both faces
					{
						FaceIndices.push( BR, TL, TR );
						FaceIndices.push( BR, BL, TL );
					}
					else
					{
						// assume all the surrounding pixels are non-masked
						var Left=true;
						var Right=true;
						var Top=true;
						var Bottom=true;
						
						// figure out which directions have a non-masked pixel
						var NumMasked=0;
						
						if ((x==0)||(FaceMaskArray[y][x-1]==0)) 
						{
							Left=false;
							NumMasked++;
						}
						if (((x+1)==NumColumns)||(FaceMaskArray[y][x+1]==0))
						{
							Right=false;
							NumMasked++;
						}
						
						if ((y==0)||(FaceMaskArray[y-1][x]==0)) 
						{
							Top=false;
							NumMasked++;
						}
						if (((y+1)==NumRows)||(FaceMaskArray[y+1][x]==0))
						{
							Bottom=false;
							NumMasked++;
						}
						// three or four surrounding masked pixels mean the face will not be added
						// 2=one triangle added
						// <=1 -> both triangles added
						if (NumMasked<=1) // only one masked still adds both triangles
						{
							FaceIndices.push( BR, TL, TR );
							FaceIndices.push( BR, BL, TL );
						}
						else if (NumMasked==2) // two add one triangle
						{
							if ((Left==false)&&(Top==false)&&(Right)&&(Bottom)) // left and top are empty, add a triangle to the bottom right
							{
								FaceIndices.push( TR, BR, BL );
							}
							else if ((Left)&&(Top==false)&&(Right==false)&&(Bottom)) // top and right are empty, add a triangle to the bottom left
							{
								FaceIndices.push( BR,BL,TL );
							}
							else if ((Left)&&(Top)&&(Right==false)&&(Bottom==false)) // bottom and right are empty, add a triangle to the top-left
							{
								FaceIndices.push( BL,TL,TR );
							}
							else if ((Left==false)&&(Top)&&(Right)&&(Bottom==false)) // bottom and left are empty, add a triangle to the top-right
							{
								FaceIndices.push( TL,TR,BR );
							}
						}
					}
				}
			}
		}
	}

	// create the index array
	
	var FaceIndexArray=null;
	
	if (positions.count<=65535) 
	{
		FaceIndexArray=new THREE.Uint16BufferAttribute( FaceIndices, 1 ); //Uint16BufferAttribute
	}
	else
	{
		FaceIndexArray=new THREE.Uint32BufferAttribute( FaceIndices, 1 );
	}
	this.setIndex( FaceIndexArray );
	
	// set the index array and attributes of the object
	
	this.addAttribute( 'position', positions );
	this.addAttribute( 'normal', normals );
	this.addAttribute( 'uv', UVAttribute );
}

/**
* This class inherites from the THREE.BufferGeometry class
*/
CM3SurfaceGeometry.prototype = Object.create( THREE.BufferGeometry.prototype );

CM3SurfaceGeometry.prototype.constructor = CM3SurfaceGeometry;

//**************************************************************************************************
// Public functions
//**************************************************************************************************
/**
* SetPositions() sets the positions of the coordinates in the surface and computes surface normals to 
* the corners described by the edges coming away from the coordinate.
* @public
* @param - TheArray[NumRows][NumColumns][3] - two dimensional array of coordinates.  The number of
* Rows and Columns must match those specified in the constructor.  TheArray coordinates has the form:
*
* x=TheArray[Row][Column][0],y=TheArray[Row][Column][1],z=TheArray[Row][Column][2]
* 
* This function can be called to update the coordinates as animaion is performed.
*/
CM3SurfaceGeometry.prototype.SetPositions = function (TheArray) 
{
	this.ThePositions=TheArray; // save for interpolation
	
	var NumRows=this.NumRows;
	var NumColumns=this.NumColumns;
	
	// Update the normal vectors
	
	var positions=this.getAttribute( 'position');
	var normals=this.getAttribute( 'normal');
	
	var index = 0; // index into the 
	var normal = new THREE.Vector3(); // temp normal vector
	
	var Vertex1,Vertex2,Vertex3;
	
//	this.TheNormals=[]; // jjg - not used
	
	for ( var y = 0; y <= NumRows; y ++ ) {

		var v = y / NumRows;

		//this.TheNormals[y]=[];
		
		for ( var x = 0; x <= NumColumns; x ++ ) {

			var u = x / NumColumns;

			try
			{
				// update the positions
				var px = TheArray[y][x][0];
				var py = TheArray[y][x][1];
				var pz = TheArray[y][x][2];
				
				positions.setXYZ( index, px, py, pz );
				
				// find the normals
				
				Vertex1=TheArray[y][x];
				
				if (x>0) 
				{
					Vertex2=TheArray[y][x-1];
				}
				else 
				{
					Vertex2=TheArray[y][x+1];
				}
				if (y>0) 
				{
					Vertex3=TheArray[y-1][x];
				}
				else 
				{
					Vertex3=TheArray[y+1][x];
				}
				var Ux=Vertex2[0]-Vertex1[0];
				var Uy=Vertex2[1]-Vertex1[1];
				var Uz=Vertex2[2]-Vertex1[2];
				
				var Vx=Vertex3[0]-Vertex1[0];
				var Vy=Vertex3[1]-Vertex1[1];
				var Vz=Vertex3[2]-Vertex1[2];
				
				var Nx=Uy*Vz-Uz*Vy;
				var Ny=Uz*Vx-Ux*Vz;
				var Nz=Ux*Vy-Uy*Vx;
			
				var normal = new THREE.Vector3(); // temp normal vector
				normal.set( Nx, Ny, Nz ).normalize();
				if ((x==0)||(y==0)) 
				{
					normals.setXYZ( index,- normal.x, -normal.y, -normal.z );
				}
				else
				{
					normals.setXYZ( index, normal.x, normal.y, normal.z );
				}
				//this.TheNormals[y][x]=normal;
			}
			catch (TheException)
			{
				var j=12;
				throw(TheException);
			}
			
			index ++;
		}
	}
	// make sure the buffers are updated
	
	normals.needsUpdate = true;
	positions.needsUpdate = true;
}
//**************************************************************************************************
// Static functions
//**************************************************************************************************
CM3SurfaceGeometry.GetArrayFromGeoJSON=function(TheGeoJSONObject,ZOffset,ZFactor)
{
	// find the parameters for the data
	
	var NumBands=TheGeoJSONObject.NumBands;
	
	var NumColumns=TheGeoJSONObject.WidthInPixels;
	var NumRows=TheGeoJSONObject.HeightInPixels;
	var TheData=TheGeoJSONObject.Data;
	var TheMask=null;
	
	if (TheGeoJSONObject.Mask!=undefined) TheMask=TheGeoJSONObject.Mask;
	
	// get the bounds 
	
	var Bounds=TheGeoJSONObject.Bounds; // 
	
	var XRange=Bounds.XMax-Bounds.XMin;
	var YRange=Bounds.YMax-Bounds.YMin;
	
	var MinZ=TheGeoJSONObject.MinPixelValues;
	var MaxZ=TheGeoJSONObject.MaxPixelValues;
	
	// 
	
	var XOffset=Bounds.XMin;
	var XFactor=XRange/(NumColumns-1);
	
	var YOffset=Bounds.YMin;
	var YFactor=YRange/(NumRows-1);
	
	// create the array and set it up based on the values in the raster
	
	var TheArray=[];
	var TheMaskArray=null;
	
	if (TheMask!=null) TheMaskArray=[];
	
	for ( var y = 0; y < NumRows; y ++ ) 
	{
		TheArray[y]=[];
		if (TheMask!=null) TheMaskArray[y]=[];

		for ( var x = 0; x < NumColumns; x ++ ) 
		{
			TheArray[y][x]=[];
			
			var px = x;//(x-NumColumns/2);
			var py = y; //(y-NumRows/2);
			var pz =TheData[NumRows-y-1][x]/2;

			TheArray[y][x][0]=(px*XFactor)+XOffset;
			TheArray[y][x][1]=((py*YFactor)+YOffset);
			TheArray[y][x][2]=(pz*ZFactor)+ZOffset;
			
			if (TheMask!=null) TheMaskArray[y][x]=TheMask[NumRows-y-1][x];
		}
	}
	var Result=[TheArray,TheMaskArray];
	
	return(Result);
}

/** 
* Adds a surface to a scene.  Requests the elements of a surface from the server and then
* 	creates a CM3SurfaceGeometry object when the data is returned and calls the specified
* 	onload function.
*
* Inputs
* - DEM_URL - URL for a js file containing a DEM that was saved as JS from BlueSpray
* - TheImage - Image to drape over the surface
* - TheBumpMap - Bump map (original DEM - DEM_URL) to apply or null for none
* - TheBumpScale - Exaggeration to apply to the bump map (10 is typical)
* - TheReflectanceMap - A map of the specular reflectance or null
*/
CM3SurfaceGeometry.LoadSurface=function(DEM_URL,Caller,OnLoadFunction)
{
	// request data from server (data was saved as a JSON raster from BlueSpray
	
	var TheRequest=new XMLHttpRequest();
	TheRequest.open("GET",DEM_URL,true);
	TheRequest.TheURL=DEM_URL;
	TheRequest.OnLoadFunction=OnLoadFunction;
	TheRequest.Caller=Caller;
	
	TheRequest.onreadystatechange=function() 
	{
		if( this.readyState == 4)  // done
		{
			if( this.status == 200) // OK
			{
				// get the JSON object from the response
				
				var TheText=TheRequest.responseText;

				var TheGeoJSONObject=JSON.parse(TheText);
				
				// create the surface and add it to the scene
				
				if (OnLoadFunction!=null) this.OnLoadFunction(TheGeoJSONObject,this.Caller);
				
//					edges = new THREE.VertexNormalsHelper( TheSurface, 20, 0x00ff00, 1 );
////					edges = new THREE.FaceNormalsHelper( mesh, 2, 0x00ff00, 1 );
//					TheScene.add(edges);
			}
			else alert("HTTP error "+this.status+" "+this.statusText+" ("+this.TheURL+")");
		}
	}
	TheRequest.send();
}

