/****************************************************************************************************
* CM3Utilities Class
*
* Utilities to help with 3D rendering
*
* @module CM3Utilities
* @Copyright HSU, Jim Graham, 2019
****************************************************************************************************/
//****************************************************************************************************
// Constructor so we can add static definitions
//****************************************************************************************************
/**
* Constructor so we can add static definitions
*
* @public, @constructs
*/
function CM3Utilities() 
{
	
}
//****************************************************************************************************
// Public static functions
//****************************************************************************************************
/**
* Constructor so we can add static definitions
*
* @public, @constructs
*/
CM3Utilities.GetRegularPolygonShape=function(NumPoints,Size,CenterX,CenterY,StartAngle)
{
	var Points2D=CMUtilities.GetRegularPolygon(NumPoints,Size,CenterX,CenterY,StartAngle);

	var TheShape = new THREE.Shape();
	
	var Xs=Points2D[0];
	var Ys=Points2D[1];
	
	TheShape.moveTo(Math.round(Xs[0]),Math.round(Ys[0]) );
	
	for (var j=1;j<Points2D.length;j++)
	{
		TheShape.lineTo(Math.round(Xs[j]),Math.round(Ys[j] ));
//		ThePoints.push( new THREE.Vector2 ( Points2D[0][j], Points2D[1][j] ) );
	}
	TheShape.lineTo(Math.round(Xs[0]),Math.round(Ys[0]) );
	
//	var TheShape = new THREE.Shape( ThePoints );
	
	return(TheShape);
}
/**
* 
*
* @public
*/
CM3Utilities.Get3DColorFromCanvasColor=function(TheColor)
{
	var Opacity=1;
	var Transparent=false;
	
	if (typeof(TheColor) === 'string')
	{
		var Red=0;
		var Green=0;
		var Blue=0;
		
		// convert from HSL if needed
		var Index1=TheColor.indexOf("hsl(");
										 
		if (Index1!=-1) // hsl color
		{
			var Index2=TheColor.indexOf(")");
			var Tokens=TheColor.substring(Index1+4,Index2);
			Tokens=Tokens.split(",");
			
			var h=Tokens[0];
			var s=Tokens[1];
			var l=Tokens[2];
			
			if (s.indexOf("%")!=-1)
			{
				s=s.replace("%","");
				s=s/100;
			}
			if (l.indexOf("%")!=-1)
			{
				l=l.replace("%","");
				l=l/100;
			}
			var RGB=CMUtilities.HSLToRGB(h,s,l);
			Red=RGB[0];
			Green=RGB[1];
			Blue=RGB[2];
		}
		else
		{
			var TheValues=CMUtilities.GetRGBAValuesFromRGBA(TheColor);
			Red=TheValues.Red;
			Green=TheValues.Green;
			Blue=TheValues.Blue;
			
			Opacity=TheValues.Transparency;
			
			if (Opacity!=1) Transparent=true;
		}
		
		TheColor="rgb("+Red+","+Green+","+Blue+")";
																		
	}
	return({RGB:TheColor,Opacity:Opacity,Transparent:Transparent});
}
//****************************************************************************************************
// Public static functions
//****************************************************************************************************
/**
 * Computes the area of a "patch" on the surface of the earth that is defined
 * by two longitudes and two latitudes.
 * 
 * The equation herein was taken from: http://mathforum.org/library/drmath/view/63767.html
 * 
* @public
 * @param Lon1 -90 to 90
 * @param Lat1
 * @param Lon2 -180 to 180
 * @param Lat2 
 */
CM3Utilities.GreatArcArea=function( Lon1, Lat1, Lon2, Lat2, Radius)
{
	var Area=0;
	
	if ((Lat1<0)&&(Lat2>0)) // must two two calcs
	{
		var Area1=CM3Utilities.GreatArcArea(Lon1,-Lat1,Lon2,0,Radius);
		var Area2=CM3Utilities.GreatArcArea(Lon1,0,Lon2,Lat2,Radius);
		
		Area=Area1+Area2;
	}
	else if ((Lat1>0)&&(Lat2<0)) // must two two calcs
	{
		var Area1=CM3Utilities.GreatArcArea(Lon1,Lat1,Lon2,0,Radius);
		var Area2=CM3Utilities.GreatArcArea(Lon1,0,Lon2,-Lat2,Radius);
		
		Area=Area1+Area2;
	}
	else  // both lats are in the same hemisphere
	{
		
		// find the absolute difference between the longitudes
		
		var LonDelta=Math.abs(Lon1-Lon2);
		if (LonDelta>180) LonDelta=360-LonDelta;
		
		// find the heights from each lattitude 
		
		var h1 = Radius*(1-Math.sin(CMProjector.DegreesToRadians(Lat1)));
		var h2 = Radius*(1-Math.sin(CMProjector.DegreesToRadians(Lat2)));
		
		var A1 = 2*Math.PI*Radius*h1;
		var A2 = 2*Math.PI*Radius*h2;

		// find the area between the two latitudes
		
		var Area360=Math.abs(A2-A1);
		
		var AreaPolygon=Area360*LonDelta/360; 
		
		Area=AreaPolygon;
	}
	return(Area);
}
/**
 * Below is the haversine formula which produces better results for small distances.
 * @param Lon1
 * @param Lat1
 * @param Lon2
 * @param Lat2
 * @param Radius - radius of the earth in the same units as returned (typically meters)
 * @return 
 */
CM3Utilities.GreatArcDistance=function( EastLongitude, Lat1, WestLongitude, Lat2, Radius)
{
	// convert to radians
	EastLongitude=EastLongitude*Math.PI/180;
	Lat1=Lat1*Math.PI/180;
	WestLongitude=WestLongitude*Math.PI/180;
	Lat2=Lat2*Math.PI/180;
	
	// change along lines of latitude and longitude
	DeltaLon=Math.abs(WestLongitude-EastLongitude);
	DeltaLat=Math.abs(Lat2-Lat1);
	
	//double Temp1=sqr(sin(DeltaLat/2))+cos(Lat1)*cos(Lat2)*sqr(sin(DeltaLon/2));
	
	Temp1=Math.pow(Math.sin(DeltaLat/2),2)+Math.cos(Lat1)*Math.cos(Lat2)*Math.pow(Math.sin(DeltaLon/2),2);
	
	Angle=2*Math.asin(Math.sqrt(Temp1));
	
	if (isNaN(Angle))
	{
		throw new Exception("Sorry, the angle could not be computed");
	}
	Distance=Radius*Angle;
	
	return(Distance);
}

//****************************************************************************************************
// Public static functions
//****************************************************************************************************

CM3Utilities.GetArcArrayInDegrees=function(StartAngle,EndAngle,NumSteps,Radius)
{
	StartAngle=StartAngle*Math.PI/180;
	EndAngle=EndAngle*Math.PI/180;
	
	var TheCurve=CM3Utilities.GetArcArrayInRadians(StartAngle,EndAngle,NumSteps,Radius);
	return(TheCurve);
}

/**
* Creates an array of THREE.Vector3 coordinates that form a smooth arc
* @public
*/
CM3Utilities.GetArcArrayInRadians=function(StartAngle,EndAngle,NumSteps,Radius)
{
	var TheCurve=[];
	
	var Angle=StartAngle;
	var Step=(EndAngle-StartAngle)/NumSteps;
	
	for (var i=0;i<=NumSteps;i++)
	{
		var x=Math.sin(Angle)*Radius;
		var y=Math.cos(Angle)*Radius;
		
		TheCurve.push(new THREE.Vector3( x,y,0 ));
		
		Angle+=Step;		
	}
	return(TheCurve);
}
//****************************************************************************************************
// Public static functions for managing 3D text
//****************************************************************************************************
