/**
* CM3View Class
*
* Class to manage a 3D view.  This includes:
* - Three.js camera
* - Three.js Renderer which renders the scene into the camera (which is tied to the view's element)
* - ViewingPlatform - optional platform to allow the camera to move indepenently from the direction of the platform.
*  		Think of an airplane with a pilot, the plane is moving in one direction but the pilot can look in other directions. 
* - ViewingControls - these are based on the Three.js classes but modified to allow for spatial actions
*
* For mulitple views, these need to move here:
* - CanvasElement - the element the scene is displayed in
* - CanvasContainer - the DIV that contains the canvas object.  This is required to capture resize events and then resize the canvas element.
*
* @module CM3View
* @Copyright HSU, Jim Graham, 2019
*/
//******************************************************************
// Definitions
//******************************************************************
CM3View.MOUSE_DRAGGING=1;
CM3View.MOUSE_IN=2;
CM3View.MOUSE_OUT=3;
CM3View.MOUSE_OVER=4;

/**
* Below are the settings definitions.
* @public, @settings
*/
CM3View.SettingDefintions=
{
	CM3View:
	{
		CameraType: { Name:"Camera Type",Type:CMBase.DATA_TYPE_ENUMERATED, Options:["Perspective","Orthagonal"],Default:"Perspective" },
		CameraUp: { Name:"Camera Up",Type:CMBase.DATA_TYPE_ENUMERATED, Options:["X","Y","Z"],Default:"Z" },
		ControlType: { Name:"Control Type",Type:CMBase.DATA_TYPE_ENUMERATED, Options:["Orbital","Flying","Helicopter"],Default:"Orbital" },
		NearDistance: { Name:"Near Distance",Type:CMBase.DATA_TYPE_FLOAT, Default:0.1 },
		FarDistance: { Name:"FarDistance",Type:CMBase.DATA_TYPE_FLOAT, Default:200000000 }, // Infinity produces nothing
		
		LookAt: { Name:"Look At",Type:CMBase.DATA_TYPE_VECTOR, Default:null }, // position to look at on initialization
	},
	PerspectiveCamera:
	{
		ViewAngle: { Name:"View Angle",Type:CMBase.DATA_TYPE_FLOAT, Default:45 },
	},
	OrthagonalCamera:
	{
		FrustumSize: { Name:"Frustum Size",Type:CMBase.DATA_TYPE_FLOAT, Default:500 },
	}
};
//*****************************************************************************************************
// Constructor
//*****************************************************************************************************
/*
* Constructor
* @public, @constructs
*/
function CM3View() 
{
	CMView.call(this);
	
	//***********************************************************************************************
	//  Settings
	//***********************************************************************************************
	this.Settings.CM3View=	
	{
		CameraType:"Perspective",
		CameraUp:"Z",
		ControlType:"Orbital",
		NearDistance:0.1,
		FarDistance:2000000
	};
	this.Settings.PerspectiveCamera=	
	{
		ViewAngle:45,
	};
	this.Settings.OrthagonalCamera=	
	{
		FrustumSize:500,
	};	
	//***********************************************************************************************
	// Private Properties
	//***********************************************************************************************
	
	this.TheCamera=null; // the camera that is displaying content in the view
	
	this.TheControls=null; // controls for the user to move the camera
	
	this.TheRenderer; // renders the scene into the view
	
	this.TheCanvasContainer; // the DOM element that contains the 3d view of a 3d scene, required to update cursor, coordinate conversion and contains the renderer
	this.CanvasElement;
	
	// Variables for selecting objects
	this.SelectedOGLObject; // the item that is slected
	this.MouseOverOGLObject; // item the mouse is over
	
	this.DraggingPlane=new THREE.Plane(); // Plane that objects are dragged within (normal plane to the camera at the item intersected with mouse?
	this.TheRayTracer=new THREE.Raycaster(); // object to trace rays from camera to 3d objects
 
	this.intersection=new THREE.Vector3();
	
	// for flying control
	this.clock=new THREE.Clock();
}

CM3View.prototype=new CMView();

CM3View.prototype.constructor=CM3View;
//***********************************************************************************************
// Protected Rendering Functions
//***********************************************************************************************
CM3View.prototype.Render=function()
{
	if (this.TheRenderer!=undefined)
	{
		var delta=this.clock.getDelta();
		
		if (this.TheControls!=null) 
		{
			this.TheControls.Update(delta); // allows for animated moves
		}
		// render the scene
		try
		{
			var TheScene=this.GetParent(CM3Scene);
			
			var TheOGLScene=TheScene.GetOGLScene();
		
			this.TheRenderer.render(TheOGLScene, this.TheCamera);
		}
		catch(err)
		{
			console.log(err);
		}
		
/*			if (Speed!=0)
		{
			var CameraDirection=TheCamera.getWorldDirection();
			CameraDirection.multiplyScalar(-Speed);
			yawObject.position.add(CameraDirection);
		}
*/	}
}

//***********************************************************************************************
// Setup
//***********************************************************************************************
/**
* Sets up the View3D for rendering.  
*/
CM3View.prototype.Setup=function(NewCanvasElement,NewContainer)
{
	// Save the container and canvas element
	this.TheCanvasContainer=NewContainer;
	this.TheCanvasElement=NewCanvasElement;
	
	this.TheCanvasContainer.TheView=this; // must have the view tied to this DOM element for KeyDown event
	
	this.TheCanvasElement.style.cursor="crosshair";
	this.SetTool(CMView.TOOL_SELECT);
	
	var TheScene=this.GetParent(CM3Scene);

	//************************************************************************
	// create the TheRenderer and insert the domElement/Canvas into the document
	
	var ViewWidth=this.TheCanvasElement.width;
	var ViewHeight=this.TheCanvasElement.height;
	
	if (this.TheRenderer==null) // has not been specified
	{
		var Parameters={ antialias: true,logarithmicDepthBuffer: true };
		if (this.TheCanvasElement!=undefined) Parameters["canvas"]=this.TheCanvasElement;
		
		var TheColor=TheScene.GetSetting("Style","fillStyle",0x888888);
		
		this.TheRenderer=new THREE.WebGLRenderer(Parameters);
		this.TheRenderer.setClearColor(TheColor, 1); // second param is opacity, 0=> transparent
		
		this.TheRenderer.setPixelRatio(window.devicePixelRatio);
		this.TheRenderer.setSize(ViewWidth, ViewHeight);
		
		// add the canvas element to the container if it was not specified
		if (this.TheCanvasElement==undefined) this.TheCanvasContainer.appendChild(this.TheRenderer.domElement);
				
		this.TheRenderer.setSize(ViewWidth,ViewHeight); // 
	}
	//************************************************************************
	// Create the camera
	
	if (this.TheCamera==null)
	{
		var Aspect=ViewWidth / ViewHeight;
		
		// perspective camera settings
		
		var CameraType=this.GetSetting("CM3View","CameraType");
		
		var NearDistance=this.GetSetting("CM3View","NearDistance",0.1);
		var FarDistance=this.GetSetting("CM3View","FarDistance");
		
		switch (CameraType)
		{
		case "Perspective":
			var ViewAngle=this.GetSetting("PerspectiveCamera","ViewAngle",45);
			var CameraUp=this.GetSetting("CM3View","CameraUp","Z");
			
			this.TheCamera=new THREE.PerspectiveCamera(ViewAngle, Aspect, NearDistance, FarDistance);
			if (CameraUp=="Z")
			{
				this.TheCamera.up.set(0, 0, 1);
			}
			else
			{
				this.TheCamera.up.set(0, 1, 0);
			}
			break;
			
		case "Orthagonal":
			var FrustumSize=this.GetSetting("OrthagonalCamera","FrustumSize");
			var aspect=window.innerWidth / window.innerHeight;
			this.TheCamera=new THREE.OrthographicCamera(FrustumSize * aspect / - 2, FrustumSize * aspect / 2, FrustumSize / 2, FrustumSize / - 2,
				NearDistance, FarDistance);
			break;
		}
		this.TheCamera.position.set(0,200,400); // is overriden when data loaded
		this.TheCamera.lookAt(TheScene.GetPosition());	
	}
	
	//************************************************************************
	// Create the control and connect it to the camera and DOM element
	
	if (this.TheControls==null) // have not been intialized
	{
		var ControlType=this.GetSetting("CM3View","ControlType","Orbital");
		
		if (ControlType=="Flying")
		{
			this.TheControls=new CM3ControlsAirplane();
		}
		else if (ControlType=="Orbital")
		{
			this.TheControls=new CM3ControlsOrbit();
		}
		else if (ControlType=="Helicopter")
		{
			this.TheControls=new CM3ControlsHelicopter();
		}
	}
	this.TheControls.SetParent(this);
	this.TheControls.Initialize(this.GetCamera(),this.TheRenderer.domElement);
}
//******************************************************************
// CMBase Settings Functions
//******************************************************************

CM3View.prototype.CMView_GetSettingsDefinitions=CMView.prototype.GetSettingsDefinitions; // save the superclass function

CM3View.prototype.GetSettingsDefinitions=function() 
{
	var Result=this.CMView_GetSettingsDefinitions(); // get the settings definitions from the superclass
	
	for (GroupKey in CM3View.SettingDefintions)
	{
		Result[GroupKey]=CM3View.SettingDefintions[GroupKey];
	}

	return(Result); 
}
CM3View.prototype.CMView_SetSetting=CMView.prototype.SetSetting;

CM3View.prototype.SetSetting=function(Group,Key,Value) 
{
	this.CMView_SetSetting(Group,Key,Value);
	
	if (Group=="Item")
	{
		if (Key=="PositionOffset")
		{
			if (this.TheControls!=null)
			{
				this.TheControls.SetPosition(Value[0],Value[1],Value[2]);
			}
		}
		else if (Key=="Rotation")
		{
			if (this.TheControls!=null)
			{
				this.TheControls.SetPosition(Value[0],Value[1],Value[2]);
			}
		}
	}
	else if (Group=="CM3View")
	{
		if (Key=="LookAt")
		{
			if (this.TheControls!=null)
			{
				this.TheControls.LookAt(Value[0],Value[1],Value[2]);
			}
		}
	}
}

//*******************************************************************************
// CMVeiw Functions
//*******************************************************************************

CM3View.prototype.SetStyle=function(NewStyle,SaveCurrentStyle) 
{
}
/**
* Saves the current style to a stack
* @public
*/
CM3View.prototype.SaveStyle=function() 
{
}
/**
* Restores the style from the stack
* @public
*/
CM3View.prototype.RestoreStyle=function() 
{
}
//*******************************************************************************
// CMView Functions
//*******************************************************************************

CM3View.prototype.Resize=function()
{
	// the canvas aspect ratio is not correct unless we set the "width" and "height" of the element
	// rather than the style.  This must be done or it will display maps distorted
	
	var TheParent=this.TheCanvasElement.parentNode;
	var ParentWidth=jQuery(TheParent).width();
	var ParentHeight=jQuery(TheParent).height();

	this.TheRenderer.setSize(ParentWidth,ParentHeight); // ths appears to reallocate the HTML canvas element
	
	this.TheCamera.aspect=ParentWidth / ParentHeight;
	
	this.TheCamera.updateProjectionMatrix();
}

//*******************************************************************************
// CMView Functions
//*******************************************************************************
/*
* Zoom in to a higher resolution (map units are doubled relative to pixels)
* @public
*/
CM3View.prototype.ZoomIn=function()
{
	if (this.TheControls!=null)
	{
		this.TheControls.ZoomIn();
	}
}
/*
* Zoom in to a lower resolution (map units are halved relative to pixels)
* @public
*/
CM3View.prototype.ZoomOut=function()
{
	if (this.TheControls!=null)
	{
		this.TheControls.ZoomOut();
	}
}
/**
* Zoom to the maxum bounds that have been set in the veiw
* @public
*/
CM3View.prototype.ZoomToMaxBounds=function()
{
	var TheScene=this.GetParent(CMScene);
	
	var TheBounds=TheScene.GetBounds();
	
	if (TheBounds!=null)
	{
		var CenterX=(TheBounds.XMax+TheBounds.XMin)/2;
		var CenterY=(TheBounds.YMax+TheBounds.YMin)/2;
		var Depth=TheBounds.YMax-TheBounds.YMin;
		
		this.SetSetting("Item","PositionOffset",[CenterX,CenterY-Depth,Depth/2]); //(x,y,z) left and right(x), up and down(z), in and out but always looking at same spot(y)
//		this.SetPosition();
	}
}
/**
* Zooms the view to the specified bounds.  Selects a zoom level that will
* contain the entire map.
* @public
* @param NewBounds - 
*/
CM3View.prototype.ZoomToBounds=function(NewBounds)
{
	if (NewBounds==null) 
	{
		alert("Sorry, you cannot call CM3View.ZoomToBounds(NewBounds) with NewBounds=null.");
	}
	else
	{
/*		var width=this.TheCanvasElement.width;
		var height=this.TheCanvasElement.height;	
		
		// Determine how much to TheScale our coordinates by
		var TheScale=Math.abs(NewBounds.XMax - NewBounds.XMin)/width;
		var yScale=Math.abs(NewBounds.YMax - NewBounds.YMin)/height;
		
		if (TheScale < yScale)  TheScale=yScale; // if xScale < yScale, use xScale, else use yScale
		
		// zoom out until we find a zoom level that is appropriate
		var OldZoomLevel=this.ZoomLevel;
		
		var MinZoom=this.GetSetting("View2D","MinZoom");
		var MaxZoom=this.GetSetting("View2D","MaxZoom");
		
		this.ZoomLevel=20;
		if (this.ZoomLevel>MaxZoom) this.ZoomLevel=MaxZoom;
		
		while ((this.GetScale()<TheScale)&&(this.ZoomLevel>MinZoom))
		{
			this.ZoomLevel--;
		}
		
		// reset the dataset requests if the zoom level changed
		if (this.ZoomLevel!=OldZoomLevel) // zoom level changed
		{
			CMDataset.ResetRequests();
		}
		// reset the center of the map to the center of the bounds
*/		

//	TheCMScene3D.GetView(0).SetPosition(X,Y,4000); // in and out of screen

		var CenterRefX=(NewBounds.XMax+NewBounds.XMin)/2;
		var CenterRefY=-(NewBounds.YMin+NewBounds.YMax)/2;; // Views are added to the scene so Y is negative
		
		var Width=NewBounds.XMax-NewBounds.XMin;
		var Z=Width;
		
		this.SetSetting("Item","PositionOffset",[CenterRefX,CenterRefY,Z]); //(x,y,z) left and right(x), up and down(z), in and out but always looking at same spot(y)
//		this.SetPosition(CenterRefX,CenterRefY,Z);
		
		this.SetSetting("Item","Rotation",[0,0,0]); 
		this.SetSetting("CM3View","LookAt",[CenterRefX,CenterRefY,0]); 
		this.SetRotation(0,0,0);
		
//		this.LookAt(CenterRefX,CenterRefY,0); 
//		var TheCenter=new THREE.Vector3(CenterRefX, CenterRefY, 0);
		
		//this.TheCamera.position.set(CenterRefX,CenterRefY,4000);
//		this.TheCamera.position.set(CenterRefX,CenterRefY,4000);
		
		// these do not appear to work
//		this.TheCamera.lookAt(TheCenter);	
//		this.TheCamera.rotateY(Math.PI/2); //Math.PI/2;
	}
}
//*****************************************************************************
// Event Handlers
//*****************************************************************************
/**
*
* this pointer is the Container object
*/
CM3View.prototype.MouseDown=function(TheEvent)
{
	var Used=false;
	
	var TheScene=this.GetParent(CM3Scene);
	
	TheEvent.preventDefault();

	var ViewWidth=this.TheRenderer.domElement.width;
	var ViewHeight=this.TheRenderer.domElement.height;
	
	// convert the coordinate to be local to the canvas
	
	var Coordinate=CMUtilities.GetElementCoordinate(TheEvent.clientX,TheEvent.clientY,this.TheCanvasContainer);
	
	// convert the coordiate to be 0 to 1 in the viewing panel (camera lens)
	
	var MouseCameraCoordinate=new THREE.Vector2(); 
	MouseCameraCoordinate.x=(Coordinate.x / ViewWidth) * 2 - 1;
	MouseCameraCoordinate.y=- (Coordinate.y / ViewHeight) * 2 + 1;

	// move the camera to the mouse coordinate
	
	this.TheRayTracer.setFromCamera(MouseCameraCoordinate, this.TheCamera); // setup ray tracer to point from mouse along camera's direction

//	Used=TheScene.MouseDown(this,TheEvent,this.TheRayTracer);
		
	Used=TheScene.SendMessageToDescendants(CMView.MESSAGE_MOUSE_DOWN,{TheView:this,TheEvent:TheEvent,TheRayTracer:this.TheRayTracer});

/*	// find something that intersects with the mouse ray
	var TheGeo=TheScene.GetGeo();
	TheGeo.OGLGeo.updateMatrixWorld ( true);
	
	var _inverseMatrix = new THREE.Matrix4();
	_inverseMatrix.getInverse( TheGeo.OGLGeo.matrixWorld );
this.TheRayTracer.ray.applyMatrix4( TheGeo.OGLGeo.matrixWorld );
	
//this.TheRayTracer.ray.applyMatrix4( TheGeo.OGLGeo.matrixWorld ); 
	
	var IntersectedObjects=this.TheRayTracer.intersectObjects(TheScene.SelectableObjects); // find the objects the ray intersects with

	if (IntersectedObjects.length>0) // if something was intersected, select it
	{
		if ((IntersectedObjects[0].object.TheCM3Item!=undefined))
		{
			this.SelectedOGLObject=IntersectedObjects[0].object; // select the intersected object as the selected object
	
			if (this.TheRayTracer.ray.intersectPlane(this.DraggingPlane, this.intersection)) 
			{
				Used=this.SelectedOGLObject.TheCM3Item.MouseDown(this,TheEvent,this.intersection,this.SelectedOGLObject);
			}
		}
	}*/
	if (Used==false) Used=this.TheControls.MouseDown(TheEvent);
	
	this.Repaint();
	
	return(Used);
}
/**
* Handles mouse movement events in the view
* @private
*/

CM3View.prototype.MouseMove=function(TheEvent)
{
	var Used=false;
	
	var TheScene=this.GetParent(CM3Scene);
	
//		if ((this.TheControls.getState()==-1)) //this.TheControls.STATE.NONE
	{
		var ViewWidth=this.TheRenderer.domElement.width;
		var ViewHeight=this.TheRenderer.domElement.height;
		
		TheEvent.preventDefault(); // make it so the default MouseCameraCoordinate behavior does not occur
	
		// find the MouseCameraCoordinate position within the element
		var Coordinate=CMUtilities.GetElementCoordinate(TheEvent.clientX,TheEvent.clientY,this.TheCanvasContainer);
		var MouseCameraCoordinate=new THREE.Vector2(); 
		MouseCameraCoordinate.x=(Coordinate.x / ViewWidth) * 2 - 1;
		MouseCameraCoordinate.y=- (Coordinate.y / ViewHeight) * 2 + 1;
	
		// setup the ray tracer to go from the camera through the mouse cursor
		this.TheRayTracer.setFromCamera(MouseCameraCoordinate, this.TheCamera);
	
		//
		var planeZ = new THREE.Plane(new THREE.Vector3(0,0,1),0); //default plane should be ((1,0,0), 0)
		var intersects = new THREE.Vector3();
		this.TheRayTracer.ray.intersectPlane(planeZ, intersects);
			
//		console.log(this.TheRayTracer.ray); 
//		console.log(intersects);
		
		this.SendMessageToListeners(CMView.MESSAGE_MOUSE_MOVE,{TheView:this,TheEvent:TheEvent,TheRayTracer:this.TheRayTracer});
		
		Used=TheScene.SendMessageToDescendants(CMView.MESSAGE_MOUSE_MOVE,{TheView:this,TheEvent:TheEvent,TheRayTracer:this.TheRayTracer});
		
/*		if (this.MouseOverOGLObject!=null)
		{
			if (this.MouseOverOGLObject.TheCM3Item==undefined)
			{
				var j=12;
			}
		}
		if (this.SelectedOGLObject)  // if something is already selected, move it
		{
			if (this.TheRayTracer.ray.intersectPlane(this.DraggingPlane, this.intersection))  // jjg - this appears to be undocumented!
			{
				Used=this.SelectedOGLObject.TheCM3Item.MouseMove(this,TheEvent,this.intersection,CM3View.MOUSE_DRAGGING);
			}
		}
		else // nothing selected, see if there is an this.intersection
		{
			// get the SelectableObjects that insersect with the ray
			var IntersectedObjects=this.TheRayTracer.intersectObjects(TheScene.SelectableObjects);
			
			// see if we have a new MouseOverOGLObject
			if (IntersectedObjects.length>0) // the mouse is over an object
			{
				if (this.MouseOverOGLObject!=IntersectedObjects[0].object) // not the currently intersected object
				{
					if (this.MouseOverOGLObject)  // moved out of the current object and into a new object
					{
						Used=this.MouseOverOGLObject.TheCM3Item.MouseMove(this,TheEvent,this.intersection,CM3View.MOUSE_OUT);
						this.MouseOverOGLObject=null;
					}
					// setup the new mouse over object
					this.MouseOverOGLObject=IntersectedObjects[0].object;
					
					Used=this.MouseOverOGLObject.TheCM3Item.MouseMove(this,TheEvent,this.intersection,CM3View.MOUSE_IN);
					
					// setup the dragging plane to be parallel to the camera and contain the object
					this.DraggingPlane.setFromNormalAndCoplanarPoint(
						this.TheCamera.getWorldDirection(this.DraggingPlane.normal), // normal to the this.DraggingPlane
						this.MouseOverOGLObject.position); // point on the this.DraggingPlane (coplanar point)
				}
				else // over the current mouse over object
				{
					Used=this.MouseOverOGLObject.TheCM3Item.MouseMove(this,TheEvent,this.intersection,CM3View.MOUSE_OVER);
				}
			} 
			else // mouse is not over anything, clear MouseOverOGLObject property
			{
				if (this.MouseOverOGLObject) 
				{
					Used=this.MouseOverOGLObject.TheCM3Item.MouseMove(this,TheEvent,this.intersection,CM3View.MOUSE_OUT);

					this.MouseOverOGLObject=null;
				}
			}
		}
*/	}
//	this.Repaint();
	
	if (Used==false) Used=this.TheControls.MouseMove(TheEvent);
	
	/*TheCamera=this.TheCamera;
	var ThePosition=THREE.Vector3(100,0,20);
	
	_vector3 = new THREE.Vector3();
	_viewMatrix = new THREE.Matrix4();
	_viewProjectionMatrix = new THREE.Matrix4();
	
	_viewMatrix.copy( TheCamera.matrixWorldInverse );
	_viewProjectionMatrix.multiplyMatrices( TheCamera.projectionMatrix, _viewMatrix );
	
	var geometry = new THREE.BoxGeometry( 1, 1, 1 );
	var material = new THREE.MeshBasicMaterial( {color: 0x00ff00} );
	var cube = new THREE.Mesh( geometry, material );
	cube.position.x=100;
	cube.position.y=0;
	cube.position.z=20;

	_vector3.setFromMatrixPosition( cube.matrixWorld );
			_vector3.applyMatrix4( _viewProjectionMatrix );

	var x = _vector3.x ;
	var y = - _vector3.y ;
		*/
//		console.log("x="+x+" y="+y);
		
//		this.TheContainer.style.left=x+"px";
//		this.TheContainer.style.top=y+"px";

	return(Used);
};
CM3View.prototype.GetScreenPosition=function(cube)
{
	TheCamera=this.TheCamera;
	//var ThePosition=THREE.Vector3(100,0,20);

	_vector3 = new THREE.Vector3();
	_viewMatrix = new THREE.Matrix4();
	_viewProjectionMatrix = new THREE.Matrix4();
	
	_viewMatrix.copy( TheCamera.matrixWorldInverse );
	_viewProjectionMatrix.multiplyMatrices( TheCamera.projectionMatrix, _viewMatrix );

/*var geometry = new THREE.BoxGeometry( 1, 1, 1 );
var material = new THREE.MeshBasicMaterial( {color: 0x00ff00} );
var cube = new THREE.Mesh( geometry, material );
cube.position.x=100;
cube.position.y=0;
cube.position.z=20;
*/
	_vector3.setFromMatrixPosition( cube.matrixWorld );
	_vector3.applyMatrix4( _viewProjectionMatrix );

	var x = _vector3.x ;
	var y = - _vector3.y;
		
//	console.log("x="+x+" y="+y);
	
	return([x,y]);
}

/**
* Handles mouse up events in the view
* @private
*/
CM3View.prototype.MouseUp=function(TheEvent)
{
	var Used=false;
	
	var TheScene=this.GetParent(CM3Scene);
	
	Used=TheScene.SendMessageToDescendants(CMView.MESSAGE_MOUSE_UP,{TheView:this,TheEvent:TheEvent,TheRayTracer:this.TheRayTracer});
//	Used=TheScene.MouseUp(this,TheEvent);
	
	// if there is a selected object, clear it and reset the cursor
/*	if (this.SelectedOGLObject!=null) 
	{
		this.SelectedOGLObject=null;
		
		this.TheCanvasContainer.style.cursor='auto';
		
		TheEvent.preventDefault();
		Used=true;
	}
		
	this.Repaint();
*/	
	if (Used==false) Used=this.TheControls.MouseUp(TheEvent);
	
	return(Used);
}
CMView.prototype.MouseWheel=function(TheEvent)
{
	var Used=false;
	
	Used=this.TheControls.MouseWheel(TheEvent);
	
	return(Used);
}

CM3View.prototype.KeyDown=function(TheEvent)
{
	var Used=false;
	
	Used=this.TheControls.KeyDown(TheEvent);
	
	return(Used);
}
CM3View.prototype.KeyUp=function(TheEvent)
{
	var Used=false;
	
	Used=this.TheControls.KeyUp(TheEvent);
	
	return(Used);
}
//************************************************************************************
// 
//************************************************************************************

CM3View.prototype.PaintBackground=function()
{
	var Used=false;
};
CM3View.prototype.PaintStart=function() 
{
	//this.Render();
}
CM3View.prototype.PaintEnd=function() 
{
	this.Render();
}
//************************************************************************************
// CM3View functions that need to move to settings
//************************************************************************************
/**
* Sets the camera to the specified x,y,z location in the 3D scene.
* @public
*/
CM3View.prototype.SetPosition=function(x,y,z) 
{
	this.SetSetting("Item","PositionOffset",[x,y,z]);
//	this.TheControls.SetPosition(x,y,z);
}

CM3View.prototype.GetPosition=function() 
{
	return(this.TheControls.GetPosition());
}
/**
* @public
*/
CM3View.prototype.SetRotation=function(x,y,z) 
{
	this.SetSetting("Item","Rotation",[x,y,z]);
//	this.TheControls.SetRotation(x,y,z);
}
/**
* Points the camera to the specified x,y,z coordinate
* @public
*/
CM3View.prototype.LookAt=function(x,y,z) 
{
	this.SetSetting("CM3View","LookAt",[x,y,z]);
//	this.TheControls.LookAt(x,y,z);
}
//************************************************************************************
// CM3View functions
//************************************************************************************

/**
* Get the current renderer from the view
*/
CM3View.prototype.GetRenderer=function() 
{
	return(this.TheRenderer);
}
CM3View.prototype.SeteRenderer=function(TheRenderer) 
{
	this.TheRenderer=TheRenderer;
}
CM3View.prototype.GetCamera=function() 
{
	return(this.TheCamera);
}
CM3View.prototype.SetCamera=function(TheCamera) 
{
	this.TheCamera=TheCamera;
}

CM3View.prototype.GetControls=function() 
{
	return(this.TheControls);
}
CM3View.prototype.SetControls=function(TheControls) 
{
	this.TheControls=TheControls;
}

CM3View.prototype.SetCursor=function(NewCursor) 
{
	this.TheCanvasContainer.style.cursor=NewCursor;
}

